UNIT BLSunit;
{ This unit contains a procedure for the solution of
  linear least squares problems with simple bounds (BLS).
  The algorithm is taken from
    P.A.Bjoerck: Least Squares Methods

    IN: Handbook of numerical analysis
        Vol.1: Solution of equations in R^n
        Ed.: P.G.Ciarlet and J.L.Lions
        Elsevier/North Holland 1987
}

INTERFACE



USES error,   { error handler                                                 }
     math,    { low level mathematical subroutines                            }
     givens;  { low level procedures for GIVENS plane rotations               }

CONST macheps=2.220446049e-16;
TYPE real=double;
     ConstrType=RECORD
                  lower_bound : Boolean;
                  upper_bound : Boolean;
                  bound       : Boolean;
                END;

{-----------------------------------------------------------------------------}

PROCEDURE BLS(
              m,n                : integer;{ dimension of problem             }
          VAR Ap;                          { pointer to problem matrix A      }
              dA                 : integer;{ leading dimension of A           }
          VAR bp;                          { pointer to right hand side b     }
          VAR Active_BoundsP;              { pointer to bounds                }
          VAR lp, up;                      { pointer to lower and upper bounds}
              try_last_active_set: Boolean;
          VAR xp;                          { pointer to solution vector x     }
          VAR Residual_sum       : real;   { || Ax - b ||_2                   }
          VAR workrealp;                   { real working array               }
          VAR workintp                     { integer working array            }
             );
{ Computation of x   : min || Ax-b ||  , l<=x<=u
                  min                2

  The problem is solved with an active set strategy. The problem matrix
  is assumed to be of full rank n. The algorithm is taken from

      Bjoerck: Least squares methods,
  in: Handbook of numerical analysis,
      Vol.1: Solution of equations in R^n ,
      Ed.: P.G.Ciarlet and J.L.Lions
      Elsevier / North Holland 1987

  Algorithm 27.1 Least squares problems with simple bounds;
                 Active set algorithm for problem BLS

  Note: If A has already upper triangular form then this structure will be
        preserved. Otherwise the matrix A and the right hand side b will be
        overwritten with the result of an orthogonalization method.

        If try_last_active_set=true then the active set resulting from the
        last call to function BLS will be used first. Then workint[n+1]..
        workint[2*n] (indizes of bound and of free variables) must have the
        old values.

  m    : Integer, number of rows of A

  n    : Integer, number of columns of A

  A    : Array[1..m,1..n] of real; A is assumed to be of full rank n
  Ap   : Pointer to the elements of A

  dA   : Integer, leading dimension of A
         if A was declared as Array[1..MMAX,1..NMAX] of real then dA:=NMAX

  b    : Array[1..m] of real; right hand side
  bp   : Pointer to the elements of b

  Active_Bounds : Array[1..n] of ConstrType
    bound                 =TRUE   if there exists a constraint on x[i]
                           FALSE  else
    upper_bound           =TRUE   if there exists an upper bound on x[i]
                           FALSE  else
    lower_bound           =TRUE   if there exists a lower bound on x[i]
                           FALSE  else
  Active_BoundsP : Pointer to the elements of Active_Bounds

  l, u : Array[1..n] of real; contain number of bounds (if any)
  lp,up: Pointer to the elements of l, u

  try_last_active_set : BOOLEAN; in consecutive calls this can speed up to
                                 find the correct active set
  x    : Array[1..n] of real; solution vector
  xp   : Pointer to the elements of x

  Residual_sum : real;           || Ax-b||_2

  workreal : Array[1..3*n] of real; real working array
  workrealp: Pointer to the elements of workreal

  workint  : Array[1..2*n] of integer; integer working array
  workintp : Pointer to the elements of workint

  Subroutines called

  On Output:

  A,b           results of the QR-transformation
  x             solution of the box constrained problem
  Residual_Sum  || A*x-b ||_2

  Active_Bounds[i].Bound   =TRUE     if i-th bound is active
                            FALSE    if i-th bound is inactive

  if Active_Bounds[i].Bound = TRUE then
    workreal[i] contains the Lagrange-Parameter to the i-th constraint

    workreal[i] > 0    if upper bound is active
    workreal[i] < 0    if lower bound is active
}

{-----------------------------------------------------------------------------}

IMPLEMENTATION
{$R-}

CONST Factor = 0.01;

{-----------------------------------------------------------------------------}

FUNCTION ind (LDA,i,j : integer) : integer;
{ Function for index computations
  Element a[i,j] corresponds to  vector [ (i-1) * LDA+j ]
  ( Arrays are stored row by row in Pascal )
  A must be declared as Array[1..m,1..LDA]                                    }

BEGIN
  ind:=(i-1)*LDA+j;
END; { ind }

{-----------------------------------------------------------------------------}

FUNCTION LTE( a,b: real): Boolean;
VAR b_plus_b_a,b_a: real;
BEGIN
  b_a := (b - a)*Factor;
  b_plus_b_a := b + b_a;     { Round to precision of machine }
  LTE := b_plus_b_a  >= b
END;

{-----------------------------------------------------------------------------}

FUNCTION GTE( a, b: real): Boolean;
VAR b_plus_b_a,b_a: real;
BEGIN
  b_a := (b - a)*Factor;
  b_plus_b_a := b + b_a;     { Round to precision of machine }
  GTE := b_plus_b_a  <= b;
END;

{-----------------------------------------------------------------------------}

PROCEDURE BLS(
              m,n                : integer;{ dimension of problem             }
          VAR Ap;                          { pointer to problem matrix A      }
              dA                 : integer;{ leading dimension of A           }
          VAR bp;                          { pointer to right hand side b     }
          VAR Active_BoundsP;              { pointer to bounds                }
          VAR lp, up;                      { pointer to lower and upper bounds}
              try_last_active_set: Boolean;
          VAR xp;                          { pointer to solution vector x     }
          VAR Residual_sum       : real;   { || Ax - b ||_2                   }
          VAR workrealp;                   { real working array               }
          VAR workintp                     { integer working array            }
             );

LABEL finished;

TYPE vector       = ARRAY[1..1] OF real;
     intvector    = ARRAY[1..1] OF integer;
     constr_array = ARRAY[1..1] OF ConstrType;

CONST nk       : integer=0;        { number of active constraints }
      firstcall: boolean= true;

VAR A                  : vector       ABSOLUTE Ap;
    b                  : vector       ABSOLUTE bp;
    Active_Bounds      : constr_array ABSOLUTE Active_BoundsP;
    l                  : vector       ABSOLUTE lp;
    u                  : vector       ABSOLUTE up;
    x                  : vector       ABSOLUTE xp;
    workreal           : vector       ABSOLUTE workrealp;
    workint            : intvector    ABSOLUTE workintp;

    { Work arrays }
    lambda_p,                    { lambda_p points to first element of lambda
                                   lambda=Lagrange multipliers                }
    q_p,                         { q_p points to first element of q, q[i]=z[i]}
    c_p,                         { c_p points to first element of c           }
    gamma_p,                     { gamma_p points to first element of gamma   }
    pp_p,                        { pp_p points to first element of pp         }

    Iterations,                  { number of iterations                       }
    MaxIterations,               { maximal number of iterations               }
    rankA,                       { (computed) rank of A                       }
    i,                           { Index for vector x, i=pp[ip]               }
    ip,                          { permutated index
                                         ip=1..n-nk: free variables
                                         ip=n-nk..n: bound variables          }
    j,jp,k,
    kp,
    imin,
    t,
    tp
                       : integer;
    xi,
    zi,
    maxl,
    lam,
    alphamin,
    x_q                : real;


    zfeasible,
    optimal,
    atbound,
    Bound_just_freed_x : Boolean;

{  workreal [ 1   ] .. workreal [  n  ]    lambda   = Array[1..n] of real
                                           lambda_p:=1;

   workreal [n+1  ] .. workreal [ 2*n ]    q        = Array[1..n] of real;
                                           q_p     :=n+1;

   workreal [2*n+1] .. workreal [ 3*n ]    c        = Array[1..n] of real;
                                           c_p     :=2*n+1;

   workint  [1    ] .. workint  [ n   ]    gamma    = Array[1..n] of integer;
                                           gamma_p :=1;

   workint  [ n+1 ] .. workint  [ 2*n ]    pp       = Array[1..n] of integer;
                                           pp_p    :=n+1;

   ==> workreal[q_p-1+i] is the i-th element of q                             }


PROCEDURE QR_Fact;
{ QR factorization of A by Givens rotations (row by row)                      }
VAR i,j,jp: integer;
BEGIN
  FOR i := 2 TO m DO
    FOR jp := 1 TO min(i-1,n) DO
      BEGIN
        j := workint[pp_p-1+jp]; { get index of row }
        IF A[ ind(dA,i,j) ] <> 0.0 THEN
          BEGIN { Annihilate A[i,j] by GIVENS-rotations : }

            setup_GIVENS_rotation( A[ ind(dA,jp,j) ],A[ ind(dA,i,j) ],xi);
            { to annihilate A[i,j] by A[jp,j]}

            apply_GIVENS_rotation( n, A[ ind(dA,jp,1) ], A[ ind(dA,i,1) ] );
            { all columns }
            A[ ind(dA,i,j) ] := 0.0;

            apply_GIVENS_rotation( 1, b[ jp], b[i]);
            { right hand side }
          END;
      END;
  m := n;
END;

BEGIN
  { set pointer to the first element of the relevant vectors }
  lambda_p:=1;
  q_p     :=n+1;
  c_p     :=2*n+1;
  gamma_p :=1;
  pp_p    :=n+1;


  IF firstcall THEN
    BEGIN
      try_last_active_set := false;{ There is no last active set }
      firstcall := false;          { prepare next call }
    END;

  { Initialization }
  IF NOT try_last_active_set THEN
    BEGIN
      nk := 0; { no active constraints }
      FOR ip := 1 TO n DO
        workint[pp_p-1+ip] := ip;
    END;
  { i in F if x[i] is a free variable
    i in B if x[i] is fixed at his lower or upper bound

    F := [1..n]; index set of free variables
    B := [];     index set of bound variables }

  MaxIterations := 10 * n;
  t := 0;
  iterations := 0;

  { Check compatibility of constraints}
  FOR i := 1 TO n DO
    WITH Active_Bounds[i] DO
      IF lower_bound AND upper_bound THEN
        IF l[i] > u[i] THEN
          Error_Handler('Error in BLS: lower > upper');

  { set initial point in interior of feasible set
    x:=(l+u)/2                                    }
  FOR i := 1 TO n DO
    WITH Active_Bounds[i] DO
      IF lower_bound THEN               { lower bound exists }

        IF upper_bound THEN
          x[i] := ( l[i] + u[i]) * 0.5  { both upper and lower bound exist }
        ELSE                            { only lower bound exists }

          IF l[i] <= 0.0 THEN
            x[i] := l[i] * 0.5
          ELSE
            x[i] := l[i] * 1.5

      ELSE                              { no lower bound exists }
        IF upper_bound THEN             { only upper bound exists }
          IF u[i] >= 0.0 THEN
            x[i] := u[i] * 0.5
          ELSE
            x[i] := u[i] * 1.5
        ELSE    x[i] := 0.0;            { no bound at all }

  { Compute QR-Factorization of A,b, store it at A,b }
  RankA := 0;
  QR_Fact;
  { Now (A,b) is overwritten by the Factorization (R, S;c)
                                                  (0, U;d)

    E_F represents part of orthogonal matrix corresponding to free variables
    E_B represents part of orthogonal matrix corresponding to bound variables }

  { Compute numerical rank of A }
  FOR i := 1 TO n DO
    BEGIN
      IF Abs(A[ ind(dA,i,workint[pp_p-1+i]) ]) > macheps THEN Inc(rankA);
    END;

  IF RankA < n THEN
    Error_Handler('Error in BLS, RankA < n');

  { Main loop:}
  REPEAT

    { Compute unconstraint minimum in free variables
                    -1
            z := E R  c
                  F
    }
    IF ( t > 0 ) AND
      (( workreal[c_p-1+n-nk] + A[ ind(dA,n-nk,t) ]) = workreal[c_p-1+n-nk])
      THEN
      BEGIN { just freed x is linear dependent }
        FOR ip := 1 TO n-nk DO
          BEGIN
            i := workint[pp_p-1+ip];
            workreal[q_p-1+i] := x[i];
          END;
        Bound_just_freed_x := true;
      END
    ELSE
      BEGIN { just freed x is linear independent }
        Bound_just_freed_x := false;
        FOR ip := 1 TO n-nk DO
          workreal[c_p-1+ip] := b[ ip];
        zfeasible := true;
        ip := n - nk;
        WHILE ip >= 1 DO
          BEGIN
            i := workint[pp_p-1+ip];
            workreal[q_p-1+i] := workreal[c_p-1+ip] / A[ ind(dA,ip,i) ];
            zi := workreal[q_p-1+i];
            FOR j := ip-1 DOWNTO 1 DO
              workreal[c_p-1+j] := workreal[c_p-1+j] - A[ ind(dA,j,i) ] *zi;
            WITH Active_Bounds[i] DO
              BEGIN
                IF lower_bound THEN
                  IF NOT GTE( workreal[q_p-1+i], l[i]) THEN
                    zfeasible := false;
                IF upper_Bound THEN
                  IF NOT LTE( workreal[q_p-1+i], u[i]) THEN
                    zfeasible := false;
              END;
            Dec(IP);
          END;

        { c:=E_F R^(-1) c }


        { if just freed component is not feasible }
        IF NOT zfeasible THEN
          IF t > 0 THEN
            IF (( workreal[q_p-1+t] < l[t] ) AND ( workint[gamma_p-1+t] = 1))
               OR
               (( workreal[q_p-1+t] > u[t])  AND ( workint[gamma_p-1+t] = -1))
            THEN
              BEGIN
                FOR ip := 1 TO n-nk DO
                  workreal[q_p-1+(workint[pp_p-1+ip])] := x[ip]; { no change in x }
                Bound_just_freed_x := true { component will again be bound }
              END;
    END; { t>0 and linear independent }


    { zi = q[i]] }
    IF zfeasible THEN
      BEGIN
        { Set x[i] := z[i] for i in F }
        FOR ip := 1 TO n-nk DO
          BEGIN
            i := workint[pp_p-1+ip];
            x[ i] := workreal[q_p-1+i]
          END{ for};

        { Check for optimality }

        { Compute LAGRANGE-multipliers lambda=E_B transp(U) d }
        IF nk = 0 THEN
          GOTO finished; { B=[], all variables are free }

        maxl := 0.0;
        optimal := true; { assume that we have found the solution }
        FOR ip := n-nk+1 TO n DO
          BEGIN { for all bound variables}
            i := workint[pp_p-1+ip];
            workreal[lambda_p-1+i] := 0.0; { lambda[i] corresponds to z[i]}
            FOR k := n-nk+1 TO n DO
              workreal[lambda_p-1+i] := workreal[lambda_p-1+i] +
                                        A[ ind(dA,k,i) ]*b[k];
            { Lagrange multipliers for bound variables }

            { Find index t with
                sign(gamma_t)*lambda_t = max_( i in B) sign(gamma_i) lambda_i }
            lam := sign( workint[gamma_p-1+i])*workreal[lambda_p-1+i];
            IF NOT LTE(lam, maxl) THEN
              BEGIN
                optimal := false; { x is not optimal }
                maxl := lam;
                t := i;           { move index t from B to F }
                tp:= ip;
              END
          END;

        IF optimal THEN
          GOTO finished; { solution found }

        { Free x[t]:
             B := B-t; F := F+t; }
        FOR ip := tp DOWNTO n-nk+2 DO
          workint[pp_p-1+ip] := workint[pp_p-1+ip-1];
        workint[pp_p-1+n-nk+1] := t;
        nk := nk -1;

        { Update Decomposition: }
        FOR ip := tp DOWNTO n-nk+1 DO
          IF A[ ind(dA,ip,t) ] <> 0.0 THEN
            BEGIN
              { Annihilate A[ip,t] by GIVENS-rotations : }
              setup_GIVENS_rotation( A[ ind(dA,ip-1,t) ],A[ ind(dA,ip,t)],xi);
              { to annihilate A[ip,t] by A[ip-1,t] }
              apply_GIVENS_rotation( n, A[ ind(dA,ip-1,1) ],A[ ind(dA,ip,1) ]);
              { all columns }
              A[ ind(dA,ip,t) ] := 0.0;

              apply_GIVENS_rotation( 1, b[ ip-1], b[ip]);
              { right hand side }
            END;
        { Factorization updated, again triangle }

        { Update right hand side
               ( c )
               ( d )  = b - A(E ,E ) * E x
                               F  B     B
          (c)
          (d) equals the residual b - Ax if the free components
              of x are set zero, the bound components equal there bounds
              (in general <> 0.0 !!)

           UPDATE (c)   (c)   (S)                                         T
                  (d):= (d) + (U) * E  x , with x = (0,0,..,x      ,0,..,0)
                                     B                       tBound
        }
        FOR j := 1 TO n DO
          b[j] := b[j] + A[ ind(dA,j,t) ] * x[t];

      END { z feasible }
    ELSE
      BEGIN { z not feasible }

        IF NOT Bound_just_freed_x THEN
          BEGIN
            alphamin := 1.0;
            imin := 0;
            FOR ip := 1 TO n-nk DO
              BEGIN { for all i in F }
                { compute alpha(i)= (x(i)-l(i)) / (x(i)-z(i))  if z(i) < l(i)
                                    (u(i)-x(i)) / (z(i)-x(i))  if z(i) > l(i) }
                i := workint[pp_p-1+ip];
                x_q := x[i]-workreal[q_p-1+i];
                WITH Active_Bounds[i] DO
                  BEGIN
                    IF x_q <> 0.0 THEN
                      BEGIN
                        IF lower_bound THEN
                          IF NOT GTE(workreal[q_p-1+i],l[i]) THEN
                            BEGIN
                              alphamin := rmin( alphamin, (x[i]-l[i])/x_q);
                              imin := i;
                            END;
                        IF upper_bound THEN
                          IF NOT LTE(workreal[q_p-1+i],u[i]) THEN
                            BEGIN
                              alphamin := rmin( alphamin, (u[i]-x[i])/(-x_q));
                              imin := i;
                            END;
                      END
                    ELSE { q = x }
                      alphamin := 0.0;
                  END; {with}
              END;
          END; { if NOT Bound_just_Freed_x }

        ip := 1;
        WHILE ip <= n-nk DO
          BEGIN { for all i in F }
            i := workint[pp_p-1+ip];
            {Set new x:}
            IF Bound_just_freed_x THEN
              IF i = t THEN
                 BEGIN
                   workint[gamma_p-1+i] := 0;
                   atbound := true
                 END
               ELSE
                 atbound := false
            ELSE
              BEGIN
                IF i = imin THEN
                  IF workreal[q_p-1+i] > x[i]  THEN
                    x[i] := u[i]
                  ELSE
                    x[i] := l[i]
                ELSE
                  x[ i] := x[ i] + alphamin * ( workreal[q_p-1+i]-x[i] );
                atbound := false;
                WITH Active_Bounds[i] DO
                  BEGIN
                    workint[gamma_p-1+i] := 0;
                    IF lower_bound THEN
                      IF LTE(x[i],l[i]) THEN
                        BEGIN
                          x[i] := l[i];
                          workint[gamma_p-1+i] := 1;
                          atbound := true
                        END;
                    IF upper_bound THEN
                      IF GTE(x[i],u[i]) THEN
                        BEGIN
                          x[i] := u[i];
                          IF workint[gamma_p-1+i] = 1 THEN
                            { lower bound=upper bound }
                            workint[gamma_p-1+i] := 0
                          ELSE
                            workint[gamma_p-1+i] := -1;
                          atbound := true; { upper bound}
                        END;
                  END; { with }

              END; { if Bound_just_freed_x }

            IF atbound THEN
              BEGIN
                { F := F-i;B:=B+i; Bound x[i] }
                FOR jp := ip TO n-nk-1 DO
                  workint[pp_p-1+jp] := workint[pp_p-1+jp+1];
                workint[pp_p-1+n-nk] := i;
                nk := nk +1;
                { Set of free and bounded variables updated }

                { Update Factorization: }
                FOR kp := ip TO n-nk DO
                  BEGIN
                    k := workint[pp_p-1+kp];
                    IF A[ ind(dA,kp+1,k) ] <> 0.0 THEN
                      BEGIN
                        { Annihilate A[kp+1,k] by GIVENS-rotations : }
                        setup_GIVENS_rotation( A[ ind(dA,kp,k) ],A[ ind(dA,kp+1,k) ],xi);
                        { to annihilate A[kp+1,k] by A[kp,k] }
                        apply_GIVENS_rotation( n, A[ind(dA,kp,1) ], A[ ind(dA,kp+1,1) ]);
                        { all columns }

                        A[ ind(dA,kp+1,k) ] := 0.0;
                        apply_GIVENS_rotation( 1, b[ kp], b[kp+1]);
                        { right hand side }
                      END{ if}
                  END; { for }

                { UPDATE right hand side: }
                FOR j := 1 TO n DO
                  b[j] := b[j] - A[ ind(dA,j,i) ] * x[i];

              END
            ELSE
              Inc(ip);

          END; { while ip }

        t := 0;  { we come from not feasible }
        Bound_just_freed_x := false;
      END; { z not feasible }

    Inc(Iterations);
    IF Iterations > MaxIterations THEN
      GOTO finished;

  UNTIL false; { Main loop }

finished:
  FOR ip := 1 TO n - nk DO
    Active_Bounds[workint[pp_p-1+ip]].bound := false;

  { compute residual and set active constraints (for next a call) }
  Residual_Sum := 0.0;
  FOR ip := n-nk+1 TO n DO
    BEGIN
      i := workint[pp_p-1+ip];
      WITH Active_Bounds[i] DO
        bound := true;
      Residual_Sum := Residual_Sum + Sqr( b[ ip]);
    END; { for }
  Residual_Sum:=Sqrt(Residual_Sum);

END;

{-----------------------------------------------------------------------------}

END.