UNIT BSpline;
{ procedures for the computation of polynomial B-Splines and l2-approximations}

INTERFACE
  USES def,       { global declarations                                       }
       error,     { error handler                                             }
       math,      { low level mathematical subroutines                        }
       GIVENS;    { low level procedures for GIVENS plane rotations           }

{-----------------------------------------------------------------------------}

PROCEDURE interval(xx       : real;              { argument to be placed      }
                   nr       : integer;           { index of right endpoint    }
                   VAR t    : knotsequence;
                   VAR left : integer;           { t[left] <= xx <t[left+1]   }
                   VAR out_of_interval : Boolean
                  );
{ procedure to place the argument xx within the breakpoint sequence t         }

{-----------------------------------------------------------------------------}

PROCEDURE BSpline_table( VAR row     : RowVector;
                         VAR t       : knotsequence;
                                  xx : real;
                             k,
                                left : integer
                       );
{ computes all nonzero B-Splines of order k to knot sequence t at point xx
  t[left] <= xx < t[left+1] is assumed

  The recurrence relation (i=left)

        B(i,j)(x)= (x-t[i])   / (t[i+j-1]-t[i])  B(i,j-1)(x) +
                   (t[i+j]-x) / (t[i+j]-t[i+1])  B(i+1,j-1)(x)
                 = w(i,j)(x) B(i,j-1)(x) + (1-w(i+1,j)(x)) B(i+1,j-1)(x)

        with               (x-t[i]) / (t[i+j-1]-t[i])     if t[i] < t[i+j-1]
              w(i,j)(x) =
                            0                             else

  for j=1,...,k is used to generate the vector
  row(1),...,row(k) = B(left-k+1,k),..., B(left,k).

  The algoritm generates simultaneously the last column of the
  following triangular array of B-Splines from left to right

                                 B (x)
                           ..     i-k+1,k
                   B (x)   ..     ..
                    i-1,2

            B (x)  B (x)   ..    B (x)
             i,1    i,2           i,k

Algorithm: C. de Boor, A practical guide to splines, p.134                    }

{-----------------------------------------------------------------------------}

FUNCTION Spline_Value ( VAR alpha  : coefficients;
                        VAR t      : knotsequence;
                            xx     : real;
                            left,k : integer) : real;

{ Evaluate Spline Approximation at xx
  t[left] <= xx < t[left+1] is assumed

  Spline_Value(xx)= B(xx) * alpha                                             }

{-----------------------------------------------------------------------------}

FUNCTION Spline_Value_Deriv (
                        VAR alpha  : coefficients;
                        VAR t      : knotsequence;
                            xx     : real;
                            r,k    : integer) : real;

{ calculates value at xx of r-th derivative of spline s
  s is a spline of order k represented by knotsequence t and coefficents
  alpha
  The r-th derivative of s exists if  t[j+k-r] > t[j], j=r+1,...,n
  t[left] <= xx < t[left+1] is assumed
  If all inner knots are simple then s^(k-1) exists and is piecewise constant }

{-----------------------------------------------------------------------------}

PROCEDURE Get_F (m,n,k     : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR y     : datavalues;
                 VAR alpha : coefficients;
                 VAR F     : datavalues
                );
{ computes F(t) = y - B(t) * alpha                                            }

{-----------------------------------------------------------------------------}

PROCEDURE GIVENS_factorization (  m,n   :integer;       { dimension }
                                    k   :integer;       { order     }
                                VAR t   :knotsequence;
                                VAR tau :dataabscissae;
                                VAR B   :leastsquarematrix;
                                VAR Q   :OrthogonalMatrix
                                );
{ produces GIVENS factorization of B row by row, stores parameter for
  plane rotation in Q, stores upper triangular matrix in B                    }

{-----------------------------------------------------------------------------}

PROCEDURE apply_GIVENS_transformation (  m,n  :integer;       { dimension }
                                          k   :integer;       { order     }
                                      VAR t   :knotsequence;
                                      VAR tau :dataabscissae;
                                      VAR y   :datavalues;
                                      VAR Q   :OrthogonalMatrix;
                                      VAR x   :coefficients;
                                      VAR RES :real
                                     );
{ applies GIVENS transformation of B stored in Q to the right hand side vector
  y. On Output x contains the first n components of Q^T * y, x=Q_1^T y        }

{-----------------------------------------------------------------------------}

PROCEDURE apply_GIVENS_transformation_all (
                                         m,n  :integer;       { dimension }
                                          k   :integer;       { order     }
                                      VAR t   :knotsequence;
                                      VAR tau :dataabscissae;
                                      VAR y   :datavalues;
                                      VAR Q   :OrthogonalMatrix
                                     );
{ applies GIVENS transformation of B stored in Q to the right hand side vector
  y. y will be overwritten by Q^T * y                                         }

{-----------------------------------------------------------------------------}

PROCEDURE apply_inverse_GIVENS_transformation (
                                         m,n  :integer;       { dimension }
                                          k   :integer;       { order     }
                                      VAR t   :knotsequence;
                                      VAR tau :dataabscissae;
                                      VAR v   :datavalues;
                                      VAR Q   :OrthogonalMatrix
                                     );
{ applies the inverse GIVENS factorization of B stored in Q to the right hand
  side vector v. v will be overwritten by Q * v                               }

{-----------------------------------------------------------------------------}

PROCEDURE Rc_x(    n,k:integer;            { number of rows, bandwidth }
               VAR R  :leastsquarematrix;  { upper triangular matrix   }
               VAR c  :coefficients;       { unknown coefficients      }
               VAR x  :coefficients        { right side                }
               );
{ solution of R c = x
   R banded upper triangular matrix with band width k
   R is stored row by row    r[i,i] in R[i,1]
                             r[i,j] in R[i,j-i+1]                             }

{-----------------------------------------------------------------------------}

PROCEDURE l2Appr(m,n,
                 k         : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR y     : datavalues;
                 VAR alpha : coefficients;
                 VAR RES   : real
                 );
{                                              +
  computes least squares solution alpha opt = B  *y
  of || y- B * alpha || --> Min ! by row by row GIVENS rotations              }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Spline_Knots_approx(
                 m,n,
                 k         : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR F     : datavalues
                 );
{
  computes derivative of spline s(x)=\sum_(j=1)^n B_(j,k)(x) * alpha(j) with
  respect to t[kk] for x=x_i i=1,...,m
  The derivative will be approximated by finite differences.

                            \partial s(x_i)
  F:= B'(t) [e^kk] alpha  = ---------------     i=1,...,m
                            \partial t(kk)
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Spline_Knots(
                 m,n,
                 k         : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR F     : datavalues
                 );
{
  computes derivative of spline s(x)=\sum_(j=1)^n B_(j,k)(x) * alpha(j) with
  respect to t[kk] for x=x_i i=1,...,m
  It is assumed that t[kk] is a simple inner knot. Otherwise the formula used
  is not valid.

                            \partial s(x_i)
  F:= B'(t) [e^kk] alpha  = ---------------     i=1,...,m
                            \partial t(kk)
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Get_BTvec(
                 m,n,k     : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR v     : datavalues;
                 VAR w     : coefficients
                );
{ computes w = B(t)^T * v                                                     }

{-----------------------------------------------------------------------------}

IMPLEMENTATION

{-----------------------------------------------------------------------------}

PROCEDURE interval(xx       : real;              { argument to be placed      }
                   nr       : integer;           { index of right endpoint    }
                   VAR t    : knotsequence;
                   VAR left : integer;           { t[left] <= xx <t[left+1]   }
                   VAR out_of_interval : Boolean
                  );
VAR i,j,k,n  : integer;

BEGIN
  {$R-}
  n:=nr-1;
  IF (xx < t[1]) OR (xx > t[nr]) THEN
    BEGIN
      IF xx < t[1] THEN
        left:=1
      ELSE
        left:=nr;
      out_of_interval:=TRUE;
      Exit;
    END
  ELSE
    out_of_interval:=FALSE;

  IF left > 0 THEN
    IF left < nr THEN
      BEGIN
        k:=left;
        IF t[k] <= xx THEN
          IF xx < t[k+1] THEN
            BEGIN
              left:=k;
              Exit;
            END
          ELSE
            BEGIN
              IF k < n THEN
                REPEAT Inc(k)
                UNTIL (t[k]<>t[k+1]) OR (k=n);  { knot multiplicity <> 1 }
              IF xx < t[k+1] THEN
                BEGIN
                  left:=k;
                  Exit;
                END;
            END;
      END;
  IF xx = t[nr] THEN
    BEGIN
      left:=n;
      Exit;
    END;

  { binary search }
  i:=1;k:=1;j:=nr;
  WHILE (k >= 1) AND (k <= n) AND (i <> j) DO
    BEGIN
      k:=(i+j) DIV 2;
      IF t[k] = xx THEN
        BEGIN
          WHILE (t[k]=t[k+1]) AND (k<n) DO
            Inc(k);
          left:=k;
          Exit;
        END;
      IF t[k] <= xx THEN
        IF xx < t[k+1] THEN
          BEGIN
            left:=k;
            Exit;
          END
        ELSE i:=k
        ELSE
          j:=k;
    END;
END; { interval }

{-----------------------------------------------------------------------------}

PROCEDURE BSpline_table( VAR row     : RowVector;
                         VAR t       : knotsequence;
                                  xx : real;
                             k,
                                left : integer
                       );
VAR j, r   : integer;
    saved, tr, tl, term : real;

BEGIN
  row[1]:=1.0;

  FOR j:= 1 TO k-1 DO
    BEGIN
      saved:=0.0;
      FOR r:= 1 TO j DO
        BEGIN
          tr:= t[left+r] - xx;
          tl:= xx - t[left+r-j];
          term:= row[r]/(tr+tl);
          row[r]:=saved+tr*term;
          saved:=tl*term;
        END;
      row[j+1]:=saved;
    END;
END; { BSpline_Table }

{-----------------------------------------------------------------------------}

FUNCTION Spline_Value ( VAR alpha  : coefficients;
                        VAR t      : knotsequence;
                            xx     : real;
                            left,k : integer) : real;

VAR  j : integer;
   sum : real;
     u : RowVector;

BEGIN
  FOR j:=1 TO k DO
    u[j]:=0.0;

  BSpline_Table(u,t,xx,k,left);
  sum:=0.0;
  FOR j:=left-k+1 TO left DO
    sum:=sum + alpha[j] * u[j-left+k];
  Spline_Value:= sum;
END; { Spline_Value }

{-----------------------------------------------------------------------------}

FUNCTION Spline_Value_Deriv (
                        VAR alpha  : coefficients;
                        VAR t      : knotsequence;
                            xx     : real;
                            r,k    : integer) : real;

LABEL return;

VAR DL,DR : ARRAY[1..NBANDS-1] OF real;
    AJ    : ARRAY[1..NBANDS] OF real;
    j, jj, ilo, JCMAX, JCMIN : integer;

BEGIN
  IF r >= k THEN
    BEGIN
      Spline_Value_deriv:=0.0;
      GOTO return;
    END;

  { locate interval with t[left] <= xx < t[left+1] }
  interval(xx,n+1,t,left,out_of_interval);
  IF Out_of_interval THEN
    BEGIN { xx lies outside the support of the spline s and s^r(xx)=0.0 }
      Spline_value_deriv:=0.0;
      GOTO return;
    END;

  IF k=1 THEN
    BEGIN
      Spline_value_deriv:=alpha[left];
      GOTO return;
    END;

  { store the k B-spline coefficients relevant for the knot interval
    t[left],t[left+1] in AJ[1], ... , AJ[k] and compute
    DL[j]=xx-t[left+1-j] and DR[j]=t[left+j]-xx for j=1,..,k-1
    Set any of the AJ not obtainable from input to zero
    Set any of the t's not obtainable equal to t[1] or [t[n+k] }

  JCMIN:=1;
  IF left<k THEN
    BEGIN
      JCMIN:=1+k-left;
      FOR j:=1 TO left DO
        dl[j]:=xx-t[left+1-j];
      FOR j:=left TO k-1 DO
        BEGIN
          aj[k-j]:=0.0;
          dl[j]:=dl[left];
        END;
    END
  ELSE
    BEGIN
      FOR j:=1 TO k-1 DO
        dl[j]:=xx-t[left+1-j];
    END;

  JCMAX:=k;
  IF n < left THEN
    BEGIN
      JCMAX:=k+n-left;
      FOR j:=1 TO JCMAX DO
        dr[j]:=t[left+j]-xx;
      FOR j:=jcmax TO k-1 DO
        BEGIN
          aj[j+1]:=0.0;
          dr[j]:=dr[jcmax];
        END;
    END
  ELSE
    BEGIN
      FOR j:=1 TO k-1 DO
        dr[j]:=t[left+j]-xx;
    END;

  FOR j:=jcmin TO jcmax DO
    aj[j]:=alpha[left-k+j];

  { difference the coefficients r times }
  IF r <> 0 THEN
    BEGIN
      FOR j:=1 TO r DO
        BEGIN
          ilo:=k-j;
          FOR jj:=1 TO k-j DO
            BEGIN
              aj[jj]:=((aj[jj+1]-aj[jj])/(dl[ilo]+dr[jj]))*(k-j);
              Dec(ilo);
            END;
        END;
    END;

  { compute value at x in t[left],t[left+1] of r-th derivative, given its
    relevant B-spline coefficients in aj[1], ... , aj[k-r] }
  IF r=k-1 THEN
    Spline_value_deriv:=aj[1]
  ELSE
    BEGIN
      FOR j:=r+1 TO k-1 DO
        BEGIN
          ilo:=k-j;
          FOR jj:=1 TO k-j DO
            BEGIN
              aj[jj]:=(aj[jj+1]*dl[ilo]+aj[jj]*dr[jj])/(dl[ilo]+dr[jj]);
              Dec(ilo);
            END;
        END;
      Spline_Value_deriv:=aj[1];
    END;

return:
END; {Spline_Value_deriv }

{-----------------------------------------------------------------------------}

PROCEDURE Get_F (m,n,k     : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR y     : datavalues;
                 VAR alpha : coefficients;
                 VAR F     : datavalues
                );
VAR i : integer;
BEGIN
  FOR i:= 1 TO m DO
    BEGIN
      interval(tau[i],n+1,t,left,out_of_interval);
      F[i]:= y[i] - Spline_Value(alpha,t,tau[i],left,k)
    END;
END; { Get_F }

{-----------------------------------------------------------------------------}

PROCEDURE GIVENS_factorization (  m,n   :integer;       { dimension }
                                    k   :integer;       { order     }
                                VAR t   :knotsequence;
                                VAR tau :dataabscissae;
                                VAR B   :leastsquarematrix;
                                VAR Q   :OrthogonalMatrix
                                );

VAR i,j,ind             :integer;
    u                   :RowVector;
    xi                  :real; { to reconstruct the parameters}

BEGIN
  { initialization }
  FOR i:=1 TO n DO
    FOR j:=1 TO k DO
      B[i,j]:=0.0;

  FOR i:=1 TO m DO
    FOR j:=1 TO k DO
      Q[i,j]:=0.0;

  { computation }
  FOR i:=1 TO m DO
    BEGIN
      { compute i-th row of B and store it to u }
      interval(tau[i],n+1,t,left,out_of_interval);
      IF out_of_interval THEN
        Error_Handler('Givens Factorization: Point out of interval');
      Bspline_table(u,t,tau[i],k,left);
      { compute values of all nonzero B-Splines to tau[i] and store it to u }

      { rotate out the i-th row of B }
      FOR j:= 1 TO min(k,n-left+k) DO
        BEGIN
          { annihilate u[j] }
          IF u[j] <> 0.0 THEN
            BEGIN
              ind:=left+j-k;
              setup_GIVENS_rotation(B[ind,1],u[j],xi);

              apply_GIVENS_rotation(k-j+1,B[ind,1],u[j]);

              { now save xi for reconstructing the parameters }
              Q[i,j]:=xi;
            END
        END;
    END;

  { The first k rows of the B-Spline matrix will just be inserted into the
    upper triangular matrix. We do this with a rotation (CosG=0, SinG=1,
    Xi=1). This saves us from a special handling of the first k rows.
    It does not affect the computation of c=Q_1^T * y using the
    procedure apply_GIVENS_transformation, i.e. we rotate not in place of c.
    However, computing products of the form Q * y and Q^T * y (the whole
    vector, not just the first elements) we must set this rotations to
    identity transformations, i.e. Xi=0, before using the transformation.

    FOR j:=1 TO k do
      Q[j,j]:=0.0;

    A routine performing the computation of (c,d)= Q^T * y (the whole
    m-vector, not just the first n components in c) can be found in
    apply_GIVENS_transformation_all.
    A routine performing the computation of Q * y can be found in
    apply_inverse_GIVENS_transformation.
  }
END; { GIVENS_factorization }

{-----------------------------------------------------------------------------}

PROCEDURE apply_GIVENS_transformation (  m,n  :integer;       { dimension }
                                          k   :integer;       { order     }
                                      VAR t   :knotsequence;
                                      VAR tau :dataabscissae;
                                      VAR y   :datavalues;
                                      VAR Q   :OrthogonalMatrix;
                                      VAR x   :coefficients;
                                      VAR RES :real
                                     );

VAR i,j,ind             : integer;
    v                   : real;

BEGIN
  { Initialization }
  Res:=0.0;
  FOR i:=1 TO n DO
    x[i]:=0.0;

  FOR i:=1 TO m DO
    BEGIN
      v:=y[i];
      interval(tau[i],n+1,t,left,out_of_interval);
      IF out_of_interval THEN
        Error_Handler('Givens factorization: Point out of interval');

      FOR j:=1 TO min(k,n-left+k) DO
        BEGIN
          ind:=left+j-k;
          reconstruct_GIVENS_rotation(Q[i,j]);
          apply_GIVENS_rotation(1,x[ind],v);
        END;
      RES:=RES+Sqr(v);
    END;
  RES:=Sqrt(RES);

END; { apply_GIVENS_transformation }

{-----------------------------------------------------------------------------}

PROCEDURE apply_GIVENS_transformation_all (
                                         m,n  :integer;       { dimension }
                                          k   :integer;       { order     }
                                      VAR t   :knotsequence;
                                      VAR tau :dataabscissae;
                                      VAR y   :datavalues;
                                      VAR Q   :OrthogonalMatrix
                                     );

VAR i,j,ind             : integer;

BEGIN
  { modify Q, for reasons see GIVENS_factorization }
  FOR i:=1 TO k DO
    Q[i,i]:=0.0;

  FOR i:=1 TO m DO
    BEGIN
      interval(tau[i],n+1,t,left,out_of_interval);
      IF out_of_interval THEN
        Error_Handler('Givens factorization: Point out of interval');

      FOR j:=1 TO min(k,n-left+k) DO
        BEGIN
          ind:=left+j-k;
          reconstruct_GIVENS_rotation(Q[i,j]);
          apply_GIVENS_rotation(1,y[ind],y[i]);
        END;
    END;

  { restore original Q }
  FOR i:=1 TO k DO
    Q[i,i]:=1.0;

END; { apply_GIVENS_transformation_all }

{-----------------------------------------------------------------------------}

PROCEDURE apply_inverse_GIVENS_transformation (
                                         m,n  :integer;       { dimension }
                                          k   :integer;       { order     }
                                      VAR t   :knotsequence;
                                      VAR tau :dataabscissae;
                                      VAR v   :datavalues;
                                      VAR Q   :OrthogonalMatrix
                                     );

VAR i,j,ind             : integer;

BEGIN
  { modify Q, for reasons see GIVENS_factorization }
  FOR i:=1 TO k DO
    Q[i,i]:=0.0;

  FOR i:=m DOWNTO 1 DO
    BEGIN
      interval(tau[i],n+1,t,left,out_of_interval);
      IF out_of_interval THEN
        Error_Handler('Givens factorization: Point out of interval');

      FOR j:=min(k,n-left+k) DOWNTO 1 DO
        BEGIN
          ind:=left+j-k;
          reconstruct_GIVENS_rotation(Q[i,j]);
          SinG:=-SinG;
          apply_GIVENS_rotation(1,v[ind],v[i]);
        END;
    END;

  { restore original Q }
  FOR i:=1 TO k DO
    Q[i,i]:=1.0;

END; { apply_inverse_GIVENS_transformation }

{-----------------------------------------------------------------------------}

PROCEDURE Rc_x(    n,k:integer;            { number of rows, bandwidth }
               VAR R  :leastsquarematrix;  { upper triangular matrix   }
               VAR c  :coefficients;       { unknown coefficients      }
               VAR x  :coefficients        { right side                }
               );

VAR i,j                 :integer;
    s                   :real;

BEGIN
  FOR i:=n DOWNTO 1 DO
    BEGIN
      IF Abs(R[i,1]) < MachEps THEN
        Error_Handler('Error in Rc_x: Matrix singular')
      ELSE
        BEGIN
          s:=x[i];
          FOR j:=min(n,i+k-1) DOWNTO i+1 DO
            s:=s-R[i,j-i+1]*c[j];
          c[i]:=s/R[i,1];
        END;
    END;
END;  { Rc_x }

{-----------------------------------------------------------------------------}

PROCEDURE l2Appr(m,n,
                 k         : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR y     : datavalues;
                 VAR alpha : coefficients;
                 VAR RES   : real
                 );
VAR  B : leastsquarematrix;
     Q : OrthogonalMatrix;
     x : coefficients;

BEGIN
  GIVENS_factorization(m,n,k,t,tau,B,Q);
  apply_GIVENS_transformation(m,n,k,t,tau,y,Q,x,RES);
  Rc_x(n,k,B,alpha,x);
END; { l2Appr }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Spline_Knots_approx(
                 m,n,
                 k         : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR F     : datavalues
                 );

VAR h, temp              : real;
    i, j, j1, j2, j3, j4 : integer;
    ttemp                : knotsequence;
    u, utemp             : RowVector;
    left, lefttemp       : integer;

BEGIN
  h:=EPSDiff1*(Abs(t[kk]) + EPSDiff2);
  temp:=t[per[kk]]+h;
  h:=temp-t[per[kk]];
  { trick to improve the accuracy of finite difference approximations }

  FOR i:=1 TO n+k DO
    ttemp[i]:=t[i];

  ttemp[ kk ]:= t[ kk ]+h;

  FOR i:=1 TO m DO
    BEGIN
      F[i]:=0.0;
      interval(tau[i],n+1,ttemp,lefttemp,out_of_interval);
      interval(tau[i],n+1,t,left,out_of_interval);
      { lefttemp <= left because of ttemp[kk] > t[kk] }
      j1:=max(lefttemp-k+1,1);
      j2:=max(left-k+1,1);
      j3:=min(lefttemp,n);
      j4:=min(left,n);

      { nonzero B-Splines to tau[i] }
      BSpline_Table(u,t,tau[i],k,left);
      BSpline_Table(utemp,ttemp,tau[i],k,lefttemp);

      FOR j:=j1 TO j2 - 1 DO
        BEGIN      { Summmation about all nonzero B-Splines to tau[i] }
          F[i]:=F[i] + utemp[j-lefttemp+k] / h * alpha[j];
          { F = ( B' [e^k] ) alpha }
        END;

      FOR j:=j2 TO j3 DO
        BEGIN      { Summmation about all nonzero B-Splines to tau[i] }
          F[i]:=F[i] + (utemp[j-lefttemp+k] - u[j-left+k] ) / h * alpha[j];
          { F = ( B' [e^k] ) alpha }
        END;

      FOR j:=j3 + 1 TO j4 DO
        BEGIN      { Summmation about all nonzero B-Splines to tau[i] }
          F[i]:=F[i] + ( - u[j-left+k] ) / h * alpha[j];
          { F = ( B' [e^k] ) alpha }
        END;
    END;

END; { Derivative_Spline_Knots_approx }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Spline_Knots(
                 m,n,
                 k         : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR F     : datavalues
                 );

VAR i,j         : integer;
    ttemp       : knotsequence;
    u           : RowVector;

BEGIN
  { set temporary knot sequence t' }
  FOR j:=1 TO kk DO
    ttemp[j]:=t[j];
  FOR j:=kk+1 to n+k+1 do
    ttemp[j]:=t[j-1];

  FOR i:=1 TO m DO
    BEGIN
      IF (tau[i] <= t[kk-k+1]) OR (tau[i] >= t[kk+k-1]) THEN
        F[i]:=0.0
      ELSE
        BEGIN
          F[i]:=0.0;
          interval(tau[i],n+2,ttemp,left,out_of_interval);

          { nonzero B-Splines to tau[i] }
          BSpline_Table(u,ttemp,tau[i],k,left);
          FOR j:=max(left-k+1,kk-k+1) TO min(left,kk) DO
            BEGIN      { Summmation about all nonzero B-Splines to tau[i] }
              F[i]:=F[i]+ u[j-left+k] * (alpha[j-1]-alpha[j])/(ttemp[j+k]-ttemp[j]);
              { F = ( B' [e^k] ) alpha }
            END;
        END;
    END;
END; { Derivative_Spline_Knots }

{-----------------------------------------------------------------------------}

PROCEDURE Get_BTvec(
                 m,n,k     : integer;
                 VAR t     : knotsequence;
                 VAR tau   : dataabscissae;
                 VAR v     : datavalues;
                 VAR w     : coefficients
                );
var i,j,ind,left : integer;
    u            :RowVector;
Begin
  for j:=1 to n do
    w[j]:=0.0;

  FOR i:=1 TO m DO
    BEGIN
      { compute i-th row of B and store it to u }
      interval(tau[i],n+1,t,left,out_of_interval);
      Bspline_table(u,t,tau[i],k,left);
      { compute values of all nonzero B-Splines to tau[i] and store it to u }

      FOR j:= 1 TO min(k,n-left+k) DO
        BEGIN
          ind:=left+j-k;
          w[ind]:=w[ind]+u[j]*v[i];
        END;
    END;
end; { Get_BTvec }

{-----------------------------------------------------------------------------}

END. { unit BSpline }