UNIT CSSpline;
{ procedures for the computation of constrained smoothing B-Splines           }

INTERFACE
  USES def,       { definitions of constants, types and global variables      }
       error,     { error handler                                             }
       math,      { low level mathematical subroutines                        }
       givens,    { low level procedures for GIVENS plane rotations           }
       smooth,    { procedures for smoothing polynomial B-Splines             }
       bspline,   { procedures to compute B-Splines and l2-approximations     }
       BLSunit;   { procedures for the solution of bounded least squares prob.}

CONST BLSnmax = NMAX;
      BLSmmax = NMAX;

TYPE BLSMatrixmxn = ARRAY[1..BLSmmax,1..BLSnmax] OF real;
     BLSvectorm   = ARRAY[1..BLSmmax] OF real;
     BLSBounds    = ARRAY[1..BLSnmax] OF real;
     BLSConstr    = ARRAY[1..BLSnmax] OF
                      RECORD
                        lower_bound: Boolean;
                        upper_bound: Boolean;
                        bound      : Boolean;
                      END;

VAR Active_Bounds     : BLSConstr;
    lower, upper      : BLSBounds;

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Spline(
        m,                                   { number of datavalues           }
        n,                                   { number of B-Splines            }
        k                   : integer;       { order of B-Splines             }
    VAR t                   : knotsequence;
    VAR Active_Bounds       : BLSconstr;
    VAR lower,
        upper               : BLSbounds;     { bounds on spline coefficients  }
        TypeOfConstraints   : char;
    VAR tau                 : dataabscissae;
    VAR y                   : datavalues;
    VAR alpha               : coefficients;  { to compute                     }
    VAR RES                 : real;          { residual                       }
        try_last_active_set,
        QR_factored         : Boolean;
    VAR B                   : LeastSquareMatrix;
    VAR Q                   : OrthogonalMatrix
    );
{ computes the coefficients alpha of an constrained spline function
  s=\sum_(j=1)^n alpha(j) B_(j,k) (x)

  0.5 || y- B alpha ||^2 ---> min

  s.t.       Lower[i] <= s^p (x) <= Upper[i]     for x in [ t[i],t[i+1] ]
                                                 i=k,...,n

  lower[j], upper[j] j=p+1,..,n are the bounds on the coefficients of the
  p-th derivative, so that s^(p) (x)=\sum_(j=p+1)^n alpha(j) B_(j,k-p)

  TypeOfConstraints       'N' - no constraints
                          'P' - positivity         0 <= s(x)
                          'M' - monotonicity       0 <= s'(x)
                          'C' - convexity          0 <= s''(x)
                          '0' - bounds for s       l <= s(x)   <= u
                          '1' - bounds for s'      l <= s'(x)  <= u
                          '2' - bounds for s''     l <= s''(x) <= u

                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Smoothing_Spline(
        m,                                   { number of datavalues           }
        n,                                   { number of B-Splines            }
        k                   : integer;       { order of B-Splines             }
    VAR t                   : knotsequence;
        r                   : integer;       { order of derivative in smoothing term }
        mu                  : real;          { smoothing parameter            }
    VAR Active_Bounds       : BLSconstr;
    VAR lower,
        upper               : BLSbounds;     { bounds on spline coefficients }
        TypeOfConstraints   : char;
    VAR tau                 : dataabscissae;
    VAR y                   : datavalues;
    VAR alpha               : coefficients;  { to compute                     }
    VAR S                   : SmoothingMatrix;{ Smoothing matrix              }
    VAR RES                 : real;          { residual                       }
        try_last_active_set,
        QR_factored         : Boolean;
    VAR B                   : LeastSquareMatrix;
    VAR Q                   : OrthogonalMatrix
    );
{ computes the coefficients alpha of an constrained smoothing spline function
  s=\sum_(j=1)^n alpha(j) B_(j,k) (x)

  || y- B alpha ||^2 + mu || S * alpha || ---> min

  s.t.       Lower[i] <= s^p (x) <= Upper[i]     for x in [ t[i],t[i+1] ]
                                                 i=k,...,n

  S*alpha is an approximation of s^(r) (x)
  lower[j], upper[j] j=p+1,..,n are the bounds on the coefficients of the
  p-th derivative, so that s^(p) (x)=\sum_(j=p+1)^n alpha(j) B_(j,k-p)

  TypeOfConstraints       'N' - no constraints
                          'P' - positivity         0 <= s(x)
                          'M' - monotonicity       0 <= s'(x)
                          'C' - convexity          0 <= s''(x)
                          '0' - bounds for s       l <= s(x)   <= u
                          '1' - bounds for s'      l <= s'(x)  <= u
                          '2' - bounds for s''     l <= s''(x) <= u

                                                                              }

{-----------------------------------------------------------------------------}

IMPLEMENTATION

{-----------------------------------------------------------------------------}

PROCEDURE UnpackBand( n, k  : integer;
                      VAR B : Leastsquarematrix;
                      VAR A : BLSMatrixmxn
                    );
{ switch from compact storage scheme in B to full storage scheme in A
  upper triangular matrix B is stored row by row
       A[i,i] in B[i,1]
       A[i,j] in B[i,j-i+1]                                                   }
VAR i,j,jend: integer;
BEGIN
  FOR i := 1 TO n  DO
    BEGIN
      FOR j := 1 TO i-1 DO
        A[i,j] := 0.0;
      jend := min( k+i-1,n);
      FOR j := i TO jend DO
        A[i,j] := B[i,j-i+1];
      FOR j := jend+1 TO n DO
        A[i,j] := 0.0;
    END;
END; { UnpackBand }

{-----------------------------------------------------------------------------}

PROCEDURE NoConstraints( n : integer; VAR TTrafo : BLSMatrixmxn);
{ Initialize transformation matrix TTrafo to zero                                 }
VAR i,j : integer;
BEGIN
  FOR i := 1 TO n DO
    FOR j := 1 TO n DO
      TTrafo[ i,j] := 0.0
END; { NoConstraints }

{-----------------------------------------------------------------------------}

PROCEDURE Bounds(     n            : integer;
                  VAR TTrafo       : BLSMatrixmxn
                );
{ Set transformation matrix TTrafo = [ D_0 ] = I                              }

VAR  i: integer;
BEGIN
  FOR i := 1 TO n DO
    TTrafo[i,i] := 1.0;
END; { Bounds }

{-----------------------------------------------------------------------------}

PROCEDURE BoundsForFirstDerivative(
                      n, k        : integer;
                  VAR t           : knotsequence;
                  VAR TTrafo      : BLSMatrixmxn
                                   );
{ Set transformation matrix
           [ D_1  ]
  TTrafo = [      ]
           [    I ]                                                           }
VAR i,row   : integer;
    divisor : real;
BEGIN
  { 1. Computation of D_1 }
  FOR row := 1 TO n-1 DO
    BEGIN
      i := row + 1;
      divisor :=  t[ i+k- 1] - t[i];
      { divisor should be greater than zero, otherwise the derivative does
        not exist }
      TTrafo[ row, row]   := - ( k-1) / divisor;
      TTrafo[ row, row+1] :=   ( k-1) / divisor;
    END{ for row};

  TTrafo[ n,n] := 1.0;
END; { BoundsForFirstDerivative }

{-----------------------------------------------------------------------------}

PROCEDURE BoundsFor2ndDerivative(
                      n, k        : integer;
                  VAR t           : knotsequence;
                  VAR TTrafo      : BLSMatrixmxn
                                 );
{ Set transformation matrix
           [ D_2  ]
  TTrafo = [      ]
           [    I ]                                                           }

VAR  row, col, i, j, r, r_run   : integer;
     divisor, h                 : real;
     S                          : SmoothingMatrix;

BEGIN
  r:=2;
  { Initialization }
  FOR row := 1 TO n DO
    FOR col := 1 TO k DO
      S[ row, col ] := 0.0;

  { 1. Computation of D_1 }
  FOR row := 1 TO n-1 DO
    BEGIN
      i := row + 1;
      divisor :=  t[ i+k- 1] - t[i];
      S[ row, 1] := - ( k-1) / divisor;
      S[ row, 2] := - S[ row, 1];
    END{ for row};
  { now D_1 is stored in S row by row }


  { 2. Multiplication with D2,...,Dl.}

  FOR r_run := 2 TO r DO
    FOR row := 1 TO n - r_run DO
      BEGIN
        i := row + r_run;
        h := t[ i+k-r_run] - t[ i];
        h := ( k- r_run) / h;
        S[ row, 1] := - S[ row, 1] * h;
        FOR col := 2 TO r_run   DO
          S[ row, col] := h * ( S[ row + 1, col -1]  -  S[ row, col]);
        S[ row, r_run+1] := h * S[ row+1, r_run];
      END{ for row};
  {endfor r_run}
  { now D_r is stored in S row by row }
  
  FOR i:=1 TO n-r DO
    FOR j:=1 TO r+1 DO
      TTrafo[i,i-1+j]:=S[i,j];

  FOR i:=n-1 TO n DO
    TTrafo[i,i]:=1.0;
END;

{-----------------------------------------------------------------------------}

PROCEDURE Transformation_A(      n : integer;
                        VAR TTrafo : BLSMAtrixmxn;
                        VAR A      : BLSMatrixmxn
                          );
{ transform the observation matrix A to A * TTrafo^(-1)
  A is overwritten by the transformed matrix
  This step is the reduction to the box constrained problem                   }
VAR i,j,k: integer;
BEGIN
  { Note that A and TTrafo are upper triangular, but A * TTrafo^(-1) not }
  FOR i := 1 TO n DO
    FOR j := i TO n DO
      BEGIN
        A[i,j] := A[i,j]/TTrafo[j,j];
        FOR k := j+1 TO n DO
           A[i,k] := A[i,k] - A[i,j] * TTrafo[ j,k]
      END;
END;

{-----------------------------------------------------------------------------}

PROCEDURE Backtransformation_x(      n : integer;
                            VAR TTrafo : BLSmatrixmxn;
                            VAR      x : coefficients;
                            VAR  alpha : coefficients
                               );
{ backtransformation of the solution of the box constrained problem to the
  solution of the original problem
  alpha=TTrafo^(-1) * x                                                       }
VAR i,j: integer;
    t : real;
BEGIN
  FOR i:=n DOWNTO 1 DO
    BEGIN
      t:=x[i];
      FOR j:=n DOWNTO i+1 DO
        t:=t-TTrafo[i,j ] * alpha[j];
      alpha[i]:=t/TTrafo[ i,i ];
    END;
END;

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Spline(
        m,                                   { number of datavalues           }
        n,                                   { number of B-Splines            }
        k                   : integer;       { order of B-Splines             }
    VAR t                   : knotsequence;
    VAR Active_Bounds       : BLSconstr;
    VAR lower,
        upper               : BLSbounds;     { bounds on spline coefficients  }
        TypeOfConstraints   : char;
    VAR tau                 : dataabscissae;
    VAR y                   : datavalues;
    VAR alpha               : coefficients;  { to compute                     }
    VAR RES                 : real;          { residual                       }
        try_last_active_set,
        QR_factored         : Boolean;
    VAR B                   : LeastSquareMatrix;
    VAR Q                   : OrthogonalMatrix
    );

TYPE  workrealtype     = ARRAY[1..1] OF real;       { We need 3*n }
      workinttype      = ARRAY[1..1] OF integer;    { We need 2*n }
      workinttype_p    =^workinttype;

CONST FirstCall : BOOLEAN       = TRUE;
      workint   : workinttype_p = NIL;

VAR workreal                             : ^workrealtype;
    A,TTrafo                             : ^BLSMatrixmxn;
    c,x                                  : coefficients;
    i,j                                  : integer;
    RES1                                 : real;
    Size                                 : LongInt;

BEGIN
  { Initialization }
  Size:=3*n*SizeOf(workrealtype)+2*SizeOf(BLSmatrixmxn);

  { if first call to this routine }
  IF FirstCall THEN
    BEGIN
      GetMem(workint,2*BLSnmax*SizeOf(workinttype));
      { allocate working space for integer variables }
      FirstCall:=False;
    END;
  { Note: The memory allocated here will be in use throughout the whole
    time the program exists }

  IF MaxAvail < Size THEN
    Error_Handler('Constr. Spline: Not enough memory!')
  ELSE
    BEGIN
      GetMem(workreal,3*n*SizeOf(workrealtype));
      GetMem(A,SizeOf(BLSMatrixmxn));
      GetMem(TTrafo,SizeOf(BLSMatrixmxn));
    END;

  RES:=0.0;

  IF QR_factored=FALSE THEN
    BEGIN
      { compute Givens factorization of B }
      GIVENS_factorization(m,n,k,t,tau,B,Q);
    END
  ELSE
    BEGIN
      { QR-factorization has already been done
        We assume that B contains the upper triangular factor and
        Q the parameters for the reconstruction of the Givens rotations }
    END;

  { apply the same transformation to the righthandside vector y }
  apply_GIVENS_transformation(m,n,k,t,tau,y,Q,c,RES);

  { Now we have transformed the original m x n problem
            B  \alpha  - y
    to the n x n problem
     R \alpha - c  where the upper triangular matrix R is stored in place
    of B                                                                      }

  { Lets now come to the solution of R * \alpha - c subject to simple bounds  }


  IF TypeOfConstraints IN ['P','M','C','0','1','2'] THEN
    BEGIN
      { change storage scheme }
      UnpackBand(n,k,B,A^);

      { initialize transformation matrix TTrafo to zero }
      NoConstraints( n,TTrafo^);

      { compute transformation matrix TTrafo }
      CASE TypeOfConstraints OF
        'P','0': Bounds( n,TTrafo^);
        'M','1': BoundsForFirstDerivative( n,k,t,TTrafo^);
        'C','2': BoundsFor2ndDerivative( n,k,t,TTrafo^);
      END;

      { transform to box constrained problem }
      Transformation_A(n,TTrafo^,A^);

      { solve bounded least squares problem
        || A x - c ||^2  subject to lower <= x <= upper                       }
      BLS(n, n, A^, BLSNmax,c,Active_Bounds, lower, upper, try_last_active_set,
           x, RES1, workreal^,workint^);

      { transform the solution x of the box constrained problem to the
        solution alpha of the original problem                                }
      Backtransformation_x(n,TTrafo^,x,alpha);

      { alpha contains now the solution of the constrained spline problem
             B  \alpha - y

        subject to lower <= D_p \alpha <= upper                               }
    END { if TypeOfConstraints }
  ELSE
    IF TypeOfConstraints='N' THEN
      BEGIN
        Rc_x( n,k,B,alpha,c);
        RES1:=0.0;
      END;

  RES:=Sqrt(Sqr(RES)+Sqr(RES1));

  FreeMem(workreal,3*n*SizeOf(workrealtype));
  FreeMem(A,SizeOf(BLSMatrixmxn));
  FreeMem(TTrafo,SizeOf(BLSMatrixmxn));
END; { Constrained_Spline }

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Smoothing_Spline(
        m,                                   { number of datavalues           }
        n,                                   { number of B-Splines            }
        k                   : integer;       { order of B-Splines             }
    VAR t                   : knotsequence;
        r                   : integer;       { order of derivative in smoothing term }
        mu                  : real;          { smoothing parameter            }
    VAR Active_Bounds       : BLSconstr;
    VAR lower,
        upper               : BLSbounds;     { bounds on spline coefficients }
        TypeOfConstraints   : char;
    VAR tau                 : dataabscissae;
    VAR y                   : datavalues;
    VAR alpha               : coefficients;  { to compute                     }
    VAR S                   : SmoothingMatrix;{ Smoothing matrix              }
    VAR RES                 : real;          { residual                       }
        try_last_active_set,
        QR_factored         : Boolean;
    VAR B                   : LeastSquareMatrix;
    VAR Q                   : OrthogonalMatrix
    );

TYPE  workrealtype     = ARRAY[1..1] OF real;       { We need 3*n }
      workinttype      = ARRAY[1..1] OF integer;    { We need 2*n }
      workinttype_p    =^workinttype;

CONST FirstCall : BOOLEAN       = TRUE;
      workint   : workinttype_p = NIL;

VAR workreal                             : ^workrealtype;
    A,TTrafo                             : ^BLSMatrixmxn;
    Rmu                                  : LeastSquareMatrix;
    c,d,G2,x                             : coefficients;
    i,j                                  : integer;
    RES1                                 : real;
    Size                                 : LongInt;

BEGIN
  { Initialization }
  Size:=3*n*SizeOf(workrealtype)+2*SizeOf(BLSmatrixmxn);

  { if first call to this routine }
  IF FirstCall THEN
    BEGIN
      GetMem(workint,2*BLSnmax*SizeOf(workinttype));
      { allocate working space for integer variables }
      FirstCall:=False;
    END;
  { Note: The memory allocated here will be in use throughout the whole
    time the program exists }

  IF MaxAvail < Size THEN
    Error_Handler('Constr. Smooth. Spline: Not enough memory!')
  ELSE
    BEGIN
      GetMem(workreal,3*n*SizeOf(workrealtype));
      GetMem(A,SizeOf(BLSMatrixmxn));
      GetMem(TTrafo,SizeOf(BLSMatrixmxn));
    END;

  RES:=0.0;

  IF QR_factored=FALSE THEN
    BEGIN
      { compute Givens factorization of B }
      GIVENS_factorization(m,n,k,t,tau,B,Q);
    END
  ELSE
    BEGIN
      { QR-factorization has already been done
        We assume that B contains the upper triangular factor and
        Q the parameters for the reconstruction of the Givens rotations }
    END;

  { apply the same transformation to the righthandside vector y }
  apply_GIVENS_transformation(m,n,k,t,tau,y,Q,c,RES);

  IF mu > 0.0 THEN
    BEGIN
      { Smoothing matrix S has been computed already }
      { GIVENS factorization of
        [       R_B     ]          [ c ]
        [               ] \alpha - [   ]
        [  \sqrt(\mu) S ]          [ 0 ]

        On Output, Rmu contains the upper triangular matrix of the QR
        factorization and d the transformed right hand side                   }

      FOR i:=1 TO n-r DO
        G2[i]:=0.0;
      GIVENS_factorization_of_RBS(n,k,r,mu,B,Rmu,S,c,G2,d,RES1);

      { Now we have transformed the original (m+n-r)x(n) problem
        [       B       ]          [ y ]
        [               ] \alpha - [   ]
        [  \sqrt(\mu) S ]          [ 0 ]
        to the n x n problem
          Rmu \alpha - d                                                      }

      FOR i:=1 TO n DO
        FOR j:=1 TO k DO
          B[i,j]:=Rmu[i,j]; { We save a copy of the triangualr factor in B }


    END { if mu > 0.0 }
  ELSE
    BEGIN
      { Now we have transformed the original m x n problem
               B  \alpha  - y
        to the n x n problem
          R \alpha - c  where the upper triangular matrix R is stored in place
        of B
        For simplicity we store R on Rmu and c on d                           }
      FOR i:=1 TO n DO
        BEGIN
          FOR j:=1 TO k DO
            Rmu[i,j]:=B[i,j];
          d[i]:=c[i];
        END;
      RES1:=0.0; { no additional sum for residual }
    END;

  { Lets now come to the solution of Rmu * \alpha - d subject to simple bounds}

  RES:=Sqrt(Sqr(RES)+Sqr(RES1));

  IF TypeOfConstraints IN ['P','M','C','0','1','2'] THEN
    BEGIN
      { change storage scheme }
      UnpackBand(n,k,Rmu,A^);

      { initialize transformation matrix TTrafo to zero }
      NoConstraints( n,TTrafo^);

      { compute transformation matrix TTrafo }
      CASE TypeOfConstraints OF
        'P','0': Bounds( n,TTrafo^);
        'M','1': BoundsForFirstDerivative( n,k,t,TTrafo^);
        'C','2': BoundsFor2ndDerivative( n,k,t,TTrafo^);
      END;

      { transform to box constrained problem }
      Transformation_A(n,TTrafo^,A^);

      { solve bounded least squares problem
        || A x - d ||^2  subject to lower <= x <= upper                       }
      BLS(n, n, A^, BLSNmax,d,Active_Bounds, lower, upper, try_last_active_set,
          x, RES1, workreal^,workint^);

      { transform the solution x of the box constrained problem to the
        solution alpha of the original problem                                }
      Backtransformation_x(n,TTrafo^,x,alpha);

      { alpha contains now the solution of the constrained smoothing
        spline problem
            [       B       ]          [ y ]
            [               ] \alpha - [   ]
            [  \sqrt(\mu) S ]          [ 0 ]

        subject to lower <= D_p \alpha <= upper                               }
    END { if TypeOfConstraints }
  ELSE
    IF TypeOfConstraints='N' THEN
      BEGIN
        Rc_x( n,k,Rmu,alpha,d);
        RES1:=0.0;
      END;

  RES:=Sqrt(Sqr(RES)+Sqr(RES1));

  FreeMem(workreal,3*n*SizeOf(workrealtype));
  FreeMem(A,SizeOf(BLSMatrixmxn));
  FreeMem(TTrafo,SizeOf(BLSMatrixmxn));
END; { Constrained_Smoothing_Spline }

{-----------------------------------------------------------------------------}

END. { Unit CSSpline }