UNIT Gauss;
{ Generalized Gauss Newton methods for nonlinear least squares problems       }
{$R-}
INTERFACE

USES crt,
     error,       { Error handler                                             }
     def,         { definitions of constants, types and global variables      }
     Jacobi,      { computation of Jacobi-Matrizes                            }
     BLAS,        { Basic Linear Algebra Subroutines                          }
     Lawson,      { procedures for constrained linear least squares problems  }
     BSpline,     { procedures to compute B-Splines and l2-approximations     }
     smooth,      { procedures to compute smoothing B-Splines                 }
     CSSpline;    { procedures for constrained smoothing B-Splines            }

{-----------------------------------------------------------------------------}

PROCEDURE Get_Constraints_abs ( VAR A     : CSTRMatrix;
                                VAR b     : CSTRVector;
                                VAR ncstr : integer;
                                    l     : integer;
                                VAR t     : knotsequence;
                                VAR per   : free_knots;
                                    eps   : real
                              );
{ procedure to compute the constraints for the spline approximation with
  free knots
                         A * ttilde >= b
  These constraints result from the numerical realization of the knot
  conditions
                   t[ per[j] ]     - t[ per[j] - 1 ] >= EPS
                   t[ per[j] + 1 ] - t[ per[j] ]     >= EPS    j = 1..l

  Here eps is an absolute measure for the minimal distance between knots.


  Output: ncstr          number of constraints     l+1 <= ncstr <= 2*l
          Matrix A       dimension  ncstr x l
          Vector b       dimension  ncstr                                     }

{-----------------------------------------------------------------------------}

PROCEDURE Get_Constraints_rel ( VAR A     : CSTRMatrix;
                                VAR b     : CSTRVector;
                                VAR ncstr : integer;
                                    l     : integer;
                                VAR t     : knotsequence;
                                VAR per   : free_knots;
                                    eps   : real
                              );
{ procedure to compute the constraints for the spline approximation with
  free knots
                         A * ttilde >= b
  These constraints result from the numerical realization of the knot
  conditions
    t[ per[j] ]     - t[ per[j] - 1 ] >= EPS (t[per[j]+1]-t[per[j]-1])
    t[ per[j] + 1 ] - t[ per[j] ]     >= EPS (t[per[j]+1]-t[per[j]-1])
                                                                    j = 1..l
  Here eps is an relative measure for the minimal distance between knots.

  Output: ncstr          number of constraints     ncstr:= 2*l (!!)
          Matrix A       dimension  ncstr x l
          Vector b       dimension  ncstr                                     }

{-----------------------------------------------------------------------------}

PROCEDURE Regularized_Generalized_Gauss_Newton (
                        m,n,                        { dimension             }
                        k           : integer;      { order of B-Splines    }
                        VAR t       : knotsequence; { contains free knots   }
                        l           : integer;      { number of free knots  }
                        VAR per     : free_knots;   { pointer to free knots }
                        VAR step    : integer;      { maximal no. of steps  }
                        VAR RetCode : integer;      { Return Code           }
                        VAR normGradient : real;    { Norm of Gradient      }
                        VAR gradTp  : real          { J^T * F * p           }
                        );
{ Approximation by splines with free knots (no smoothing, no constraints on
  derivatives)

  Regularized generalized Gauss-Newton Iteration for linearly constrained
  problem. The regularization is due to an idea of Dennis/Schnabel for
  unconstrained optimization with Newton's method. It is similar to an
  Trust Region approach, but will not be used for globalization of the
  scheme.

  On return
  t:        knotsequence at solution point
  step:     contains the number of steps
  RetCode:  Return Code
            =1     || F^(k+1) ||_2     <= EPS0
                   zero residual ?

            =2     || J^k^T * F^k ||_2 <= EPSGrad
                   stationary point ?

            =3     |F^k^T * J^k * s^k| <= EPSDesc
                   almost no descent direction found

            =4     || x^(k+1)-x^k ||_2 <= EPS1 * ( || x^k ||_2 + EPS2 )
                   steps too small

            =5     | || F^(k+1) ||_2 - || F^k ||_2 | <= EPS3 || F^k ||_2

            =6     step > MAXSTEP, too much work

            =7     failed otherwise                                           }

{-----------------------------------------------------------------------------}

PROCEDURE Smoothing_Regularized_Generalized_Gauss_Newton (
                        m,n,                        { dimension             }
                        k           : integer;      { order of B-Splines    }
                        VAR t       : knotsequence; { contains free knots   }
                        r           : integer;      { order of derivative   }
                        mu          : real;         { Smoothing parameter   }
                        l           : integer;      { number of free knots  }
                        VAR per     : free_knots;   { pointer to free knots }
                        VAR step    : integer;      { maximal no. of steps  }
                        VAR RetCode : integer;      { Return Code           }
                        VAR normGradient : real;    { Norm of Gradient      }
                        VAR gradTp  : real          { J^T * F * p           }
                        );
{ Approximation by smoothing splines with free knots (no constraints on
  derivatives)

  Regularized generalized Gauss-Newton Iteration for linearly constrained
  problem. The regularization is due to an idea of Dennis/Schnabel for
  unconstrained optimization with Newton's method. It is similar to an
  Trust Region approach, but will not be used for globalization of the
  scheme.

  On return
  t:        knotsequence at solution point
  step:     contains the number of steps
  RetCode:  Return Code
            =1     || F^(k+1) ||_2     <= EPS0
                   zero residual ?

            =2     || J^k^T * F^k ||_2 <= EPSGrad
                   stationary point ?

            =3     |F^k^T * J^k * s^k| <= EPSDesc
                   almost no descent direction found

            =4     || x^(k+1)-x^k ||_2 <= EPS1 * ( || x^k ||_2 + EPS2 )
                   steps too small

            =5     | || F^(k+1) ||_2 - || F^k ||_2 | <= EPS3 || F^k ||_2

            =6     step > MAXSTEP, too much work

            =7     failed otherwise                                           }

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Regularized_Generalized_Gauss_Newton (
                   m,n,                              { dimension             }
                   k                 : integer;      { order of B-Splines    }
               VAR t                 : knotsequence; { contains free knots   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                   l                 : integer;      { number of free knots  }
               VAR per               : free_knots;   { pointer to free knots }
               VAR step              : integer;      { maximal no. of steps  }
               VAR RetCode           : integer;      { Return Code           }
               VAR normGradient      : real;         { Norm of Gradient      }
               VAR gradTp            : real          { J^T * F * p           }
                        );
{ Approximation by splines with free knots and constraints on
  derivatives (no smoothing)

  Regularized generalized Gauss-Newton Iteration for linearly constrained
  problem. The regularization is due to an idea of Dennis/Schnabel for
  unconstrained optimization with Newton's method. It is similar to an
  Trust Region approach, but will not be used for globalization of the
  scheme.

  On return
  t:        knotsequence at solution point
  step:     contains the number of steps
  RetCode:  Return Code
            =1     || F^(k+1) ||_2     <= EPS0
                   zero residual ?

            =2     || J^k^T * F^k ||_2 <= EPSGrad
                   stationary point ?

            =3     |F^k^T * J^k * s^k| <= EPSDesc
                   almost no descent direction found

            =4     || x^(k+1)-x^k ||_2 <= EPS1 * ( || x^k ||_2 + EPS2 )
                   steps too small

            =5     | || F^(k+1) ||_2 - || F^k ||_2 | <= EPS3 || F^k ||_2

            =6     step > MAXSTEP, too much work

            =7     failed otherwise                                           }

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Smoothing_Regularized_Generalized_Gauss_Newton (
                   m,n,                              { dimension             }
                   k                 : integer;      { order of B-Splines    }
               VAR t                 : knotsequence; { contains free knots   }
                   r                 : integer;      { order of derivative   }
                   mu                : real;         { Smoothing parameter   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                   l                 : integer;      { number of free knots  }
               VAR per               : free_knots;   { pointer to free knots }
               VAR step              : integer;      { maximal no. of steps  }
               VAR RetCode           : integer;      { Return Code           }
               VAR normGradient      : real;         { Norm of Gradient      }
               VAR gradTp            : real          { J^T * F * p           }
                        );
{ Approximation by smoothing splines with free knots and constraints on
  derivatives

  Regularized generalized Gauss-Newton Iteration for linearly constrained
  problem. The regularization is due to an idea of Dennis/Schnabel for
  unconstrained optimization with Newton's method. It is similar to an
  Trust Region approach, but will not be used for globalization of the
  scheme.

  On return
  t:        knotsequence at solution point
  step:     contains the number of steps
  RetCode:  Return Code
            =1     || F^(k+1) ||_2     <= EPS0
                   zero residual ?

            =2     || J^k^T * F^k ||_2 <= EPSGrad
                   stationary point ?

            =3     |F^k^T * J^k * s^k| <= EPSDesc
                   almost no descent direction found

            =4     || x^(k+1)-x^k ||_2 <= EPS1 * ( || x^k ||_2 + EPS2 )
                   steps too small

            =5     | || F^(k+1) ||_2 - || F^k ||_2 | <= EPS3 || F^k ||_2

            =6     step > MAXSTEP, too much work

            =7     failed otherwise                                           }

{-----------------------------------------------------------------------------}

IMPLEMENTATION
{$define mixed} {backtracking with quadratic/cubic polynomials}

VAR OutputFile : Text;

PROCEDURE Diagnostic ( step     : integer; { number of iterate            }
                       normF    : real;    { norm of residual             }
                       cond     : real;    { condition number of Jacobian }
                       gradTp   : real;    { J^T * F *p                   }
                       stepback : integer; { number of backtracking steps }
                       gamma    : real;    { step length in line-search   }
                       relerror : real;    { || x^k+1 - x^k || / || x^k ||}
                       relFerror: real;    { | || F^k+1 || - || F^k || | / || F^k || }
                       mode     : STRING;  { Gauss-Newton or regul. GN    }
                       job      : integer  { what to do                   }
                     );
{ Print diagnostic of generalized Gauss-Newton method

         -1   open file
  job =   0   write to file
         +1   close file                                                      }

BEGIN
  CASE job OF
    -1: BEGIN
          Assign(OutputFile,'diag.txt');
          Rewrite(OutputFile);
            Write(OutputFile,' k ','  Residual norm   ','    Condition number ','   Projected Gradient ');
          WriteLn(OutputFile,'      rel. Error     ','    rel. Error in F  ','   Backtracking   ','            mode ');
          WriteLn(OutputFile,'':107,'steps   steplength');
          WriteLn(OutputFile);
        END;
     0: BEGIN
            Write(OutputFile,step : 2, normF : 20,' ',cond : 20,' ',gradTp : 20,' ');
          WriteLn(OutputFile,relError:20,' ',relFerror:20,' ',stepback : 2,'  ',gamma : 20,' ',mode);
        END;
     1: BEGIN
          Close(OutputFile);
        END;
  END;
END; { Diagnostic }

{-----------------------------------------------------------------------------}

PROCEDURE Get_Constraints_abs ( VAR A     : CSTRMatrix;
                                VAR b     : CSTRVector;
                                VAR ncstr : integer;
                                    l     : integer;
                                VAR t     : knotsequence;
                                VAR per   : free_knots;
                                    eps   : real
                              );

VAR i, j : integer;

BEGIN

  FOR i:=1 TO MAXNCSTR DO
    BEGIN
      b[i]:= 0.0;
      FOR j:=1 TO LMAX DO
        A[i,j]:= 0.0;
    END;                    { A:=0 ; b:=0 }

  IF l=1 THEN
    BEGIN
      ncstr:= 1;
      A[ncstr,1]:=  1.0;
      b[ncstr]  :=  t[per[1]-1] + EPS;
      Inc(ncstr);
      A[ncstr,1]:= -1.0;
      b[ncstr]  := -t[per[1]+1] + EPS;
    END
  ELSE
    BEGIN
      ncstr:= 1;             { first free knot separat }
      A[ncstr,1]:=  1.0;
      b[ncstr]  :=  t[per[1]-1] + EPS;
      Inc(ncstr);
      IF (per[1]+1 = per[2]) THEN
        BEGIN                { right knot is free }
          A[ncstr,1]:=-1.0;
          A[ncstr,2]:= 1.0;
          b[ncstr]  := EPS;
        END
      ELSE
        BEGIN                { right knot is not free }
          A[ncstr,1]:=-1.0;
          b[ncstr]  :=-t[per[1]+1] + EPS;
        END;

      FOR j:=2 TO l-1 DO     { now the inner free knots }
        BEGIN

          { left knot }
          IF (per[j]-1=per[j-1]) THEN
            BEGIN
              { left knot is also free, nothing to do }
            END
          ELSE
            BEGIN            { left knot is not free }
              Inc(ncstr);
              A[ncstr,j]:=  1.0;
              b[ncstr]  :=  t[per[j]-1] + EPS;
            END;

          { right knot }
          Inc(ncstr);
          IF (per[j]+1 = per[j+1]) THEN
            BEGIN            { right knot is free }
              A[ncstr,j]  :=-1.0;
              A[ncstr,j+1]:= 1.0;
              b[ncstr]    := EPS;
            END
          ELSE
            BEGIN            { right knot is not free }
              A[ncstr,j]:=-1.0;
              b[ncstr]  :=-t[per[j]+1] + EPS;
            END;

        END; { for j }

      IF (per[l]-1=per[l-1]) THEN       { last free knot }
        BEGIN
          { left knot is also free, nothing to do }
        END
      ELSE
        BEGIN                { left knot is not free }
          Inc(ncstr);
          A[ncstr,l]:= 1.0;
          b[ncstr]  := t[per[l]-1] + EPS;
        END;

      Inc(ncstr);
      A[ncstr,l]:= -1.0;
      b[ncstr]  := -t[per[l]+1] + EPS;
    END; { else l=1 }

END; { Get_Constraints_abs }

{-----------------------------------------------------------------------------}

PROCEDURE Get_Constraints_rel( VAR A     : CSTRMatrix;
                               VAR b     : CSTRVector;
                               VAR ncstr : integer;
                                   l     : integer;
                               VAR t     : knotsequence;
                               VAR per   : free_knots;
                                   eps   : real
                             );

VAR i, j : integer;
    eps1 : real;

BEGIN

  FOR i:=1 TO MAXNCSTR DO
    BEGIN
      b[i]:= 0.0;
      FOR j:=1 TO LMAX DO
        A[i,j]:= 0.0;
    END;                    { A:=0 ; b:=0 }

  eps1:=1.0-eps;
  ncstr:=2*l;

  IF l=1 THEN               { only one free knot }
    BEGIN
      A[1,1]:=  1.0;
      b[1]  :=  eps1*t[per[1]-1] + eps*t[per[1]+1];
      A[2,1]:= -1.0;
      b[2]  := -eps*t[per[1]-1] -eps1*t[per[1]+1];
    END
  ELSE
    BEGIN  { We handle the first and last free knot seperatly }

      IF (per[1]+1 = per[2]) THEN
        BEGIN                { right knot is free }
          A[1,1]:= 1.0;
          A[1,2]:= -eps;
          b[1]  := eps1*t[per[1]-1];
          A[2,1]:=-1.0;
          A[2,2]:=eps1;
          b[2]  :=-eps*t[per[1]-1];
        END
      ELSE
        BEGIN                { right knot is not free }
          A[1,1]:= 1.0;
          b[1]  := eps1*t[per[1]-1]+eps*t[per[1]+1];
          A[2,1]:=-1.0;
          b[2]  :=-eps1*t[per[1]+1]-eps*t[per[1]-1];
        END;

      FOR j:=2 TO l-1 DO     { now the inner free knots }
        BEGIN
        { We have to consider four cases
             left and right knot free
             left knot free, but not right knot
             right knot free, but not left knot
             left and right knot not free
        }

          IF (per[j]-1=per[j-1]) AND (per[j]+1=per[j+1]) THEN
            BEGIN { left and right knot free }
              A[2*j-1,j-1]:=-eps1;
              A[2*j-1,j]  :=1.0;
              A[2*j-1,j+1]:=-eps;
              b[2*j-1]    :=0.0;
              A[2*j,j-1]  :=eps;
              A[2*j,j]    :=-1.0;
              A[2*j,j+1]  :=eps1;
              b[2*j]      :=0.0;
            END;

          IF (per[j]-1=per[j-1]) AND (per[j]+1<>per[j+1]) THEN
            BEGIN { left knot free, but not right knot }
              A[2*j-1,j-1]:=-eps1;
              A[2*j-1,j]  :=1.0;
              b[2*j-1]    :=eps*t[per[j]+1];
              A[2*j,j-1]  :=eps;
              A[2*j,j]    :=-1.0;
              b[2*j]      :=-eps1*t[per[j]+1];
            END;

          IF (per[j]-1<>per[j-1]) AND (per[j]+1=per[j+1]) THEN
            BEGIN { right knot free, but not left knot }
              A[2*j-1,j]  :=1.0;
              A[2*j-1,j+1]:=-eps;
              b[2*j-1]    :=eps1*t[per[j]-1];
              A[2*j,j]    :=-1.0;
              A[2*j,j+1]  :=eps1;
              b[2*j]      :=-eps*t[per[j]-1];
            END;

          IF (per[j]-1<>per[j-1]) AND (per[j]+1<>per[j+1]) THEN
            BEGIN { left and right knot not free }
              A[2*j-1,j]  :=1.0;
              b[2*j-1]    :=eps1*t[per[j]-1]+eps*t[per[j]+1];
              A[2*j,j]    :=-1.0;
              b[2*j]      :=-eps1*t[per[j]+1]-eps*t[per[j]-1];
            END;


        END; { for j }

      IF (per[l]-1=per[l-1]) THEN       { last free knot }
        BEGIN { left knot free }
          A[2*l-1,l-1]:=-eps1;
          A[2*l-1,l]  :=1.0;
          b[2*l-1]    :=eps*t[per[l]+1];
          A[2*l,l-1]  :=eps;
          A[2*l,l]    :=-1.0;
          b[2*l]      :=-eps1*t[per[l]+1];
        END
      ELSE
        BEGIN { left knot not free }
          A[2*l-1,l]  :=1.0;
          b[2*l-1]    :=eps1*t[per[l]-1]+eps*t[per[l]+1];
          A[2*l,l]    :=-1.0;
          b[2*l]      :=-eps1*t[per[l]+1]-eps*t[per[l]-1];
        END;

    END; { else l<>1 }

END; { Get_Constraints_rel }

{-----------------------------------------------------------------------------}

PROCEDURE Regularized_Generalized_Gauss_Newton (
                        m,n,                        { dimension             }
                        k           : integer;      { order of B-Splines    }
                        VAR t       : knotsequence; { contains free knots   }
                        l           : integer;      { number of free knots  }
                        VAR per     : free_knots;   { pointer to free knots }
                        VAR step    : integer;      { maximal no. of steps  }
                        VAR RetCode : integer;      { Return Code           }
                        VAR normGradient : real;    { Norm of Gradient      }
                        VAR gradTp  : real          { J^T * F * p           }
                        );

CONST alph=0.2;      { in (0,1.0) factor to reduce step size in line search }
      delta=0.2;     { in (0,0.5) sufficient descent in line search ?       }

LABEL return;

TYPE  workrealtype     = ARRAY[1..1] OF real;
      { We need (l+1)*(2+ncstr+2*ncstr) }
      workinttype      = ARRAY[1..1] OF integer;
      { We need (l+ncstr) }
      Reg_JacobiMatrix = ARRAY[1..MMAX+LMAX,1..LMAX] OF real;
      { We need (m+l)*l }
      Reg_F            = ARRAY[1..1] OF real;
      { We need (m+l) }

VAR cond,gamma,normF,normFnew,normdiff,
    normXold,temp,Hnorm,Hij,lambda       : real;
    F,Fsave                              : datavalues;
    tnew                                 : knotsequence;
    x,p                                  : coefficients;
    maxstep,step1,i,jj,ncstr,LSIError,kk : integer;
    Descent_Test,convergent              : Boolean;
    Size                                 : LongInt;
    workreal                             : ^workrealtype;
    workint                              : ^workinttype;
    J,JSave                              : ^JacobiMatrix;
    Acstr,Asave                          : ^CSTRMatrix;
    bcstr,bsave                          : ^CSTRVector;
    Jreg                                 : ^Reg_JacobiMatrix;
    Freg                                 : ^Reg_F;
    mode                                 : STRING[12];

{$ifdef mixed}
    gamma_new,f1,f2,aa,bb,disc,gamma_prev,normFprev  : real;
{$endif}

BEGIN
  { Output of intermediate results in File}
  { Open file }
  Diagnostic(0,0,0,0,0,0,0,0,'',-1);

  { Initialization }
  ncstr:=2*l;    { maximal number of constraints to allocate working space }
  Size:=0;
  Inc(Size,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  Inc(Size,(l+ncstr)*SizeOf(workinttype));
  Inc(Size,2*SizeOf(JacobiMatrix));
  Inc(Size,2*SizeOf(CSTRMatrix));
  Inc(Size,2*SizeOf(CSTRVector));
  Inc(Size,SizeOf(Reg_JacobiMatrix));
  Inc(Size,(m+l)*SizeOf(Reg_F));

  IF MaxAvail < Size THEN
    Error_Handler('Gauss-Newton-Method: Not enough memory!')
  ELSE
    BEGIN
      GetMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
      GetMem(workint,(l+ncstr)*SizeOf(workinttype));
      GetMem(J,SizeOf(JacobiMatrix));
      GetMem(Jsave,SizeOf(JacobiMatrix));
      GetMem(Acstr,SizeOf(CSTRMatrix));
      GetMem(Asave,SizeOf(CSTRMatrix));
      GetMem(bcstr,SizeOf(CSTRVector));
      GetMem(bsave,SizeOf(CSTRVector));
      GetMem(Jreg,SizeOf(Reg_JacobiMatrix));
      GetMem(Freg,(m+l)*SizeOf(Reg_F));
    END;

  { Get linear inequality constraints for problem }
  IF Constraints_abs
    THEN
      Get_Constraints_abs(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots)
    ELSE
      Get_Constraints_rel(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots);

  gamma:=1.0;
  maxstep:=step;
  step :=0;
  Copy1(n+k,t,tnew); { tnew := t, We change only t[per] in this procedure! }
  l2Appr(m,n,k,t,tau,y,alpha,RES); { compute F(t) }
  normFnew:=RES;

  REPEAT
    Get_F(m,n,k,t,tau,y,alpha,F);
    normF:= normFnew;

    IF Kaufman_Model THEN
      Kaufman_Jacobi_Matrix(m,n,k,t,l,per,J^)
    ELSE
      Golub_Jacobi_Matrix(m,n,k,t,l,per,J^,F);

    { compute the gradient J^T * F }
    FOR jj:=1 TO l DO
      BEGIN
        temp:=0.0;
        FOR i:=1 TO m DO
          temp:=temp + J^[i,jj]*F[i];
        x[jj]:=temp;
      END;
    normGradient:=nrm2(l,x,1);

    { Save the Jacobian J and the residual F for further computations }
    FOR i:=1 TO m DO
      BEGIN
        FOR jj:=1 TO l DO
          Jsave^[i,jj]:=J^[i,jj];
        Fsave[i]:=F[i];
      END;

    { restore the constraints which could have been overwritten in a previous
      step }
    FOR i:=1 TO ncstr DO
      BEGIN
        FOR jj:=1 TO l DO
          Acstr^[i,jj]:=Asave^[i,jj];
        bcstr^[i]:=bsave^[i];
      END;

    { Compute Acstr * \tilde(t) }
    FOR i:=1 TO ncstr DO
      BEGIN
        temp:=0.0;
        FOR jj:=1 TO l DO
          temp:=temp + Acstr^[i,jj] * t[ per[jj] ];
        { Now temp contains i-th component of Acstr * \tilde(t) }
        bcstr^[i]:=bcstr^[i]-temp; { build right hand side }
      END;


    FOR i:=1 TO m DO
      F[i]:=-F[i];

    { solve linear least squares problem with linear inequality constraints }
    LSI(J^,LMAX,m,l,Acstr^,LMAX,ncstr,p,F,bcstr^,workreal^,workint^,LSIError);
    { J^, Acstr^, F, and bcstr^ will be overwritten with transformed values!
      If J is nearly rank deficient then only J has been changed, all other
      vectors and matrices (esp. Acstr^ and bcstr^ have their old values.    }

    { estimate condition number of triangular factor stored in place of J^ }
    IF Abs(J^[l,l])> macheps THEN
      cond:=Condition_Estimate (J^,LMAX,l,workreal^)
    ELSE
      cond:=2.0/Sqrt(macheps);

    mode:='     GN-step';
    IF LSIError <> 1 THEN
      BEGIN
        mode:='reg. GN-step';
        IF LSIError<>5 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! ');
            GOTO return;
          END;
        { We try a regularized Gauss-Newton step. }

        { Compute || J^T * J ||_1

               | H |_1 = max_(j=1..l) \sum_(i=1)^l abs(H(i,j))
        }
        Hnorm:=0.0;
        FOR jj:=1 TO l DO
          BEGIN
            temp:=0.0;
            FOR i:=1 TO l DO
              BEGIN
                Hij:=0.0;
                FOR kk:=1 TO m DO
                  Hij:=Hij+Jsave^[kk,i] * Jsave^[kk,jj];
                temp:=temp+Abs(Hij);
              END;
            { temp contains now the sum of absolute values of the j-th column }
            IF Hnorm < temp THEN
              Hnorm:=temp;
          END; { for jj:=1 to l do }

        lambda:=Sqrt(l*macheps) * Hnorm;
        { Form new Matrix }
        FOR i:=1 TO m DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[i,jj]:=Jsave^[i,jj];
            Freg^[i]:=-Fsave[i];
          END;

        FOR i:=1 TO l DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[m+i,jj]:=0.0;
            Freg^[m+i]:=0.0;
          END;

        FOR i:=1 TO l DO
          Jreg^[m+i,i]:=Sqrt(lambda);

        {        (       J           )            (  F )
          Jreg = (                   )     Freg = (    )
                 (  sqrt(lambda) * I )            (  0 )
        }

        LSI(Jreg^,LMAX,m+l,l,Acstr^,LMAX,ncstr,p,Freg^,bcstr^,workreal^,workint^,LSIError);

        { estimate condition number of triangular factor stored in place of Jreg^ }
        IF Abs(Jreg^[l,l])> macheps THEN
          cond:=Condition_Estimate (Jreg^,LMAX,l,workreal^)
        ELSE
          cond:=2.0/Sqrt(macheps);

        IF LSIError<>1 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! Matrix has not full rank! ');
            GOTO return;
          END;
      END;  { end of computation of regul. Gauss-Newton step }

    { s[1], ... ,s[l] contains descent direction for constrained problem
      We work instead with p=-s }
    FOR i:=1 TO l DO
      p[i]:=-p[i];

    gradTp:=dot(l,x,1,p,1);         { F^T * J * p }
    IF gradTp < 0.0 THEN
      GOTO return; { -p is no longer descent direction }

    FOR jj:=1 TO l DO
        x[jj]:=t[ per[jj] ];
    normXold:= nrm2(l,x,1);

    step1:=0;

    { begin line search }
    REPEAT
      FOR jj:=1 TO l DO
        tnew[ per[jj] ]:=t[ per[jj] ] - gamma * p[jj]; { x = x - gamma * p }

      l2Appr(m,n,k,tnew,tau,y,alpha,RES);
      normFnew:=RES;

      { Armijo-Goldstein descent test }
      IF (Sqr(normF)-Sqr(normFnew)) >= 2.0*delta*gamma*gradTp THEN
        Descent_Test:=true
      ELSE Descent_Test:=False;

{$ifndef mixed}
      gamma:=gamma*alph;
{$else}
      IF Descent_test=False THEN
        BEGIN { reduce step length }
          IF gamma=1.0 THEN { first backtrack, quadratic fit }
            gamma_new:=0.5*gradTp/(0.5*Sqr(normFnew)-0.5*Sqr(normF)+gradTp)
          ELSE
            BEGIN  { all subsequent backtracks, cubic fit }
              f1:=0.5*Sqr(normFnew)-0.5*Sqr(normF)+gamma*gradTp;
              f2:=0.5*Sqr(normFprev)-0.5*Sqr(normF)+gamma_prev*gradTp;
              aa:=1.0/(gamma-gamma_prev)*(f1/Sqr(gamma)-f2/Sqr(gamma_new));
              bb:=1.0/(gamma-gamma_prev)*(-gamma_prev*f1/Sqr(gamma)-gamma*f2/Sqr(gamma_prev));
              disc:=Sqr(bb)+3.0*aa*gradTp;
              IF aa = 0.0 THEN
                gamma_new:=gradTp/(2.0*bb)
              ELSE
                gamma_new:=(-bb+Sqrt(disc))/(3.0*aa);
              IF gamma_new>0.5*gamma THEN
                gamma_new:=0.5*gamma;
            END; { cubic fit }
          gamma_prev:=gamma;
          normFprev:=normFnew;
          IF gamma_new<=0.1*gamma THEN
            gamma:=0.1*gamma
          ELSE
            gamma:=gamma_new;
        END; { reduce step length }
{$endif}

      step1:=step1+1;
      normdiff:=Abs(gamma)*nrm2(l,p,1);
    UNTIL (Descent_Test=True) OR (normdiff<=EPS1*(normXold+EPS2));
    { end line search }

    FOR jj:=1 TO l DO
       t[ per[jj] ]:=tnew[ per[jj] ]; { t := tnew }

    convergent:= (normFnew<=EPS0) OR (normdiff<=EPS1*(normXold+EPS2)) OR
    (Abs(normFnew-normF)<=EPS3*normF) OR (normGradient<=EPSGrad) OR (Abs(gradTp)<=EPSDesc);

{$ifndef mixed}
    gamma:=gamma/alph;
{$endif}

    Diagnostic(step,normFnew,cond,gradTp,step1,gamma,normdiff/normXold,(normFnew-normF)/normF,mode,0);

{$ifdef mixed}
    gamma:=1.0;
{$else}
    IF gamma/alph <= 1.0 THEN gamma:= gamma/alph ELSE gamma:=1.0;
{$endif}

    step:=step+1;
  UNTIL (step>MAXSTEP) OR convergent;

  { Close file }
  Diagnostic(0,0,0,0,0,0,0,0,'',1);

return:
  step:=step-1;
  IF normFnew<=EPS0 THEN
    RetCode:=1
  ELSE
  IF normGradient<=EPSGrad THEN
    RetCode:=2
  ELSE
  IF Abs(gradTp)<=EPSDesc THEN
    RetCode:=3
  ELSE
  IF normdiff<=EPS1*(normXold+EPS2) THEN
    RetCode:=4
  ELSE
  IF Abs(normFnew-normF) <=EPS3*normF THEN
    RetCode:=5
  ELSE
  IF step >= MAXSTEP THEN
    RetCode:=6
  ELSE
    RetCode:=7;  { failed otherwise }

  ncstr:=2*l;    { maximal number of constraints to deallocate working space }

  FreeMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  FreeMem(workint,(l+ncstr)*SizeOf(workinttype));
  FreeMem(J,SizeOf(JacobiMatrix));
  FreeMem(Jsave,SizeOf(JacobiMatrix));
  FreeMem(Acstr,SizeOf(CSTRMatrix));
  FreeMem(Asave,SizeOf(CSTRMatrix));
  FreeMem(bcstr,SizeOf(CSTRVector));
  FreeMem(bsave,SizeOf(CSTRVector));
  FreeMem(Jreg,SizeOf(Reg_JacobiMatrix));
  FreeMem(Freg,(m+l)*SizeOf(Reg_F));
END; { Regularized_Generalized_Gauss_Newton }

{-----------------------------------------------------------------------------}

PROCEDURE Smoothing_Regularized_Generalized_Gauss_Newton (
                        m,n,                        { dimension             }
                        k           : integer;      { order of B-Splines    }
                        VAR t       : knotsequence; { contains free knots   }
                        r           : integer;      { order of derivative   }
                        mu          : real;         { Smoothing parameter   }
                        l           : integer;      { number of free knots  }
                        VAR per     : free_knots;   { pointer to free knots }
                        VAR step    : integer;      { maximal no. of steps  }
                        VAR RetCode : integer;      { Return Code           }
                        VAR normGradient : real;    { Norm of Gradient      }
                        VAR gradTp  : real          { J^T * F * p           }
                        );

CONST alph=0.2;      { in (0,1.0) factor to reduce step size in line search }
      delta=0.2;     { in (0,0.5) sufficient descent in line search ?       }

LABEL return;

TYPE  workrealtype     = ARRAY[1..1] OF real;
      { We need (l+1)*(2+ncstr+2*ncstr) }
      workinttype      = ARRAY[1..1] OF integer;
      { We need (l+ncstr) }
      Reg_JacobiMatrix = ARRAY[1..MMAX+NMAX+LMAX,1..LMAX] OF real;
      { We need (m+n-r+l)*l }
      Reg_G            = ARRAY[1..1] OF real;
      { We need (m+n-r+l) }

VAR cond,gamma,normG,normGnew,normdiff,
    normXold,temp,Hnorm,Hij,lambda       : real;
    G,Gsave                              : SmoothingVector;
    tnew                                 : knotsequence;
    x,p                                  : coefficients;
    maxstep,step1,i,jj,ncstr,LSIError,kk : integer;
    Descent_Test,convergent              : Boolean;
    Size                                 : LongInt;
    workreal                             : ^workrealtype;
    workint                              : ^workinttype;
    J,JSave                              : ^SmoothJacobiMatrix;
    Acstr,Asave                          : ^CSTRMatrix;
    bcstr,bsave                          : ^CSTRVector;
    Jreg                                 : ^Reg_JacobiMatrix;
    Greg                                 : ^Reg_G;
    S                                    : SmoothingMatrix;
    mode                                 : STRING[12];

{$ifdef mixed}
    gamma_new,f1,f2,aa,bb,disc,gamma_prev,normGprev  : real;
{$endif}

BEGIN
  { Output of intermediate results in File}
  { Open file }
  Diagnostic(0,0,0,0,0,0,0,0,'',-1);

  { Initialization }
  ncstr:=2*l;    { maximal number of constraints to allocate working space }
  Size:=0;
  Inc(Size,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  Inc(Size,(l+ncstr)*SizeOf(workinttype));
  Inc(Size,2*SizeOf(SmoothJacobiMatrix));
  Inc(Size,2*SizeOf(CSTRMatrix));
  Inc(Size,2*SizeOf(CSTRVector));
  Inc(Size,SizeOf(Reg_JacobiMatrix));
  Inc(Size,(m+n-r+l)*SizeOf(Reg_G));

  IF MaxAvail < Size THEN
    Error_Handler('Gauss-Newton-Method: Not enough memory!')
  ELSE
    BEGIN
      GetMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
      GetMem(workint,(l+ncstr)*SizeOf(workinttype));
      GetMem(J,SizeOf(SmoothJacobiMatrix));
      GetMem(Jsave,SizeOf(SmoothJacobiMatrix));
      GetMem(Acstr,SizeOf(CSTRMatrix));
      GetMem(Asave,SizeOf(CSTRMatrix));
      GetMem(bcstr,SizeOf(CSTRVector));
      GetMem(bsave,SizeOf(CSTRVector));
      GetMem(Jreg,SizeOf(Reg_JacobiMatrix));
      GetMem(Greg,(m+n-r+l)*SizeOf(Reg_G));
    END;

  { Get linear inequality constraints for problem }
  IF Constraints_abs
    THEN
      Get_Constraints_abs(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots)
    ELSE
      Get_Constraints_rel(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots);

  gamma:=1.0;
  maxstep:=step;
  step :=0;
  Copy1(n+k,t,tnew); { tnew := t, We change only t[per] in this procedure! }
  Compute_Smoothing_Matrix(t,n,k,r,S);
  Compute_Smoothing_Spline(m,n,k,t,r,mu,tau,y,S,alpha,RES); { compute G(t) }
  normGnew:=RES;

  REPEAT
    Compute_Smoothing_Matrix(t,n,k,r,S);
    Get_G(m,n,k,r,mu,t,tau,y,alpha,S,G);
    normG:= normGnew;

    IF Kaufman_Model THEN
      Kaufman_Smoothing_Jacobi_Matrix(m,n,k,t,r,mu,l,per,J^)
    ELSE
      Golub_Smoothing_Jacobi_Matrix(m,n,k,t,r,mu,l,per,J^,G);

    { compute the gradient J^T * G }
    FOR jj:=1 TO l DO
      BEGIN
        temp:=0.0;
        FOR i:=1 TO m+n-r DO
          temp:=temp + J^[i,jj]*G[i];
        x[jj]:=temp;
      END;
    normGradient:=nrm2(l,x,1);

    { Save the Jacobian J and the residual G for further computations }
    FOR i:=1 TO m+n-r DO
      BEGIN
        FOR jj:=1 TO l DO
          Jsave^[i,jj]:=J^[i,jj];
        Gsave[i]:=G[i];
      END;

    { restore the constraints which could have been overwritten in a previous
      step }
    FOR i:=1 TO ncstr DO
      BEGIN
        FOR jj:=1 TO l DO
          Acstr^[i,jj]:=Asave^[i,jj];
        bcstr^[i]:=bsave^[i];
      END;

    { Compute Acstr * \tilde(t) }
    FOR i:=1 TO ncstr DO
      BEGIN
        temp:=0.0;
        FOR jj:=1 TO l DO
          temp:=temp + Acstr^[i,jj] * t[ per[jj] ];
        { Now temp contains i-th component of Acstr * \tilde(t) }
        bcstr^[i]:=bcstr^[i]-temp; { build right hand side }
      END;


    FOR i:=1 TO m+n-r DO
      G[i]:=-G[i];

    { solve linear least squares problem with linear inequality constraints }
    LSI(J^,LMAX,m+n-r,l,Acstr^,LMAX,ncstr,p,G,bcstr^,workreal^,workint^,LSIError);
    { J^, Acstr^, G, and bcstr^ will be overwritten with transformed values!
      If J is nearly rank deficient then only J has been changed, all other
      vectors and matrices (esp. Acstr^ and bcstr^ have their old values.    }

    { estimate condition number of triangular factor stored in place of J^ }
    IF Abs(J^[l,l])> macheps THEN
      cond:=Condition_Estimate (J^,LMAX,l,workreal^)
    ELSE
      cond:=2.0/Sqrt(macheps);

    mode:='     GN-step';
    IF LSIError <> 1 THEN
      BEGIN
        mode:='reg. GN-step';
        IF LSIError<>5 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! ');
            GOTO return;
          END;
        { We try a regularized Gauss-Newton step. }

        { Compute || J^T * J ||_1

          | H |_1 = max_(j=1..l) \sum_(i=1)^l abs(H(i,j))
        }
        Hnorm:=0.0;
        FOR jj:=1 TO l DO
          BEGIN
            temp:=0.0;
            FOR i:=1 TO l DO
              BEGIN
                Hij:=0.0;
                FOR kk:=1 TO m+n-r DO
                  Hij:=Hij+Jsave^[kk,i] * Jsave^[kk,jj];
                temp:=temp+Abs(Hij);
              END;
            { temp contains now the sum of absolute values of the j-th column }
            IF Hnorm < temp THEN
              Hnorm:=temp;
          END; { for jj:=1 to l do }

        lambda:=Sqrt(l*macheps) * Hnorm;
        { Form new Matrix }
        FOR i:=1 TO m+n-r DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[i,jj]:=Jsave^[i,jj];
            Greg^[i]:=-Gsave[i];
          END;

        FOR i:=1 TO l DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[m+n-r+i,jj]:=0.0;
            Greg^[m+n-r+i]:=0.0;
          END;

        FOR i:=1 TO l DO
          Jreg^[m+n-r+i,i]:=Sqrt(lambda);

        {        (       J           )            (  F )
          Jreg = (                   )     Freg = (    )
                 (  sqrt(lambda) * I )            (  0 )
        }

        LSI(Jreg^,LMAX,m+n-r+l,l,Acstr^,LMAX,ncstr,p,Greg^,bcstr^,workreal^,workint^,LSIError);

        { estimate condition number of triangular factor stored in place of Jreg^ }
        IF Abs(Jreg^[l,l])> macheps THEN
          cond:=Condition_Estimate (Jreg^,LMAX,l,workreal^)
        ELSE
          cond:=2.0/Sqrt(macheps);

        IF LSIError<>1 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! Matrix has not full rank! ');
            GOTO return;
          END;
      END;  { end of computation of regul. Gauss-Newton step }


    { s[1], ... ,s[l] contains descent direction for constrained problem
      We work instead with p=-s }
    FOR i:=1 TO l DO
      p[i]:=-p[i];

    gradTp:=dot(l,x,1,p,1);   { G^T * J * p }
    IF gradTp < 0.0 THEN
      GOTO return; { -p is no longer descent direction }

    FOR jj:=1 TO l DO
        x[jj]:=t[ per[jj] ];
    normXold:= nrm2(l,x,1);

    step1:=0;

    { begin line search }
    REPEAT
      FOR jj:=1 TO l DO
        tnew[ per[jj] ]:=t[ per[jj] ] - gamma * p[jj];

      Compute_Smoothing_Matrix(tnew,n,k,r,S);
      Compute_Smoothing_Spline(m,n,k,tnew,r,mu,tau,y,S,alpha,RES);
      normGnew:= RES;

      IF (Sqr(normG)-Sqr(normGnew)) >= 2.0*delta*gamma*gradTp THEN
        Descent_Test:=true
      ELSE Descent_Test:=False;

{$ifndef mixed}
      gamma:=gamma*alph;
{$else}
      IF Descent_test=False THEN
        BEGIN { reduce step length }
          IF gamma=1.0 THEN { first backtrack, quadratic fit }
            gamma_new:=0.5*gradTp/(0.5*Sqr(normGnew)-0.5*Sqr(normG)+gradTp)
          ELSE
            BEGIN  { all subsequent backtracks, cubic fit }
              f1:=0.5*Sqr(normGnew)-0.5*Sqr(normG)+gamma*gradTp;
              f2:=0.5*Sqr(normGprev)-0.5*Sqr(normG)+gamma_prev*gradTp;
              aa:=1.0/(gamma-gamma_prev)*(f1/Sqr(gamma)-f2/Sqr(gamma_new));
              bb:=1.0/(gamma-gamma_prev)*(-gamma_prev*f1/Sqr(gamma)-gamma*f2/Sqr(gamma_prev));
              disc:=Sqr(bb)+3.0*aa*gradTp;
              IF aa = 0.0 THEN
                gamma_new:=gradTp/(2.0*bb)
              ELSE
                gamma_new:=(-bb+Sqrt(disc))/(3.0*aa);
              IF gamma_new>0.5*gamma THEN
                gamma_new:=0.5*gamma;
            END; { cubic fit }
          gamma_prev:=gamma;
          normGprev:=normGnew;
          IF gamma_new<=0.1*gamma THEN
            gamma:=0.1*gamma
          ELSE
            gamma:=gamma_new;
        END; { reduce step length }
{$endif}

      step1:=step1+1;
      normdiff:=Abs(gamma)*nrm2(l,p,1);
    UNTIL (Descent_Test=True) OR (normdiff<=EPS1*(normXold+EPS2));
    { end line search }

    FOR jj:=1 TO l DO
       t[ per[jj] ]:=tnew[ per[jj] ]; { t := tnew }

    convergent:= (normGnew<=EPS0) OR (normdiff<=EPS1*(normXold+EPS2)) OR
    (Abs(normGnew-normG)<=EPS3*normG) OR (normGradient<=EPSGrad) OR (Abs(gradTp)<=EPSDesc);

{$ifndef mixed}
    gamma:=gamma/alph;
{$endif}

    Diagnostic(step,normGnew,cond,gradTp,step1,gamma,normdiff/normXold,(normGnew-normG)/normG,mode,0);

{$ifdef mixed}
    gamma:=1.0;
{$else}
    IF gamma/alph <= 1.0 THEN gamma:= gamma/alph ELSE gamma:=1.0;
{$endif}

    step:=step+1;
  UNTIL (step>MAXSTEP) OR convergent;

  { Close file }
  Diagnostic(0,0,0,0,0,0,0,0,'',1);

return:
  step:=step-1;
  IF normGnew<=EPS0 THEN
    RetCode:=1
  ELSE
  IF normGradient<=EPSGrad THEN
    RetCode:=2
  ELSE
  IF Abs(gradTp)<=EPSDesc THEN
    RetCode:=3
  ELSE
  IF normdiff<=EPS1*(normXold+EPS2) THEN
    RetCode:=4
  ELSE
  IF Abs(normGnew-normG) <=EPS3*normG THEN
    RetCode:=5
  ELSE
  IF step >= MAXSTEP THEN
    RetCode:=6
  ELSE
    RetCode:=7;  { failed otherwise }

  ncstr:=2*l;    { maximal number of constraints to deallocate working space }

  FreeMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  FreeMem(workint,(l+ncstr)*SizeOf(workinttype));
  FreeMem(J,SizeOf(SmoothJacobiMatrix));
  FreeMem(Jsave,SizeOf(SmoothJacobiMatrix));
  FreeMem(Acstr,SizeOf(CSTRMatrix));
  FreeMem(Asave,SizeOf(CSTRMatrix));
  FreeMem(bcstr,SizeOf(CSTRVector));
  FreeMem(bsave,SizeOf(CSTRVector));
  FreeMem(Jreg,SizeOf(Reg_JacobiMatrix));
  FreeMem(Greg,(m+n-r+l)*SizeOf(Reg_G));
END; { Smoothing_Regularized_Generalized_Gauss_Newton }

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Regularized_Generalized_Gauss_Newton (
                   m,n,                              { dimension             }
                   k                 : integer;      { order of B-Splines    }
               VAR t                 : knotsequence; { contains free knots   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                   l                 : integer;      { number of free knots  }
               VAR per               : free_knots;   { pointer to free knots }
               VAR step              : integer;      { maximal no. of steps  }
               VAR RetCode           : integer;      { Return Code           }
               VAR normGradient      : real;         { Norm of Gradient      }
               VAR gradTp            : real          { J^T * F * p           }
                        );

CONST alph=0.2;      { in (0,1.0) factor to reduce step size in line search }
      delta=0.2;     { in (0,0.5) sufficient descent in line search ?       }

LABEL return;

TYPE  workrealtype     = ARRAY[1..1] OF real;
      { We need (l+1)*(2+ncstr+2*ncstr) }
      workinttype      = ARRAY[1..1] OF integer;
      { We need (l+ncstr) }
      Reg_JacobiMatrix = ARRAY[1..MMAX+LMAX,1..LMAX] OF real;
      { We need (m+l)*l }
      Reg_F            = ARRAY[1..1] OF real;
      { We need (m+l) }

VAR B                                    : LeastSquareMatrix;
    Q                                    : OrthogonalMatrix;
    cond,gamma,normF,normFnew,normdiff,
    normXold,temp,Hnorm,Hij,lambda       : real;
    F,Fsave                              : datavalues;
    tnew                                 : knotsequence;
    x,p                                  : coefficients;
    maxstep,step1,i,jj,ncstr,LSIError,kk : integer;
    Descent_Test,convergent              : Boolean;
    Size                                 : LongInt;
    workreal                             : ^workrealtype;
    workint                              : ^workinttype;
    J,JSave                              : ^JacobiMatrix;
    Acstr,Asave                          : ^CSTRMatrix;
    bcstr,bsave                          : ^CSTRVector;
    Jreg                                 : ^Reg_JacobiMatrix;
    Freg                                 : ^Reg_F;
    mode                                 : STRING[12];
{$ifdef mixed}
    gamma_new,f1,f2,aa,bb,disc,gamma_prev,normFprev  : real;
{$endif}

BEGIN
  { Output of intermediate results in File}
  { Open file }
  Diagnostic(0,0,0,0,0,0,0,0,'',-1);

  { Initialization }
  ncstr:=2*l;    { maximal number of constraints to allocate working space }
  Size:=0;
  Inc(Size,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  Inc(Size,(l+ncstr)*SizeOf(workinttype));
  Inc(Size,2*SizeOf(JacobiMatrix));
  Inc(Size,2*SizeOf(CSTRMatrix));
  Inc(Size,2*SizeOf(CSTRVector));
  Inc(Size,SizeOf(Reg_JacobiMatrix));
  Inc(Size,(m+l)*SizeOf(Reg_F));

  IF MaxAvail < Size THEN
    Error_Handler('Gauss-Newton-Method: Not enough memory!')
  ELSE
    BEGIN
      GetMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
      GetMem(workint,(l+ncstr)*SizeOf(workinttype));
      GetMem(J,SizeOf(JacobiMatrix));
      GetMem(Jsave,SizeOf(JacobiMatrix));
      GetMem(Acstr,SizeOf(CSTRMatrix));
      GetMem(Asave,SizeOf(CSTRMatrix));
      GetMem(bcstr,SizeOf(CSTRVector));
      GetMem(bsave,SizeOf(CSTRVector));
      GetMem(Jreg,SizeOf(Reg_JacobiMatrix));
      GetMem(Freg,(m+l)*SizeOf(Reg_F));
    END;

  { Get linear inequality constraints for problem }
  IF Constraints_abs
    THEN
      Get_Constraints_abs(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots)
    ELSE
      Get_Constraints_rel(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots);

  gamma:=1.0;
  maxstep:=step;
  step :=0;
  Copy1(n+k,t,tnew); { tnew := t, We change only t[per] in this procedure! }
  { compute spline coefficients alpha }
  Constrained_Spline(m,n,order,t,active_bounds,lower,upper,
    TypeOfConstraints,tau,y,alpha,RES,false,false,B,Q);
  normFnew:=RES; { compute F(t) }

  REPEAT
    { compute residual function F }
    Get_F(m,n,k,t,tau,y,alpha,F);
    normF:= normFnew;

    IF Kaufman_Model THEN
       Kaufman_Constrained_Jacobi_Matrix(m,n,k,t,
         active_bounds,lower,upper,TypeOfConstraints,l,per,J^)
    ELSE
       Golub_Constrained_Jacobi_Matrix(m,n,k,t,
          active_bounds,lower,upper,TypeOfConstraints,l,per,J^,F);

    { compute the gradient J^T * F }
    FOR jj:=1 TO l DO
      BEGIN
        temp:=0.0;
        FOR i:=1 TO m DO
          temp:=temp + J^[i,jj]*F[i];
        x[jj]:=temp;
      END;
    normGradient:=nrm2(l,x,1);

    { Save the Jacobian J and the residual F for further computations }
    FOR i:=1 TO m DO
      BEGIN
        FOR jj:=1 TO l DO
          Jsave^[i,jj]:=J^[i,jj];
        Fsave[i]:=F[i];
      END;

    { restore the constraints which could have been overwritten in a previous
      step }
    FOR i:=1 TO ncstr DO
      BEGIN
        FOR jj:=1 TO l DO
          Acstr^[i,jj]:=Asave^[i,jj];
        bcstr^[i]:=bsave^[i];
      END;

    { Compute Acstr * \tilde(t) }
    FOR i:=1 TO ncstr DO
      BEGIN
        temp:=0.0;
        FOR jj:=1 TO l DO
          temp:=temp + Acstr^[i,jj] * t[ per[jj] ];
        { Now temp contains i-th component of Acstr * \tilde(t) }
        bcstr^[i]:=bcstr^[i]-temp; { build right hand side }
      END;


    FOR i:=1 TO m DO
      F[i]:=-F[i];

    { solve linear least squares problem with linear inequality constraints }
    LSI(J^,LMAX,m,l,Acstr^,LMAX,ncstr,p,F,bcstr^,workreal^,workint^,LSIError);
    { J^, Acstr^, F, and bcstr^ will be overwritten with transformed values!
      If J is nearly rank deficient then only J has been changed, all other
      vectors and matrices (esp. Acstr^ and bcstr^ have their old values.    }

    { estimate condition number of triangular factor stored in place of J^ }
    IF Abs(J^[l,l])> macheps THEN
      cond:=Condition_Estimate (J^,LMAX,l,workreal^)
    ELSE
      cond:=2.0/Sqrt(macheps);

    mode:='     GN-step';
    IF LSIError <> 1 THEN
      BEGIN
        mode:='reg. GN-step';
        IF LSIError<>5 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! ');
            GOTO return;
          END;
        { We try a regularized Gauss-Newton step. }

        { Compute || J^T * J ||_1

          | H |_1 = max_(j=1..l) \sum_(i=1)^l abs(H(i,j))
        }
        Hnorm:=0.0;
        FOR jj:=1 TO l DO
          BEGIN
            temp:=0.0;
            FOR i:=1 TO l DO
              BEGIN
                Hij:=0.0;
                FOR kk:=1 TO m DO
                  Hij:=Hij+Jsave^[kk,i] * Jsave^[kk,jj];
                temp:=temp+Abs(Hij);
              END;
            { temp contains now the sum of absolute values of the j-th column }
            IF Hnorm < temp THEN
              Hnorm:=temp;
          END; { for jj:=1 to l do }

        lambda:=Sqrt(l*macheps) * Hnorm;
        { Form new Matrix }
        FOR i:=1 TO m DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[i,jj]:=Jsave^[i,jj];
            Freg^[i]:=-Fsave[i];
          END;

        FOR i:=1 TO l DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[m+i,jj]:=0.0;
            Freg^[m+i]:=0.0;
          END;

        FOR i:=1 TO l DO
          Jreg^[m+i,i]:=Sqrt(lambda);

        {        (       J           )            (  F )
          Jreg = (                   )     Freg = (    )
                 (  sqrt(lambda) * I )            (  0 )
        }

        LSI(Jreg^,LMAX,m+l,l,Acstr^,LMAX,ncstr,p,Freg^,bcstr^,workreal^,workint^,LSIError);

        { estimate condition number of triangular factor stored in place of Jreg^ }
        IF Abs(Jreg^[l,l])> macheps THEN
          cond:=Condition_Estimate (Jreg^,LMAX,l,workreal^)
        ELSE
          cond:=2.0/Sqrt(macheps);

        IF LSIError<>1 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! Matrix has not full rank! ');
            GOTO return;
          END;
      END;  { end of computation of regul. Gauss-Newton step }


    { s[1], ... ,s[l] contains descent direction for constrained problem
      We work instead with p=-s }
    FOR i:=1 TO l DO
      p[i]:=-p[i];

    gradTp:=dot(l,x,1,p,1);   { F^T * J * p }
    IF gradTp < 0.0 THEN
      GOTO return; { -p is no longer descent direction }

    FOR jj:=1 TO l DO
        x[jj]:=t[ per[jj] ];
    normXold:= nrm2(l,x,1);

    step1:=0;
    { begin line search }
    REPEAT
      FOR jj:=1 TO l DO
        tnew[ per[jj] ]:=t[ per[jj] ] - gamma * p[jj];

      { compute spline coefficients alpha }
      Constrained_Spline(m,n,order,tnew,active_bounds,lower,upper,
       TypeOfConstraints,tau,y,alpha,RES,false,false,B,Q);

      normFnew:= RES;

      { Armijo-Goldstein descent test }
      IF (Sqr(normF)-Sqr(normFnew)) >= 2.0*delta*gamma*gradTp THEN
        Descent_Test:=true
      ELSE Descent_Test:=False;

{$ifndef mixed}
      gamma:=gamma*alph;
{$else}
      IF Descent_test=False THEN
        BEGIN { reduce step length }
          IF gamma=1.0 THEN { first backtrack, quadratic fit }
            gamma_new:=0.5*gradTp/(0.5*Sqr(normFnew)-0.5*Sqr(normF)+gradTp)
          ELSE
            BEGIN  { all subsequent backtracks, cubic fit }
              f1:=0.5*Sqr(normFnew)-0.5*Sqr(normF)+gamma*gradTp;
              f2:=0.5*Sqr(normFprev)-0.5*Sqr(normF)+gamma_prev*gradTp;
              aa:=1.0/(gamma-gamma_prev)*(f1/Sqr(gamma)-f2/Sqr(gamma_new));
              bb:=1.0/(gamma-gamma_prev)*(-gamma_prev*f1/Sqr(gamma)-gamma*f2/Sqr(gamma_prev));
              disc:=Sqr(bb)+3.0*aa*gradTp;
              IF aa = 0.0 THEN
                gamma_new:=gradTp/(2.0*bb)
              ELSE
                gamma_new:=(-bb+Sqrt(disc))/(3.0*aa);
              IF gamma_new>0.5*gamma THEN
                gamma_new:=0.5*gamma;
            END; { cubic fit }
          gamma_prev:=gamma;
          normFprev:=normFnew;
          IF gamma_new<=0.1*gamma THEN
            gamma:=0.1*gamma
          ELSE
            gamma:=gamma_new;
        END; { reduce step length }
{$endif}

      step1:=step1+1;
      normdiff:=Abs(gamma)*nrm2(l,p,1);
    UNTIL (Descent_Test=True) OR (normdiff<=EPS1*(normXold+EPS2));
    { end line search }

    FOR jj:=1 TO l DO
       t[ per[jj] ]:=tnew[ per[jj] ]; { t := tnew }

    convergent:= (normFnew<=EPS0) OR (normdiff<=EPS1*(normXold+EPS2)) OR
    (Abs(normFnew-normF)<=EPS3*normF) OR (normGradient<=EPSGrad) OR (Abs(gradTp)<=EPSDesc);

{$ifndef mixed}
    gamma:=gamma/alph;
{$endif}

    Diagnostic(step,normFnew,cond,gradTp,step1,gamma,normdiff/normXold,(normFnew-normF)/normF,mode,0);

{$ifdef mixed}
    gamma:=1.0;
{$else}
    IF gamma/alph <= 1.0 THEN gamma:= gamma/alph ELSE gamma:=1.0;
{$endif}

    step:=step+1;
  UNTIL (step>MAXSTEP) OR convergent;

  { Close file }
  Diagnostic(0,0,0,0,0,0,0,0,'',1);

return:
  step:=step-1;
  IF normFnew<=EPS0 THEN
    RetCode:=1
  ELSE
  IF normGradient<=EPSGrad THEN
    RetCode:=2
  ELSE
  IF Abs(gradTp)<=EPSDesc THEN
    RetCode:=3
  ELSE
  IF normdiff<=EPS1*(normXold+EPS2) THEN
    RetCode:=4
  ELSE
  IF Abs(normFnew-normF) <=EPS3*normF THEN
    RetCode:=5
  ELSE
  IF step >= MAXSTEP THEN
    RetCode:=6
  ELSE
    RetCode:=7;  { failed otherwise }

  ncstr:=2*l;    { maximal number of constraints to deallocate working space }

  FreeMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  FreeMem(workint,(l+ncstr)*SizeOf(workinttype));
  FreeMem(J,SizeOf(JacobiMatrix));
  FreeMem(Jsave,SizeOf(JacobiMatrix));
  FreeMem(Acstr,SizeOf(CSTRMatrix));
  FreeMem(Asave,SizeOf(CSTRMatrix));
  FreeMem(bcstr,SizeOf(CSTRVector));
  FreeMem(bsave,SizeOf(CSTRVector));
  FreeMem(Jreg,SizeOf(Reg_JacobiMatrix));
  FreeMem(Freg,(m+l)*SizeOf(Reg_F));
END; { Constrained_Regularized_Generalized_Gauss_Newton }

{-----------------------------------------------------------------------------}

PROCEDURE Constrained_Smoothing_Regularized_Generalized_Gauss_Newton (
                   m,n,                              { dimension             }
                   k                 : integer;      { order of B-Splines    }
               VAR t                 : knotsequence; { contains free knots   }
                   r                 : integer;      { order of derivative   }
                   mu                : real;         { Smoothing parameter   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                   l                 : integer;      { number of free knots  }
               VAR per               : free_knots;   { pointer to free knots }
               VAR step              : integer;      { maximal no. of steps  }
               VAR RetCode           : integer;      { Return Code           }
               VAR normGradient      : real;         { Norm of Gradient      }
               VAR gradTp            : real          { J^T * F * p           }
                        );

CONST alph=0.2;      { in (0,1.0) factor to reduce step size in line search }
      delta=0.2;     { in (0,0.5) sufficient descent in line search ?       }

LABEL return;

TYPE  workrealtype     = ARRAY[1..1] OF real;
      { We need (l+1)*(2+ncstr+2*ncstr) }
      workinttype      = ARRAY[1..1] OF integer;
      { We need (l+ncstr) }
      Reg_JacobiMatrix = ARRAY[1..MMAX+NMAX+LMAX,1..LMAX] OF real;
      { We need (m+n-r+l)*l }
      Reg_G            = ARRAY[1..1] OF real;
      { We need (m+n-r+l) }

VAR cond,gamma,normG,normGnew,normdiff,
    normXold,temp,Hnorm,Hij,lambda       : real;
    G,Gsave                              : SmoothingVector;
    tnew                                 : knotsequence;
    x,p                                  : coefficients;
    maxstep,step1,i,jj,ncstr,LSIError,kk : integer;
    Descent_Test,convergent              : Boolean;
    Size                                 : LongInt;
    workreal                             : ^workrealtype;
    workint                              : ^workinttype;
    J,JSave                              : ^SmoothJacobiMatrix;
    Acstr,Asave                          : ^CSTRMatrix;
    bcstr,bsave                          : ^CSTRVector;
    Jreg                                 : ^Reg_JacobiMatrix;
    Greg                                 : ^Reg_G;
    B                                    : LeastSquareMatrix;
    Q                                    : OrthogonalMatrix;
    S                                    : SmoothingMatrix;
    mode                                 : STRING[12];

{$ifdef mixed}
    gamma_new,f1,f2,aa,bb,disc,gamma_prev,normGprev  : real;
{$endif}

BEGIN
  { Output of intermediate results in File}
  { Open file }
  Diagnostic(0,0,0,0,0,0,0,0,'',-1);

  { Initialization }
  ncstr:=2*l;    { maximal number of constraints to allocate working space }
  Size:=0;
  Inc(Size,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  Inc(Size,(l+ncstr)*SizeOf(workinttype));
  Inc(Size,2*SizeOf(SmoothJacobiMatrix));
  Inc(Size,2*SizeOf(CSTRMatrix));
  Inc(Size,2*SizeOf(CSTRVector));
  Inc(Size,SizeOf(Reg_JacobiMatrix));
  Inc(Size,(m+n-r+l)*SizeOf(Reg_G));

  IF MaxAvail < Size THEN
    Error_Handler('Gauss-Newton-Method: Not enough memory!')
  ELSE
    BEGIN
      GetMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
      GetMem(workint,(l+ncstr)*SizeOf(workinttype));
      GetMem(J,SizeOf(SmoothJacobiMatrix));
      GetMem(Jsave,SizeOf(SmoothJacobiMatrix));
      GetMem(Acstr,SizeOf(CSTRMatrix));
      GetMem(Asave,SizeOf(CSTRMatrix));
      GetMem(bcstr,SizeOf(CSTRVector));
      GetMem(bsave,SizeOf(CSTRVector));
      GetMem(Jreg,SizeOf(Reg_JacobiMatrix));
      GetMem(Greg,(m+n-r+l)*SizeOf(Reg_G));
    END;

  { Get linear inequality constraints for problem }
  IF Constraints_abs
    THEN
      Get_Constraints_abs(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots)
    ELSE
      Get_Constraints_rel(Asave^,bsave^,ncstr,l,t,per,Min_Dist_Knots);

  gamma:=1.0;
  maxstep:=step;
  step :=0;
  Copy1(n+k,t,tnew); { tnew := t, We change only t[per] in this procedure! }
  Compute_Smoothing_Matrix(t,n,k,r,S);
  { compute spline coefficients alpha }
  Constrained_Smoothing_Spline(m,n,order,t,r,mu,
       active_bounds,lower,upper,TypeOfConstraints,tau,y,alpha,S,RES,false,false,B,Q);
  normGnew:=RES; { compute G(t) }

  REPEAT

    Compute_Smoothing_Matrix(t,n,k,r,S);

    { compute residual function G }
    Get_G(m,n,k,r,mu,t,tau,y,alpha,S,G);
    normG:= normGnew;

    IF Kaufman_Model THEN
       Kaufman_Constrained_Smoothing_Jacobi_Matrix(m,n,k,t,r,mu,
         active_bounds,lower,upper,TypeOfConstraints,l,per,J^)
    ELSE
      Golub_Constrained_Smoothing_Jacobi_Matrix(m,n,k,t,r,mu,
          active_bounds,lower,upper,TypeOfConstraints,l,per,J^,G);

    { compute the gradient J^T * G }
    FOR jj:=1 TO l DO
      BEGIN
        temp:=0.0;
        FOR i:=1 TO m+n-r DO
          temp:=temp + J^[i,jj]*G[i];
        x[jj]:=temp;
      END;
    normGradient:=nrm2(l,x,1);

    { Save the Jacobian J and the residual G for further computations }
    FOR i:=1 TO m+n-r DO
      BEGIN
        FOR jj:=1 TO l DO
          Jsave^[i,jj]:=J^[i,jj];
        Gsave[i]:=G[i];
      END;

    { restore the constraints which could have been overwritten in a previous
      step }
    FOR i:=1 TO ncstr DO
      BEGIN
        FOR jj:=1 TO l DO
          Acstr^[i,jj]:=Asave^[i,jj];
        bcstr^[i]:=bsave^[i];
      END;

    { Compute Acstr * \tilde(t) }
    FOR i:=1 TO ncstr DO
      BEGIN
        temp:=0.0;
        FOR jj:=1 TO l DO
          temp:=temp + Acstr^[i,jj] * t[ per[jj] ];
        { Now temp contains i-th component of Acstr * \tilde(t) }
        bcstr^[i]:=bcstr^[i]-temp; { build right hand side }
      END;


    FOR i:=1 TO m+n-r DO
      G[i]:=-G[i];

    { solve linear least squares problem with linear inequality constraints }
    LSI(J^,LMAX,m+n-r,l,Acstr^,LMAX,ncstr,p,G,bcstr^,workreal^,workint^,LSIError);
    { J^, Acstr^, G, and bcstr^ will be overwritten with transformed values!
      If J is nearly rank deficient then only J has been changed, all other
      vectors and matrices (esp. Acstr^ and bcstr^ have their old values.    }

    { estimate condition number of triangular factor stored in place of J^ }
    IF Abs(J^[l,l])> macheps THEN
      cond:=Condition_Estimate (J^,LMAX,l,workreal^)
    ELSE
      cond:=2.0/Sqrt(macheps);

    mode:='     GN-step';
    IF LSIError <> 1 THEN
      BEGIN
        mode:='reg. GN-step';
        IF LSIError<>5 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! ');
            GOTO return;
          END;
        { We try a regularized Gauss-Newton step. }

        { Compute || J^T * J ||_1

          | H |_1 = max_(j=1..l) \sum_(i=1)^l abs(H(i,j))
        }
        Hnorm:=0.0;
        FOR jj:=1 TO l DO
          BEGIN
            temp:=0.0;
            FOR i:=1 TO l DO
              BEGIN
                Hij:=0.0;
                FOR kk:=1 TO m+n-r DO
                  Hij:=Hij+Jsave^[kk,i] * Jsave^[kk,jj];
                temp:=temp+Abs(Hij);
              END;
            { temp contains now the sum of absolute values of the j-th column }
            IF Hnorm < temp THEN
              Hnorm:=temp;
          END; { for jj:=1 to l do }

        lambda:=Sqrt(l*macheps) * Hnorm;
        { Form new Matrix }
        FOR i:=1 TO m+n-r DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[i,jj]:=Jsave^[i,jj];
            Greg^[i]:=-Gsave[i];
          END;

        FOR i:=1 TO l DO
          BEGIN
            FOR jj:=1 TO l DO
              Jreg^[m+n-r+i,jj]:=0.0;
            Greg^[m+n-r+i]:=0.0;
          END;

        FOR i:=1 TO l DO
          Jreg^[m+n-r+i,i]:=Sqrt(lambda);

        {        (       J           )            (  F )
          Jreg = (                   )     Freg = (    )
                 (  sqrt(lambda) * I )            (  0 )
        }

        LSI(Jreg^,LMAX,m+n-r+l,l,Acstr^,LMAX,ncstr,p,Greg^,bcstr^,workreal^,workint^,LSIError);

        { estimate condition number of triangular factor stored in place of Jreg^ }
        IF Abs(Jreg^[l,l])> macheps THEN
          cond:=Condition_Estimate (Jreg^,LMAX,l,workreal^)
        ELSE
          cond:=2.0/Sqrt(macheps);

        IF LSIError<>1 THEN
          BEGIN
            GotoXY(20,5);
            WriteLn(Chr(7)+' Error in LSI! Matrix has not full rank! ');
            GOTO return;
          END;
      END;  { end of computation of regul. Gauss-Newton step }


    { s[1], ... ,s[l] contains descent direction for constrained problem
      We work instead with p=-s }
    FOR i:=1 TO l DO
      p[i]:=-p[i];

    gradTp:=dot(l,x,1,p,1);   { G^T * J * p }
    IF gradTp < 0.0 THEN
      GOTO return; { -p is no longer descent direction }

    FOR jj:=1 TO l DO
        x[jj]:=t[ per[jj] ];
    normXold:= nrm2(l,x,1);

    step1:=0;
    { begin line search }
    REPEAT
      FOR jj:=1 TO l DO
        tnew[ per[jj] ]:=t[ per[jj] ] - gamma * p[jj];

      Compute_Smoothing_Matrix(tnew,n,k,r,S);
      { compute spline coefficients alpha }
      Constrained_Smoothing_Spline(m,n,order,tnew,r,mu,
           active_bounds,lower,upper,TypeOfConstraints,tau,y,alpha,S,RES,false,false,B,Q);

      normGnew:= RES;

      { Armijo-Goldstein descent test }
      IF (Sqr(normG)-Sqr(normGnew)) >= 2.0*delta*gamma*gradTp THEN
        Descent_Test:=true
      ELSE Descent_Test:=False;

{$ifndef mixed}
      gamma:=gamma*alph;
{$else}
      IF Descent_test=False THEN
        BEGIN { reduce step length }
          IF gamma=1.0 THEN { first backtrack, quadratic fit }
            gamma_new:=0.5*gradTp/(0.5*Sqr(normGnew)-0.5*Sqr(normG)+gradTp)
          ELSE
            BEGIN  { all subsequent backtracks, cubic fit }
              f1:=0.5*Sqr(normGnew)-0.5*Sqr(normG)+gamma*gradTp;
              f2:=0.5*Sqr(normGprev)-0.5*Sqr(normG)+gamma_prev*gradTp;
              aa:=1.0/(gamma-gamma_prev)*(f1/Sqr(gamma)-f2/Sqr(gamma_new));
              bb:=1.0/(gamma-gamma_prev)*(-gamma_prev*f1/Sqr(gamma)-gamma*f2/Sqr(gamma_prev));
              disc:=Sqr(bb)+3.0*aa*gradTp;
              IF aa = 0.0 THEN
                gamma_new:=gradTp/(2.0*bb)
              ELSE
                gamma_new:=(-bb+Sqrt(disc))/(3.0*aa);
              IF gamma_new>0.5*gamma THEN
                gamma_new:=0.5*gamma;
            END; { cubic fit }
          gamma_prev:=gamma;
          normGprev:=normGnew;
          IF gamma_new<=0.1*gamma THEN
            gamma:=0.1*gamma
          ELSE
            gamma:=gamma_new;
        END; { reduce step length }
{$endif}

      step1:=step1+1;
      normdiff:=Abs(gamma)*nrm2(l,p,1);
    UNTIL (Descent_Test=True) OR (normdiff<=EPS1*(normXold+EPS2));
    { end line search }

    FOR jj:=1 TO l DO
       t[ per[jj] ]:=tnew[ per[jj] ]; { t := tnew }

    convergent:= (normGnew<=EPS0) OR (normdiff<=EPS1*(normXold+EPS2)) OR
    (Abs(normGnew-normG)<=EPS3*normG) OR (normGradient<=EPSGrad) OR (Abs(gradTp)<=EPSDesc);

{$ifndef mixed}
    gamma:=gamma/alph;
{$endif}

    Diagnostic(step,normGnew,cond,gradTp,step1,gamma,normdiff/normXold,(normGnew-normG)/normG,mode,0);

{$ifdef mixed}
    gamma:=1.0;
{$else}
    IF gamma/alph <= 1.0 THEN gamma:= gamma/alph ELSE gamma:=1.0;
{$endif}

    step:=step+1;
  UNTIL (step>MAXSTEP) OR convergent;

  { Close file }
  Diagnostic(0,0,0,0,0,0,0,0,'',1);

return:
  step:=step-1;
  IF normGnew<=EPS0 THEN
    RetCode:=1
  ELSE
  IF normGradient<=EPSGrad THEN
    RetCode:=2
  ELSE
  IF Abs(gradTp)<=EPSDesc THEN
    RetCode:=3
  ELSE
  IF normdiff<=EPS1*(normXold+EPS2) THEN
    RetCode:=4
  ELSE
  IF Abs(normGnew-normG) <=EPS3*normG THEN
    RetCode:=5
  ELSE
  IF step >= MAXSTEP THEN
    RetCode:=6
  ELSE
    RetCode:=7;  { failed otherwise }

  ncstr:=2*l;    { maximal number of constraints to deallocate working space }

  FreeMem(workreal,(l+1)*(2+ncstr+2*ncstr)*SizeOf(workrealtype));
  FreeMem(workint,(l+ncstr)*SizeOf(workinttype));
  FreeMem(J,SizeOf(SmoothJacobiMatrix));
  FreeMem(Jsave,SizeOf(SmoothJacobiMatrix));
  FreeMem(Acstr,SizeOf(CSTRMatrix));
  FreeMem(Asave,SizeOf(CSTRMatrix));
  FreeMem(bcstr,SizeOf(CSTRVector));
  FreeMem(bsave,SizeOf(CSTRVector));
  FreeMem(Jreg,SizeOf(Reg_JacobiMatrix));
  FreeMem(Greg,(m+n-r+l)*SizeOf(Reg_G));
END; { Constrained_Smoothing_Regularized_Generalized_Gauss_Newton }

{-----------------------------------------------------------------------------}

END.