UNIT Jacobi;
{ Unit to provide the Jacobi Matrix for Gauss Newton methods                  }

INTERFACE

USES def,         { definitions of constants, types and global variables      }
     error,       { Error handler                                             }
     BLAS,        { Basic Linear Algebra Subroutines                          }
     math,        { Low level mathematical subroutines                        }
     smooth,      { procedures to compute smoothing B-Splines                 }
     BSpline,     { procedures to compute B-Splines and l2-approximations     }
     CSSpline,    { procedures for constrained smoothing B-Splines            }
     linpack;     { Routines from LINPACK for QR-factorization of a matrix    }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix;
                             VAR F   : datavalues
                             );

{ computes an approximation of the Jacobimatrix J \approx F'(t)
                                 kk
                     F( t + h * e  ) - F( t )
  kk-th column of J:=--------------------------
                               h

  where F(t) = y - B(t) * alpha

  alpha solves the least squares problem

  0.5 || y - B(t) * alpha ||^2 ---> min
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Jacobi_Matrix(
                             m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix
                             );

{ computes an approximation of the Jacobimatrix J \approx F'(t) via the
  Kaufman-Model

                                 +                   +
  kk-th column of J:= - [ I -B B   ] ( B' [ e^k] ) B   y         kk=1,..,l

  where F(t) = y - B(t) * alpha

  alpha solves the least squares problem

  0.5 || y - B(t) * alpha ||^2 ---> min
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter   }
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix;
                             VAR G   : SmoothingVector
                             );

{ computes an approximation of the Jacobimatrix J \approx G'(t)
                                 kk
                     G( t + h * e  ) - G( t )
  kk-th column of J:=--------------------------
                               h

               [ y ]    [             B(t) ]
  where G(t) = [   ]  - [                  ] * alpha
               [ 0 ]    [ \mu^(1/2) * S(t) ]

  alpha solves the smoothing problem

  0.5 || y - B(t) * alpha ||^2 + \mu * 0.5 || S(t) * alpha ||^2 ---> min
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter }
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix
                             );
{ computes an approximation of the Jacobimatrix Jsmooth \approx G'(t) for the
  smoothing problem via the Kaufman-Model

  kk-th column of Jsmooth
                                        +                                 +
  [( I )   (          B) (          B) ] (          B')      (          B) ( y )
- [(   ) - (           ) (           ) ] (            )[e^kk](           ) (   )
  [( I )   (\mu^(1/2) S) (\mu^(1/2) S) ] (\mu^(1/2) S')      (\mu^(1/2) S) ( 0 )

  for kk=1,..,l

               [ y ]    [             B(t) ]
  where G(t) = [   ]  - [                  ] * alpha
               [ 0 ]    [ \mu^(1/2) * S(t) ]

  alpha solves the smoothing problem

  0.5 || y - B(t) * alpha ||^2 + \mu * 0.5 || S(t) * alpha ||^2 ---> min
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Constrained_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix;
                             VAR F   : datavalues
                             );

{ computes an approximation of the Jacobimatrix J \approx F'(t)
                                 kk
                     F( t + h * e  ) - F( t )
  kk-th column of J:=--------------------------
                               h

  where F(t) = y - B(t) * alpha(t)

  alpha(t) solves the constrained problem

  0.5 || y - B(t) * alpha ||^2  ---> min

  s.t.           L <= D_p(t) * alpha <= U
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Constrained_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix;
                             VAR G   : SmoothingVector
                             );

{ computes an approximation of the Jacobimatrix J \approx G'(t)
                                 kk
                     G( t + h * e  ) - G( t )
  kk-th column of J:=--------------------------
                               h

               [ y ]    [             B(t) ]
  where G(t) = [   ]  - [                  ] * alpha(t)
               [ 0 ]    [ \mu^(1/2) * S(t) ]

  alpha(t) solves the constrained smoothing problem

  0.5 || y - B(t) * alpha ||^2 + \mu * 0.5 || S(t) * alpha ||^2 ---> min

  s.t.           L <= D_p(t) * alpha <= U
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Constrained_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix
                             );

{ computes an approximation of the Jacobimatrix J \approx F'(t) for the
  constrained approximation problem via the Kaufman-Model

  kk-th column of J

  J e^kk := ( I - P_(BN) ) ( J_\ttilde e^kk + B \bar(R)^+ \bar(Gamma) e^kk )

  where F(t) = y - B(t) * alpha(t)

  alpha(t) solves the constrained problem

  0.5 || y - B(t) * alpha ||^2 ---> min

  s.t.           L <= D_p(t) * alpha <= U
                                                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Constrained_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix
                             );

{ computes an approximation of the Jacobimatrix J \approx G'(t) for the
  constrained smoothing problem via the Kaufman-Model

  kk-th column of J

  J e^kk := ( I - P_(BN) ) ( J_\ttilde e^kk + B \bar(R)^+ \bar(Gamma) e^kk )

               [ y ]    [             B(t) ]
  where G(t) = [   ]  - [                  ] * alpha(t)
               [ 0 ]    [ \mu^(1/2) * S(t) ]

  alpha(t) solves the constrained smoothing problem

  0.5 || y - B(t) * alpha ||^2 + \mu * 0.5 || S(t) * alpha ||^2 ---> min

  s.t.           L <= D_p(t) * alpha <= U
                                                                              }

{-----------------------------------------------------------------------------}

IMPLEMENTATION

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix;
                             VAR F   : datavalues
                             );
{ computes an approximation of the Jacobimatrix                               }

VAR Fnew      : datavalues;
    h,temp    : real;
    i,kk      : integer;

BEGIN
  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      h:=EPSDiff1*(Abs(t[ per[kk] ]) + EPSDiff2);
      temp:=t[per[kk]]+h;
      h:=temp-t[per[kk]];
      { trick to improve the accuracy of finite difference approximations }

      temp:=t[per[kk]];
      t[ per[kk] ]:= t[ per[kk] ]+h;
      l2Appr(m,n,k,t,tau,y,alpha,Res);
      {                                                       kk
       compute least squares solution alpha to knots t + h * e   }
      Get_F(m,n,k,t,tau,y,alpha,Fnew);

      FOR i:=1 TO m DO
        J[ i, kk ] := (Fnew[i] - F[i])/h;

      { reconstruction of t }
      t[ per[kk] ]:=temp;
    END;
END; { Golub_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Jacobi_Matrix(
                             m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix
                             );
{ computes an approximation of the Jacobimatrix via the Kaufman-Model         }

VAR Q           : OrthogonalMatrix;
    B           : LeastSquareMatrix;
    F           : datavalues;
    Jcolumn     : datavalues;
    x,alphatemp : coefficients;
    i,kk        : integer;
BEGIN
  { QR-factorization of B }
  GIVENS_factorization(m,n,k,t,tau,B,Q);
  apply_GIVENS_transformation(m,n,k,t,tau,y,Q,x,RES);
  Rc_x(n,k,B,alpha,x);

  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      Derivative_Spline_Knots(m,n,k,t,tau,alpha,per[kk],F);
      {                     +
         F = ( B' [e^k] ) B   y
      }

      apply_GIVENS_transformation(m,n,k,t,tau,F,Q,x,RES);

      Rc_x(n,k,B,alphatemp,x);
      {               +                +
        alphatemp = B   ( B' [e^k] ) B   y
      }

      Get_F (m,n,k,t,tau,F,alphatemp,Jcolumn);

      { save kk-th column of J }
      FOR i:=1 TO m DO
        J[i,kk]:=-Jcolumn[i];
    END;
END; { Kaufman_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter   }
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix;
                             VAR G   : SmoothingVector
                             );

{ computes an approximation of the Jacobimatrix                               }

VAR Gnew      : SmoothingVector;
    Snew      : SmoothingMatrix;
    h,temp    : real;
    i,kk      : integer;

BEGIN
  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      h:=EPSDiff1*(Abs(t[ per[kk] ]) + EPSDiff2);
      temp:=t[per[kk]]+h;
      h:=temp-t[per[kk]];
      { trick to improve the accuracy of finite difference approximations }

      temp:=t[per[kk]];
      t[ per[kk] ]:= t[ per[kk] ]+h;
      Compute_Smoothing_Matrix(t,n,k,r,Snew);
      Compute_Smoothing_spline(m,n,k,t,r,mu,tau,y,Snew,alpha,RES);
      Get_G(m,n,k,r,mu,t,tau,y,alpha,Snew,Gnew);

      FOR i:=1 TO m+n-r DO
        Jsmooth[ i, kk ] := (Gnew[i] - G[i])/h;

      { reconstruction of t }
      t[ per[kk] ]:=temp;
    END;
END; { Golub_Smoothing_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter }
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix
                             );

VAR Q         : OrthogonalMatrix;
    B         : LeastSquareMatrix;
    S,Stemp   : SmoothingMatrix;
    G         : Smoothingvector;
    c,d       : coefficients;
    G1        : datavalues;
    G2        : coefficients;
    Rmu       : LeastSquareMatrix;
    temp      : real;
    i,j,kk    : integer;

BEGIN
  { QR-factorization of LeastSquareMatrix B }
  GIVENS_factorization(m,n,k,t,tau,B,Q);
  { B contains the upper triangular matrix R of the QR-factorization in
  packed form }

  Compute_Smoothing_Matrix(t,n,k,r,S);

  { apply the same transformation to the righthandside vector y }
  apply_GIVENS_transformation(m,n,k,t,tau,y,Q,c,RES);
  { c contains now Q_1^T y }


  { GIVENS factorization of
  [       R_B     ]          [ c ]
  [               ] \alpha - [   ]
  [  \sqrt(\mu) S ]          [ 0 ]

  On Output, Rmu contains the upper triangular matrix of the QR
  factorization and d the transformed right hand side
  }
  FOR i:=1 TO n-r DO
    G2[i]:=0.0;
  GIVENS_factorization_of_RBS(n,k,r,mu,B,Rmu,S,c,G2,d,RES);

  { Solve upper triangular system Rmu * alpha = d }
  Rc_x(n,k,Rmu,alpha,d);

  { alpha contains now
                 +
    (          B) ( y )
    (           ) (   )
    (\mu^(1/2) S) ( 0 )                                                       }

  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      Derivative_Spline_Knots(m,n,k,t,tau,alpha,per[kk],G1);
      {
         G1 = ( B' [e^k] )   alpha
      }

      Derivative_Smoothing_Matrix_Knots(n,k,t,r,alpha,per[kk],G2);
      {
         G2 = ( S' [e^k] )  alpha
      }

      { multiplication with sqrt(mu) }
      FOR i:=1 TO n-r DO
        G2[i]:=Sqrt(mu)*G2[i];

      { Now                                        +
        ( G1 )     (          B')     (          B) ( y )
        (    )   = (            )[e^k](           ) (   )
        ( G2 )     (\mu^(1/2) S')     (\mu^(1/2) S) ( 0 )   has been computed }

      apply_GIVENS_transformation(m,n,k,t,tau,G1,Q,c,RES);

      { We must apply the Givens_factorization_of_RBS to the right hand side
        (c,G2)^T in contrast to the "normal" case with the right hand side
        (c=Q_1^Ty , 0)^T
        ==> Givens_factorization_of_RBS                                       }

      GIVENS_factorization_of_RBS(n,k,r,mu,B,Rmu,S,c,G2,d,RES);
      Rc_x(n,k,Rmu,c,d);

      { c contains now
                     +                                +
        (          B) (          B')     (          B) ( y )
        (           ) (            )[e^k](           ) (   )
        (\mu^(1/2) S) (\mu^(1/2) S')     (\mu^(1/2) S) ( 0 )                  }


      FOR i:= 1 TO m DO
        BEGIN
          interval(tau[i],n+1,t,left,out_of_interval);
          G[i]:= G1[i] - Spline_Value(c,t,tau[i],left,k)
        END;

      FOR i:=1 TO n-r DO
        BEGIN
          temp:=0.0;
          FOR j:=1 TO r+1 DO
            temp:=temp+S[i,j] * c[i+j-1];
          G[m+i]:=G2[i]-Sqrt(mu)*temp;
        END;

      { Assign the kk-th column }
      FOR i:=1 TO m+n-r DO
        Jsmooth[i,kk]:=-G[i];
    END;
END; { Kaufman_Smoothing_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Constrained_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix;
                             VAR F   : datavalues
                             );

{ computes an approximation of the Jacobimatrix                               }

VAR B         : LeastSquareMatrix;
    Q         : OrthogonalMatrix;
    Fnew      : datavalues;
    h,temp    : real;
    i,kk      : integer;

BEGIN
  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      h:=EPSDiff1*(Abs(t[ per[kk] ]) + EPSDiff2);
      temp:=t[per[kk]]+h;
      h:=temp-t[per[kk]];
      { trick to improve the accuracy of finite difference approximations }

      temp:=t[per[kk]];
      t[ per[kk] ]:= t[ per[kk] ]+h;

      { compute spline coefficients alpha }
      Constrained_Spline(m,n,order,t,active_bounds,lower,upper,
       TypeOfConstraints,tau,y,alpha,RES,true,false,B,Q);

      { compute residual function F }
      Get_F(m,n,k,t,tau,y,alpha,Fnew);

      FOR i:=1 TO m DO
        J[ i, kk ] := (Fnew[i] - F[i])/h;

      { reconstruction of t }
      t[ per[kk] ]:=temp;
    END;
END; { Golub_Constrained_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Golub_Constrained_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix;
                             VAR G   : SmoothingVector
                             );

{ computes an approximation of the Jacobimatrix                               }

VAR Gnew      : SmoothingVector;
    Snew      : SmoothingMatrix;
    B         : LeastSquareMatrix;
    Q         : OrthogonalMatrix;
    h,temp    : real;
    i,kk      : integer;

BEGIN
  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      h:=EPSDiff1*(Abs(t[ per[kk] ]) + EPSDiff2);
      temp:=t[per[kk]]+h;
      h:=temp-t[per[kk]];
      { trick to improve the accuracy of finite difference approximations }

      temp:=t[per[kk]];
      t[ per[kk] ]:= t[ per[kk] ]+h;

      Compute_Smoothing_Matrix(t,n,k,r,Snew);
      { compute spline coefficients alpha }
      Constrained_Smoothing_Spline(m,n,order,t,r,mu,
           active_bounds,lower,upper,TypeOfConstraints,tau,y,alpha,Snew,RES,true,false,B,Q);

      { compute residual function G }
      Get_G(m,n,k,r,mu,t,tau,y,alpha,Snew,Gnew);

      FOR i:=1 TO m+n-r DO
        Jsmooth[ i, kk ] := (Gnew[i] - G[i])/h;

      { reconstruction of t }
      t[ per[kk] ]:=temp;
    END;
END; { Golub_Constrained_Smoothing_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Gradients_of_Subproblem_A(
                                    n,
                                    k,
                                    p : integer;
                                VAR t : knotsequence;
                                VAR Active_Bounds : BLSConstr;
                                VAR D : SmoothingMatrix;
                                VAR nactiv : integer;
                                VAR RBar : MatrixNxN);
{ compute negative gradients of active constraints to subproblem (A) at
  knot sequence t

           [   D_p ]
  Rbar = - [       ]
           [ - D_p ]_(i \in \cal(I))

  On Input:       n         number of B-splines
                  k         order of B-splines
                  p         order of derivative in shape constraints
                  t         knot sequence
             Active_Bounds  information to identify the active constraints
                  D         derivative matrix D_p

  On Output:   nactiv       number of active constraints
               Rbar         negative gradients of active constraints
                                                                              }
VAR i,j : integer;
BEGIN
  { initialization }
  FOR i:=1 TO n DO
    FOR j:=1 TO n DO
      Rbar[i,j]:=0.0;

  nactiv:=0;

  FOR i:=1 TO n-p DO
    BEGIN
      WITH active_bounds[i] DO
        IF (bound=true) AND (lower_bound=true) THEN
          BEGIN                    { get active constraints at lower bound }
            Inc(nactiv);
            FOR j:=1 TO p+1 DO
              Rbar[nactiv,i+j-1]:=-D[i,j];
          END;
    END;
  FOR i:=1 TO n-p DO
    BEGIN
      WITH active_bounds[i] DO
        IF (bound=true) AND (upper_bound=true) THEN
          BEGIN                    { get active constraints at upper bound }
            Inc(nactiv);
            FOR j:=1 TO p+1 DO
              Rbar[nactiv,i+j-1]:=D[i,j];
          END;
    END;
END; { Compute_Gradients_of_Subproblem_A }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Gradients_of_Subproblem_A_approx(
                                    n,
                                    k,
                                    p : integer;
                                VAR t : knotsequence;
                                VAR alpha : coefficients;
                                  kk  : integer;
                                VAR Active_Bounds : BLSConstr;
                                VAR D : SmoothingMatrix;
                                    nactiv : integer;
                                VAR F  : coefficients);
{ compute derivative of negative gradients of active constraints to
  subproblem (A) at knot sequence t with respect to t[kk] * alpha

                             [   D_p ] '
  F:=Rbar' [e^kk] alpha  = - [       ]                  [ e^kk] * alpha
                             [ - D_p ]_(i \in \cal(I))

  On Input:       n         number of B-splines
                  k         order of B-splines
                  p         order of derivative in shape constraints
                  t         knot sequence
                  alpha     coefficients
                  kk        index of knot to differentiate
             Active_Bounds  information to identify the active constraints
                  D         derivative matrix D_p
               nactiv       number of active constraints
 On Output:
               F
                                                                              }
VAR i     : integer;
    Ftemp : coefficients;

BEGIN
  Derivative_Derivative_Matrix_Knots_approx(n,k,t,p,alpha,kk,D,Ftemp);

  { initialization }
  nactiv:=0;

  FOR i:=1 TO n-p DO
    BEGIN
      WITH active_bounds[i] DO
        IF (bound=true) AND (lower_bound=true) THEN
          BEGIN                    { get active constraints at lower bound }
            Inc(nactiv);
            F[nactiv]:=-Ftemp[i];
          END;
    END;

  FOR i:=1 TO n-p DO
    BEGIN
      WITH active_bounds[i] DO
        IF (bound=true) AND (upper_bound=true) THEN
          BEGIN                    { get active constraints at upper bound }
            Inc(nactiv);
            F[nactiv]:=Ftemp[i];
          END;
    END;
END; { Derivatives_Gradients_of_Subproblem_A_approx }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Gradients_of_Subproblem_A(
                                    n,
                                    k,
                                    p : integer;
                                VAR t : knotsequence;
                                VAR alpha : coefficients;
                                  kk  : integer;
                                VAR Active_Bounds : BLSConstr;
                                    nactiv : integer;
                                VAR F  : coefficients);
{ compute derivative of negative gradients of active constraints to
  subproblem (A) at knot sequence t with respect to t[kk] * alpha

                             [   D_p ] '
  F:=Rbar' [e^kk] alpha  = - [       ]                  [ e^kk] * alpha
                             [ - D_p ]_(i \in \cal(I))

  On Input:       n         number of B-splines
                  k         order of B-splines
                  p         order of derivative in shape constraints
                  t         knot sequence
                  alpha     coefficients
                  kk        index of knot to differentiate
             Active_Bounds  information to identify the active constraints
               nactiv       number of active constraints
 On Output:
               F
                                                                              }
VAR i     : integer;
    Ftemp : coefficients;

BEGIN
  Derivative_Derivative_Matrix_Knots(n,k,t,p,alpha,kk,Ftemp);

  { initialization }
  nactiv:=0;

  FOR i:=1 TO n-p DO
    BEGIN
      WITH active_bounds[i] DO
        IF (bound=true) AND (lower_bound=true) THEN
          BEGIN                    { get active constraints at lower bound }
            Inc(nactiv);
            F[nactiv]:=-Ftemp[i];
          END;
    END;

  FOR i:=1 TO n-p DO
    BEGIN
      WITH active_bounds[i] DO
        IF (bound=true) AND (upper_bound=true) THEN
          BEGIN                    { get active constraints at upper bound }
            Inc(nactiv);
            F[nactiv]:=Ftemp[i];
          END;
    END;
END; { Derivatives_Gradients_of_Subproblem_A }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Constrained_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR J   : Jacobimatrix
                             );

{ computes an approximation of the Jacobimatrix                               }

TYPE qrauxtype         = ARRAY[1..1] OF real;
      { We need nactiv for RbarT and n-nactiv for R0N --> n-nactiv }
     rhstype           = ARRAY[1..1] OF real;
      { We need n }
     qytype            = ARRAY[1..1] OF real;
      { We need n }


VAR Q0         : OrthogonalMatrix;
    B          : LeastSquareMatrix;
    D          : SmoothingMatrix; { matrix of shape constraints }

    p,                             { order of derivative in shape constriants }
    nactiv,                        { number of active constraints             }
    result,                        { return code of QRSL }
    i,jj,kk,s,job   : integer;

    Rbar       : ^MatrixNxN;    { negative gradients of active constraints }
    RbarT      : ^MatrixNxN;    { contains Rbar^T }
    Nmatrix    : ^MatrixNxN;    { null space basis }
    R0N        : ^MatrixNxN;    { R0 * N }
    v1,v5      : datavalues;    { temporary }
    v2,v4      : coefficients;  { temporary }
    Size       : LongInt;
    qraux      : ^qrauxtype;
    rhs        : ^rhstype;
    qy         : ^qytype;
    dummy,temp : real;
BEGIN
  CASE TypeOfConstraints OF
    'P','0': p:=0;
    'M','1': p:=1;
    'C','2': p:=2;
  END;

  { allocate temporary working space }
  Size:=3*n*SizeOf(Real)+4*SizeOf(MatrixNxN);

  { For the QR-factorization of RbarT we need just qraux[1..nactiv], but for
    the factorization of R0N we need qraux[1..n-nactiv]. We allocate n! }
  IF MaxAvail < Size THEN
    Error_Handler('Jacobi: Not enough memory!')
  ELSE
    BEGIN
      GetMem(Rbar,SizeOf(MatrixNxN));
      GetMem(RbarT,SizeOf(MatrixNxN));
      GetMem(Nmatrix,SizeOf(MatrixNxN));
      GetMem(R0N,SizeOf(MatrixNxN));
      GetMem(qraux,n*SizeOf(qrauxtype));
      GetMem(rhs,n*SizeOf(rhstype));
      GetMem(qy,n*SizeOf(qytype));
    END;

  { QR-factorization of B }
  GIVENS_factorization(m,n,k,t,tau,B,Q0);
  { B contains the upper triangular factor R0 in packed form }

  { compute constrained spline using the above QR-factorization }
  Constrained_Spline(m,n,order,t,active_bounds,lower,upper,
    TypeOfConstraints,tau,y,alpha,RES,true,true,B,Q0);

  { alpha contains now the solution of subproblem (A) at knot sequence t
    Active_Bounds indicates the set of active constraints }

  { compute derivative matrix D_p }
  Compute_Derivative_Matrix(t,n,k,p,D);

  { compute negative gradients of active constraints }
  Compute_Gradients_of_Subproblem_A(n,k,p,t,active_bounds,D,nactiv,Rbar^);

  { now Rbar contains the negative gradients of active constraints
    Rbar is a band matrix of band width p+1, but is stored as full matrix

    We store Rbar^T also as full matrix and compute the
    Householder QR-factorization }

  { store RbarT }
  FOR i:=1 TO n DO
    FOR jj:=1 TO nactiv DO
      RbarT^[i,jj]:=Rbar^[jj,i];

  { prepare QR factorization of RbarT }
  job:=0;   { no pivoting }

  { QR-decomposition without pivoting }
  QRDC(RbarT^,NMAX,n,nactiv,dummy,qraux^,dummy,job);

  { Now the upper triangle of RbarT contains the matrix R1,
    the lower triangle of RbarT and the vector qraux contain the Householder
    transformations in product form Q1 }

  { compute a null space basis N (= Q_12) of Rbar }

  FOR i:=1 TO n DO
    rhs^[i]:=0.0;     { initialize right hand side }

  FOR jj:=1 TO n-nactiv DO
    BEGIN
      rhs^[nactiv+jj]:=1.0;
      job:=10000;     { Q * rhs }
      result:=QRSL(RbarT^,NMAX,n,nactiv,job,qraux^,rhs^,qy^,dummy,dummy,dummy,dummy);
      { cut jj-th column out of Q_2, qy contains this column }

      FOR i:=1 TO n DO
        Nmatrix^[i,jj]:=qy^[i];

      { restore rhs }
      rhs^[nactiv+jj]:=0.0;
    END;
  { null space basis Nmatrix has been computed }

  { compute R0 * N (matrix in general full!)
    R0 stored in place of B in packed form }
  FOR i:=1 TO n DO
    FOR jj:=1 TO n-nactiv DO
      BEGIN
        temp:=0.0;
        FOR s:=i TO min(n,i+k-1) DO
          temp:=temp+B[i,s-i+1]*Nmatrix^[s,jj];
        R0N^[i,jj]:=temp;
      END;

  { prepare QR factorization of R0N }
  job:=0;   { no pivoting }

  { QR-decomposition without pivoting }
  QRDC(R0N^,NMAX,n,n-nactiv,dummy,qraux^,dummy,job);

  { Now the upper triangle of R0N contains the matrix R2,
    the lower triangle of R0N and the vector qraux contain the Householder
    transformations in product form Q2 }

  { We have all necessary information to compute the Jacobian column by
    column }

  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      Derivative_Spline_Knots(m,n,k,t,tau,alpha,per[kk],v1);
      {
         v1 = ( B' [e^k] )   alpha \in \RealR^m
      }


      Derivative_Gradients_of_Subproblem_A(n,k,p,t,alpha,per[kk],active_bounds,
         nactiv,v2);

      FOR i:=1 TO nactiv DO
        v2[i]:=-v2[i];
      {
         v2 = - ( Rbar' [e^k] )   alpha = GammaBar e^k \in \RealR^nactiv
      }

      job:=11; { solve R1^T * x = v2, R1 upper triangular }
      result:=TRSL(RbarT^,NMAX,nactiv,v2,job);
      {
         v2 = R1^(-T) * v2
      }

      job:=01; { solve R1 * y = x, R1 upper triangular }
      result:=TRSL(RbarT^,NMAX,nactiv,v2,job);
      {
         v2 = (R1^T * R1)^(-1) * v2 = (R1^T * R1)^(-1) * GammaBar * e^k
         v2 \in \RealR^nactiv
      }

      FOR i:=1 TO n DO
        BEGIN
          temp:=0.0;
          FOR jj:=1 TO nactiv DO
            temp:=temp+Rbar^[jj,i]*v2[jj];
          v4[i]:=temp;
        END;
      {
         v4 = Rbar^T * v3 = Rbar^+ * v2 = Rbar^+ GammaBar * e^k \in \RealR^n
      }

      FOR i:=1 TO m DO
        BEGIN
          interval(tau[i],n+1,t,left,out_of_interval);
          v5[i]:=Spline_Value(v4,t,tau[i],left,k);
        END;
      {
         v5 = B * v4 = B * Rbar^+ * GammaBar * e^k \in \RealR^m
      }

      FOR i:=1 TO m DO
        v5[i]:=-v1[i] + v5[i];
      {
        v5 = Jttilde * e^k + B * Rbar^+ * GammaBar * e^k \in \RealR^m
      }

      apply_GIVENS_transformation(m,n,k,t,tau,v5,Q0,v4,RES);
      {
        v4 = [ I_n | 0 ] Q0^T * v5
           = [ I_n | 0 ] Q0^T * (Jttilde * e^k + B * Rbar^+ * GammaBar * e^k)
        v4 \in \RealR^n
      }

      job:=01000;  { Q2^T * v4 }
      result:=QRSL(R0N^,NMAX,n,n-nactiv,job,qraux^,v4,dummy,v4,dummy,dummy,dummy);
      {
        v4 = Q2^T * v5
           = Q2^T *[ I_n | 0 ] Q0^T * (Jttilde * e^k + B * Rbar^+ * GammaBar * e^k)
        v4 \in \RealR^n
      }

      FOR i:=n-nactiv+1 TO n DO
        v4[i]:=0.0;
      {       [ I   0 ]
        v4 =  [       ] * v4
              [ 0   0 ]

        v4 \in \RealR^n
      }

      job:=10000;  { Q2 * v4 }
      result:=QRSL(R0N^,NMAX,n,n-nactiv,job,qraux^,v4,v4,dummy,dummy,dummy,dummy);
      {
        v4 = Q2 * v4
        v4 \in \RealR^n
      }

      FOR i:=1 TO n DO
        v1[i]:=v4[i];
      FOR i:=n+1 TO m DO
        v1[i]:=0.0;
      {      [ I ]
        v1 = [   ]  * v4  \in \RealR^m
             [ 0 ]

             [ I ]        [ I  0 ]
        v1 = [   ] * Q2 * [      ] * Q2^T * [ I | 0 ] * Q0^T * (Jttilde * e^k + B * Rbar^+ * GammaBar * e^k)
             [ 0 ]        [ 0  0 ]
      }

      apply_inverse_GIVENS_transformation(m,n,k,t,tau,v1,Q0);
      { v1 = Q0 * v1  \in \RealR^m

        v1 = P_(BN) * (Jttilde * e^k + B * Rbar^+ * GammaBar * e^k)
      }

      FOR i:=1 TO m DO
        v1[i]:=v5[i]-v1[i];
      {
        v1 = P_(BN)^\perp * (Jttilde * e^k + B * Rbar^+ * GammaBar * e^k)
      }

      { Assign the kk-th column }
      FOR i:=1 TO m DO
        J[i,kk]:=v1[i];

    END;

  FreeMem(Rbar,SizeOf(MatrixNxN));
  FreeMem(RbarT,SizeOf(MatrixNxN));
  FreeMem(Nmatrix,SizeOf(MatrixNxN));
  FreeMem(R0N,SizeOf(MatrixNxN));
  FreeMem(qraux,n*SizeOf(qrauxtype));
  FreeMem(rhs,n*SizeOf(rhstype));
  FreeMem(qy,n*SizeOf(qytype));
END; { Kaufman_Constrained_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Kaufman_Constrained_Smoothing_Jacobi_Matrix
                            (m,n,                   { dimension }
                             k       : integer;     { order     }
                             VAR t   : knotsequence;
                             r       : integer;     { order of derivative   }
                             mu      : real;        { smoothing parameter   }
               VAR Active_bounds     : BLSConstr;
               VAR lower,
                   upper             : BLSbounds;
                   TypeOfConstraints : char;
                             l       : integer;     { number of free knots  }
                             VAR per : free_knots;  { pointer to free knots }
                             VAR Jsmooth : SmoothJacobimatrix
                             );

{ computes an approximation of the Jacobimatrix                               }

TYPE qrauxtype         = ARRAY[1..1] OF real;
      { We need nactiv for RbarT and n-nactiv for R0N --> n-nactiv }
     rhstype           = ARRAY[1..1] OF real;
      { We need n }
     qytype            = ARRAY[1..1] OF real;
      { We need n }

VAR Q0         : OrthogonalMatrix;
    B          : LeastSquareMatrix;
    S          : SmoothingMatrix;
    D          : SmoothingMatrix; { matrix of shape constraints }

    p,                             { order of derivative in shape constriants }
    nactiv,                        { number of active constraints             }
    result,                        { return code of QRSL }
    i,j,jj,kk,job : integer;

    Rbar       : ^MatrixNxN;    { negative gradients of active constraints }
    RbarT      : ^MatrixNxN;    { contains Rbar^T }
    Nmatrix    : ^MatrixNxN;    { null space basis }
    RtildeN    : ^MatrixNxN;    { Rtilde * N }
    v1,v5      : SmoothingVector;{ temporary }
    G1         : datavalues;
    G2         : coefficients;
    v2,v4      : coefficients;  { temporary }
    Size       : LongInt;
    qraux      : ^qrauxtype;
    rhs        : ^rhstype;
    qy         : ^qytype;
    dummy,temp : real;

BEGIN
  CASE TypeOfConstraints OF
    'P','0': p:=0;
    'M','1': p:=1;
    'C','2': p:=2;
  END;

  { allocate temporary working space }
  Size:=3*n*SizeOf(Real)+4*SizeOf(MatrixNxN);

  { For the QR-factorization of RbarT we need just qraux[1..nactiv], but for
    the factorization of R0N we need qraux[1..n-nactiv]. We allocate n! }
  IF MaxAvail < Size THEN
    Error_Handler('Jacobi: Not enough memory!')
  ELSE
    BEGIN
      GetMem(Rbar,SizeOf(MatrixNxN));
      GetMem(RbarT,SizeOf(MatrixNxN));
      GetMem(Nmatrix,SizeOf(MatrixNxN));
      GetMem(RtildeN,SizeOf(MatrixNxN));
      GetMem(qraux,n*SizeOf(qrauxtype));
      GetMem(rhs,n*SizeOf(rhstype));
      GetMem(qy,n*SizeOf(qytype));
    END;

  Compute_Smoothing_Matrix(t,n,k,r,S);

  { QR-factorization of B }
  GIVENS_factorization(m,n,k,t,tau,B,Q0);
  { B contains the upper triangular factor R0 in packed form }

  { compute constrained spline using the above QR-factorization }
  Constrained_Smoothing_Spline(m,n,order,t,r,mu,
      active_bounds,lower,upper,TypeOfConstraints,tau,y,alpha,S,RES,true,true,B,Q0);

  { alpha contains now the solution of subproblem (A) at knot sequence t
    Active_Bounds indicates the set of active constraints }

  { compute derivative matrix D_p }
  Compute_Derivative_Matrix(t,n,k,p,D);

  { compute negative gradients of active constraints }
  Compute_Gradients_of_Subproblem_A(n,k,p,t,active_bounds,D,nactiv,Rbar^);

  { now Rbar contains the negative gradients of active constraints
    Rbar is a band matrix of band width p+1, but is stored as full matrix

    We store Rbar^T also as full matrix and compute the
    Householder QR-factorization }

  { store RbarT }
  FOR i:=1 TO n DO
    FOR j:=1 TO nactiv DO
      RbarT^[i,j]:=Rbar^[j,i];

  { prepare QR factorization of RbarT }
  job:=0;   { no pivoting }

  { QR-decomposition without pivoting }
  QRDC(RbarT^,NMAX,n,nactiv,dummy,qraux^,dummy,job);

  { Now the upper triangle of RbarT contains the matrix R1,
    the lower triangle of RbarT and the vector qraux contain the Householder
    transformations in product form Q1 }

  { compute a null space basis N (= Q_12) of Rbar }

  FOR i:=1 TO n DO
    rhs^[i]:=0.0;     { initialize right hand side }

  FOR j:=1 TO n-nactiv DO
    BEGIN
      rhs^[nactiv+j]:=1.0;
      job:=10000;     { Q * rhs }
      result:=QRSL(RbarT^,NMAX,n,nactiv,job,qraux^,rhs^,qy^,dummy,dummy,dummy,dummy);
      { cut j-th column out of Q_2, qy contains this column }

      FOR i:=1 TO n DO
        Nmatrix^[i,j]:=qy^[i];

      { restore rhs }
      rhs^[nactiv+j]:=0.0;
    END;
  { null space basis Nmatrix has been computed }

  { compute Rtilde * N (matrix in general full!)
    Rtilde stored in place of B in packed form }
  FOR i:=1 TO n DO
    FOR j:=1 TO n-nactiv DO
      BEGIN
        temp:=0.0;
        FOR jj:=i TO min(n,i+k-1) DO
          temp:=temp+B[i,jj-i+1]*Nmatrix^[jj,j];
        RtildeN^[i,j]:=temp;
      END;

  { prepare QR factorization of RtildeN }
  job:=0;   { no pivoting }

  { QR-decomposition without pivoting }
  QRDC(RtildeN^,NMAX,n,n-nactiv,dummy,qraux^,dummy,job);

  { Now the upper triangle of RtildeN contains the matrix R2,
    the lower triangle of RtildeN and the vector qraux contain the Householder
    transformations in product form Q2 }

  { We have all necessary information to compute the Jacobian column by
    column }

  FOR kk:=1 TO l DO             { compute kk-th column of J }
    BEGIN
      Derivative_Spline_Knots(m,n,k,t,tau,alpha,per[kk],G1);
      {
         G1 = ( B' [e^k] )   alpha
      }

      Derivative_Smoothing_Matrix_Knots(n,k,t,r,alpha,per[kk],G2);
      {
         G2 = ( S' [e^k] )  alpha
      }

      { multiplication with sqrt(mu) }
      FOR i:=1 TO n-r DO
        G2[i]:=Sqrt(mu)*G2[i];

      FOR i:=1 TO m do
        v1[i]:=G1[i];
      For i:=1 to n-r do
        v1[m+i]:=G2[i];

      {
             (          B')
        v1 = (            ) [e^k] alpha \in \RealR^(m+n-r)
             (\mu^(1/2) S')
      }

      Derivative_Gradients_of_Subproblem_A(n,k,p,t,alpha,per[kk],active_bounds,
         nactiv,v2);

      FOR i:=1 TO nactiv DO
        v2[i]:=-v2[i];
      {
         v2 = - ( Rbar' [e^k] )   alpha = GammaBar e^k \in \RealR^nactiv
      }

      job:=11; { solve R1^T * x = v2, R1 upper triangular }
      result:=TRSL(RbarT^,NMAX,nactiv,v2,job);
      {
         v2 = R1^(-T) * v2
      }

      job:=01; { solve R1 * y = x, R1 upper triangular }
      result:=TRSL(RbarT^,NMAX,nactiv,v2,job);
      {
         v2 = (R1^T * R1)^(-1) * v2 = (R1^T * R1)^(-1) * GammaBar * e^k
         v2 \in \RealR^nactiv
      }

      FOR i:=1 TO n DO
        BEGIN
          temp:=0.0;
          FOR jj:=1 TO nactiv DO
            temp:=temp+Rbar^[jj,i]*v2[jj];
          v4[i]:=temp;
        END;
      {
         v4 = Rbar^T * v3 = Rbar^+ * v2 = Rbar^+ GammaBar * e^k \in \RealR^n
      }

      { We save Rtilde which is stored in place of B in Rbar. }
      for i:=1 to n do
        for j:=1 to n do
          Rbar^[i,j]:=0.0;
      for i:=1 to n do
        for j:=1 to k do
          Rbar^[i,i+j-1]:=B[i,j];

      FOR i:=1 TO m DO
        BEGIN
          interval(tau[i],n+1,t,left,out_of_interval);
          v5[i]:=Spline_Value(v4,t,tau[i],left,k);
        END;

      FOR i:=1 TO n-r DO
        BEGIN
          temp:=0.0;
          FOR jj:=1 TO r+1 DO
            temp:=temp+S[i,jj] * v4[i+jj-1];
          v5[m+i]:=Sqrt(mu)*temp;
        END;
      {
             (          B )
        v5 = (            ) * Rbar^+ * GammaBar * e^k \in \RealR^(m+n-r)
             (\mu^(1/2) S )
      }

      FOR i:=1 TO m+n-r DO
        v5[i]:=-v1[i] + v5[i];
      {
                             (          B )
        v5 = Jttilde * e^k + (            ) * Rbar^+ * GammaBar * e^k \in \RealR^(m+n-r)
                             (\mu^(1/2) S )
      }

      for i:=1 to m do
        G1[i]:=v5[i];
      for i:=1 to n-r do
        G2[i]:=v5[m+i];

      Get_BTvec(m,n,k,t,tau,G1,v2);
      {
        v2 = B^T * v5(1:m)
      }
      Get_STvec(m,n,k,r,t,G2,v4,S);
      {
        v4 = S^T * v5(m+1:m+n-r)
      }
      for i:=1 to n do
        v4[i]:=v2[i]+sqrt(mu)*v4[i];
      {
             (          B ) ^T
        v4 = (            )    * v5 \in \RealR^(n)
             (\mu^(1/2) S )
      }


      job:=11; { solve Rtilde^T * x = v4, Rtilde upper triangular }
      result:=TRSL(Rbar^,NMAX,n,v4,job);
      {
         v4 = Rtilde^(-T) * v4
      }

      job:=01000;  { Q2^T * v4 }
      result:=QRSL(RtildeN^,NMAX,n,n-nactiv,job,qraux^,v4,dummy,v4,dummy,dummy,dummy);
      {
        v4 = Q2^T * v4
                                 (          B ) ^T                  (          B )
           = Q2^T *Rtilde^(-T) * (            )    (Jttilde * e^k + (            ) * Rbar^+ * GammaBar * e^k)
                                 (\mu^(1/2) S )                     (\mu^(1/2) S )
        v4 \in \RealR^n
      }

      FOR i:=n-nactiv+1 TO n DO
        v4[i]:=0.0;
      {       [ I   0 ]
        v4 =  [       ] * v4
              [ 0   0 ]

        v4 \in \RealR^n
      }

      job:=10000;  { Q2 * v4 }
      result:=QRSL(RtildeN^,NMAX,n,n-nactiv,job,qraux^,v4,v4,dummy,dummy,dummy,dummy);
      {
        v4 = Q2 * v4
        v4 \in \RealR^n
      }

      job:=01; { solve Rtilde * x = v4, Rtilde upper triangular }
      result:=TRSL(Rbar^,NMAX,n,v4,job);
      {
         v4 = Rtilde^(-1) * v4
      }

      FOR i:=1 TO m DO
        BEGIN
          interval(tau[i],n+1,t,left,out_of_interval);
          v1[i]:=Spline_Value(v4,t,tau[i],left,k);
        END;

      FOR i:=1 TO n-r DO
        BEGIN
          temp:=0.0;
          FOR jj:=1 TO r+1 DO
            temp:=temp+S[i,jj] * v4[i+jj-1];
          v1[m+i]:=Sqrt(mu)*temp;
        END;

      { v1 \in \RealR^(m+n-r) }

      FOR i:=1 TO m DO
        v1[i]:=v5[i]-v1[i];
      {
        v1 = P_(BmuS N)^\perp * (Jttilde * e^k + BmuS * Rbar^+ * GammaBar * e^k)
      }

      { Assign the kk-th column }
      FOR i:=1 TO m+n-r DO
        Jsmooth[i,kk]:=v1[i];

    END;

  FreeMem(Rbar,SizeOf(MatrixNxN));
  FreeMem(RbarT,SizeOf(MatrixNxN));
  FreeMem(Nmatrix,SizeOf(MatrixNxN));
  FreeMem(RtildeN,SizeOf(MatrixNxN));
  FreeMem(qraux,n*SizeOf(qrauxtype));
  FreeMem(rhs,n*SizeOf(rhstype));
  FreeMem(qy,n*SizeOf(qytype));
END; { Kaufman_Constrained_Smoothing_Jacobi_Matrix }

{-----------------------------------------------------------------------------}

END.