UNIT knotrem;
{ automatic knot removal procedures                                           }

INTERFACE
USES def,         { global declarations                                       }
     BSpline,     { procedures to compute B-Splines and l2-approximations     }
     Smooth,      { procedures to compute smoothing B-Splines                 }
     Gauss,       { Gauss-Newton-like methods for nonlinear problems          }
     various,     { various procedures for menu handling                      }
     BLAS,        { Basic Linear Algebra Subroutines                          }
     crt;

{-----------------------------------------------------------------------------}

PROCEDURE Automatic_Knot_Removal_Phase1;
{ Recursively
    - rank knots according to their importance for the approximation which is
      measured by the jump in the first discontinuous derivative
    - remove the less important knot
  until knot sequence is not longer acceptable, i.e., residual sum is greater
  than a prescribed bound                                                     }

{-----------------------------------------------------------------------------}

PROCEDURE Automatic_Knot_Removal_Phase2;
{ Recursively
    - optimize knot sequence
    - rank knots according to their importance for the approximation which is
      measured by the jump in the first discontinuous derivative
    - remove the less important knot
  until knot sequence is not longer acceptable, i.e., residual sum is greater
  than a prescribed bound                                                     }

{-----------------------------------------------------------------------------}

PROCEDURE Remove_Knot;
{ remove the knot with the smallest jump in the first discontinous derivative }

{-----------------------------------------------------------------------------}

IMPLEMENTATION

{-----------------------------------------------------------------------------}

PROCEDURE Automatic_Knot_Removal_Phase1;
VAR ch                                                   : char;
    upper_bound,jump_left,jump_right,jump,jump_min       : real;
    already_removed,No_progress                          : Boolean;
    step1,jump_min_ind,jj,nold,lold                      : integer;
    tsave                                                : knotsequence;
    S                                                    : SmoothingMatrix;

BEGIN
  TextBackground(MyBackground);
  ClrScr;
  Box(20,60,2,4,MyMenuBackground,MyMenuForeground);
  Msg(30,3,'Knot removal procedure');
  Box(5,75,8,22,Black,White);
  Msg(9,14,'Enter upper bound for residual sum: ');
  ReadLn(Upper_bound);
  Msg(9,18,'Smoothing spline (y/n): ');
  ReadLn(ch);
  IF UpCase(ch) IN ['Y','J'] THEN
    How:=True   { smoothing }
  ELSE
    How:=False; { smoothing }

  already_removed:=False;
  No_Progress:=False; {Assume that we can reduce the number of knots }
  step1:=0;
  REPEAT
    Box(20,60,5,7,White,Black);
    TextColor(Black);
    Msg(22,6,'Removing knots! No. inner knots: ');
    Write(l);

    Inc(step1);

    { We do not locate a good knot sequence }

    { Compute the residual sum and see if the number and location of knots is
      already feasible }
    IF How=False THEN
      l2Appr(m,n,order,t,tau,y,alpha,RES)
    ELSE
      BEGIN
        Compute_Smoothing_Matrix(t,n,order,r,S);
        Compute_Smoothing_spline(m,n,order,t,r,mu,tau,y,S,alpha,RES);
      END;
    IF res < Upper_Bound THEN
      BEGIN
        { number and locations of knots are feasible, so we try to
          reduce one more knot }

        { rank the free knots according to the absolute size of the
          jump in the first (order-1) noncontinous derivative }
        jump_left:=spline_value_deriv(alpha,t,(t[per[1]]+t[per[1]-1])*0.5,order-1,order);
        jump_right:=spline_value_deriv(alpha,t,t[per[1]],order-1,order);

        jump_min:=Abs(jump_left-jump_right); { jump at t[per[1]] }
        jump_min_ind:=1;
        FOR jj:=2 TO l DO
          BEGIN
            jump_left:=spline_value_deriv(alpha,t,(t[per[jj]]+t[per[jj]-1])*0.5,order-1,order);
            jump_right:=spline_value_deriv(alpha,t,t[per[jj]],order-1,order);

            jump:=Abs(jump_left-jump_right); { jump at t[per[jj]] }
            IF jump < jump_min THEN
              BEGIN
                jump_min:=jump;
                jump_min_ind:=jj;
              END;
          END;
        { Now per[jump_min_ind] contains the index of the free knot
          with the smallest jump in the (order-1)-th  derivative. This
          knot will now be removed. }

        nold:=n;
        lold:=l;
        Copy1(n+order,t,tsave);
        { Save the old knot sequence so that we can recover the knot
          sequence if we had no success. }

        { Set new dimension n, knot sequence t, pointer to free knots
          and number of free knots }
        Dec(n);
        Dec(l);
        FOR jj:=per[jump_min_ind] TO n+order DO
          t[jj]:=t[jj+1];
        Already_removed:=true;
        { Now we try the next iteration }
      END {if res<upper_bound}
    ELSE
      BEGIN
        { We cannot remove any further knots. If we have already
          removed a knot restore the old knot sequence. If we have
          not removed any knot then Upper_bound or the number of
          knots is too small }

        No_Progress:=True;
        { No further progress in reducing the number of knots can be made }

        IF already_removed=true THEN
          BEGIN
            { restore the last knot sequence }
            n:=nold;
            l:=lold;
            Copy1(n+order,tsave,t);
          END;
      END; { Res >= Upper_bound }
  UNTIL (No_progress=true) OR (step1>STEPMAX) OR (l=0);
END; { Automatic_Knot_Removal_Phase1 }

{-----------------------------------------------------------------------------}

PROCEDURE Automatic_Knot_Removal_Phase2;
VAR ch                                                   : char;
    upper_bound,jump_left,jump_right,jump,jump_min       : real;
    already_removed,No_progress                          : Boolean;
    step1,jump_min_ind,jj,nold,lold                      : integer;
    tsave                                                : knotsequence;
    S                                                    : SmoothingMatrix;

BEGIN
  TextBackground(MyBackground);
  ClrScr;
  Box(20,60,2,4,MyMenuBackground,MyMenuForeground);
  Msg(30,3,'Knot removal procedure');
  Box(5,75,8,22,Black,White);
  Msg(9,14,'Enter upper bound for residual sum: ');
  ReadLn(Upper_bound);
  Msg(9,16,'rel. constraints (y/n): ');
  ReadLn(ch);
  IF UpCase(ch) IN ['Y','J'] THEN
    Constraints_abs:=FALSE
  ELSE
    Constraints_abs:=TRUE;
  Msg(9,18,'Smoothing spline (y/n): ');
  ReadLn(ch);
  IF UpCase(ch) IN ['Y','J'] THEN
    How:=True   { smoothing    }
  ELSE
    How:=False; { no smoothing }
  Msg(9,20,'Kaufman model (y/n): ');
  ReadLn(ch);
  IF UpCase(ch) IN ['Y','J'] THEN
    Kaufman_Model:=True   { Kaufman model       }
  ELSE
    Kaufman_Model:=False; { Golub/Pereyra model }

  already_removed:=False;
  No_Progress:=False; {Assume that we can reduce the number of knots }
  step:=stepmax;
  step1:=0;
  REPEAT
    Box(20,60,5,7,White,Black);
    TextColor(Black);
    Msg(22,6,'Removing knots! No. inner knots: ');
    Write(l);

    Inc(step1);

    step:=stepmax;
    { Locate good knot sequence }
    IF How=False THEN
      Regularized_Generalized_Gauss_Newton(
                m,n,order,t,l,per,step,RetCode,normGradient,Grad_p)
    ELSE
      Smoothing_Regularized_Generalized_Gauss_Newton(
                m,n,order,t,r,mu,l,per,step,RetCode,normGradient,Grad_p);

    { Now we have computed a good knot sequence, compute the residual sum
      and see if the number and location of knots is already feasible     }
    IF How=False THEN
      l2Appr(m,n,order,t,tau,y,alpha,RES)
    ELSE
      BEGIN
        Compute_Smoothing_Matrix(t,n,order,r,S);
        Compute_Smoothing_spline(m,n,order,t,r,mu,tau,y,S,alpha,RES);
      END;
    IF res < Upper_Bound THEN
      BEGIN
        { number and locations of knots are feasible, so we try to
          reduce one more knot }

        { rank the free knots according to the absolute size of the
          jump in the first (order-1) noncontinous derivative }
        jump_left:=spline_value_deriv(alpha,t,(t[per[1]]+t[per[1]-1])*0.5,order-1,order);
        jump_right:=spline_value_deriv(alpha,t,t[per[1]],order-1,order);

        jump_min:=Abs(jump_left-jump_right); { jump at t[per[1]] }
        jump_min_ind:=1;
        FOR jj:=2 TO l DO
          BEGIN
            jump_left:=spline_value_deriv(alpha,t,(t[per[jj]]+t[per[jj]-1])*0.5,order-1,order);
            jump_right:=spline_value_deriv(alpha,t,t[per[jj]],order-1,order);

            jump:=Abs(jump_left-jump_right); { jump at t[per[jj]] }
            IF jump < jump_min THEN
              BEGIN
                jump_min:=jump;
                jump_min_ind:=jj;
              END;
          END;
        { Now per[jump_min_ind] contains the index of the free knot
          with the smallest jump in the (order-1)-th  derivative. This
          knot will now be removed. }

        nold:=n;
        lold:=l;
        Copy1(n+order,t,tsave);
        { Save the old knot sequence so that we can recover the knot
          sequence if we had no success. }

        { Set new dimension n, knot sequence t, pointer to free knots
          and number of free knots }
        Dec(n);
        Dec(l);
        FOR jj:=per[jump_min_ind] TO n+order DO
          t[jj]:=t[jj+1];
        Already_removed:=true;
        { Now we try the next iteration }
      END {if res<upper_bound}
    ELSE
      BEGIN
        { We cannot remove any further knots. If we have already
          removed a knot restore the old knot sequence. If we have
          not removed any knot then Upper_bound or the number of
          knots is too small }

        No_Progress:=True;
        { No further progress in reducing the number of knots can be made }

        IF already_removed=true THEN
          BEGIN
            { restore the last knot sequence }
            n:=nold;
            l:=lold;
            Copy1(n+order,tsave,t);
          END;
      END; { Res >= Upper_bound }
  UNTIL (No_progress=true) OR (step1>STEPMAX) OR (l=0);
END; { Automatic_Knot_Removal_Phase2 }

{-----------------------------------------------------------------------------}

PROCEDURE Remove_Knot;
VAR jump_min_ind, jj                   : integer;
    jump_left, jump_right, jump_min, jump : real;
BEGIN
  jump_left:=spline_value_deriv(alpha,t,(t[per[1]]+t[per[1]-1])*0.5,order-1,order);
  jump_right:=spline_value_deriv(alpha,t,t[per[1]],order-1,order);

  jump_min:=Abs(jump_left-jump_right); { jump at t[per[1]] }
  jump_min_ind:=1;
  FOR jj:=2 TO l DO
    BEGIN
      jump_left:=spline_value_deriv(alpha,t,(t[per[jj]]+t[per[jj]-1])*0.5,order-1,order);
      jump_right:=spline_value_deriv(alpha,t,t[per[jj]],order-1,order);
      jump:=Abs(jump_left-jump_right); { jump at t[per[jj]] }
      IF jump < jump_min THEN
        BEGIN
          jump_min:=jump;
          jump_min_ind:=jj;
        END;
    END;

  { Now per[jump_min_ind] contains the index of the free knot
    with the smallest jump in the (order-1)-th  derivative. This
    knot will now be removed. }

  { diagnose output }
  Box(20,60,6,8,MyMenuBackground,MyMenuForeground);
  Msg(23,7,' Removing knot number: ');
  WriteLn(per[jump_min_ind]);
  Wait(10000);

  Dec(n);
  Dec(l);
  FOR jj:=per[jump_min_ind] TO n+order DO
    t[jj]:=t[jj+1];
END; { Remove_Knot }

{-----------------------------------------------------------------------------}

END. { unit Knotrem }