UNIT Lawson;
{ This unit contains procedures for the solution of constrained linear
  least squares problems.
  The algorithms are taken from
      C.L.Lawson; R.J. Hanson: Solving least squares problems
  and adapted to Turbo-Pascal.                                                }

{$R-}
INTERFACE

USES linpack;     { Routines from LINPACK for QR-factorization of a matrix    }

TYPE real=double;
CONST macheps=2.220446049e-16;

{-----------------------------------------------------------------------------}

PROCEDURE LSI ( VAR Ap;dA,m,n:integer;VAR Cp;dC,ncstr:integer;VAR xp;
                VAR bp;VAR hp;VAR workrealp;VAR workintp;VAR Mode:Integer);

{ Computation of the solution of a

  Linear Least Squares Problem with linear inequality constraints (LSI)


  || A*x - b ||  -----> min     subject to  C*x >= h

        A : Array[1..m,1..n] of real;
        x : Array[1..n] of real;
        b : Array[1..m] of real;
        C : Array[1..ncstr,1..n] of real;
        h : Array[1..ncstr] of real;

  The problem LSI will be transformed to an equivalent Least Distance Problem
  via Householder transformations with column pivoting.
  The matrix A is assumed to have full rank n. Otherwise the procedure will
  abort.

  Note: The matrices A, C and the vectors b, h will be overwritten.

  A:    Array[1..m,1..n] of real; stored as vector with m*n elements ;
  Ap:   Pointer to the elements of A

  dA:   Integer, leading dimension of A
        if A was declared as Array[1..MMAX,1..NMAX] of real then dA:=NMAX

  m:    Integer, number of rows of A

  n:    Integer, number of columns of A

  C:    Array[1..ncstr,1..n] of real; stored as vector with ncstr*n elements;
  Cp:   Pointer to the elements of C

  dC:   Integer, leading dimension of C
        if C was declared as Array[1..NCSTR_MAX,1..NMAX] of real then dC:=NMAX

 ncstr: Integer, number of rows of C; number of constraints

  x:    Array[1..n] of real; on output x contains the solution of the problem
  xp:   Pointer to the elements of x

  b:    Array[1..m] of real; right hand side of linear least squares problem
        On Output b will be overwritten with Q^T b
  bp:   Pointer to the elements of b

  h:    Array[1..ncstr] of real; right hand side of constraints
  hp:   Pointer to the elements of h

  workreal:    Array[1.. (n+1) * (2+ncstr) + 2 * ncstr] of real;
               real working array
  workrealp:   Pointer to the elements of workreal

  workint:    Array[1.. n+ncstr) ] of integer; integer working array
  workintp:   Pointer to the elements of workint

  Mode:       Integer, status flag after termination
               mode=1  succesfull
               mode=2  bad dimension
               mode=3  maximum number of iterations exceeded
               mode=4  inequality constraints incompatible
               mode=5  A (nearly) rankdeficient

  Subroutines called QRDC, QRSL, SolveUTS, LDP                                }

{-----------------------------------------------------------------------------}

PROCEDURE LDP ( VAR Gp; dG, m, n:integer; VAR hp ; VAR xp;
                VAR XNORM:real; VAR wp; VAR INDEXp; VAR MODE:integer);

{ Least Distance Programming
  Lawson; Hanson: Solving Least Squares Problems

  This subroutine computes the solution of the following constrained least
  squares problem:

    minimize || x ||  subject to    G * x >= h

  It reduces the Least Distance Problem to a related Non-Negative Least
  Squares Problem.


  G:    Array[1..m,1..n] of real; stored as vector with m*n elements ;
        Either m>=n or m<n is permitted and there is no restriction on the
        rank of G; the contents of G are not modified
  Gp:   Pointer to the elements of G

  dG:   Integer, leading dimension of G
        if G was declared as Array[1..MMAX,1..NMAX] of real then dG:=NMAX

  m:    Integer, number of rows of G

  n:    Integer, number of columns of G

  h:    Array[1..m] of real; contains vector h of the problem; the contents
        of the array h are not modified
  hp:   Pointer to the elements of h

  x:    Array[1..n] of real; on output x contains the solution of the problem
  xp:   Pointer to the elements of x

 xnorm: on output xnorm contains the norm of the solution vector

  w:    Array[1.. (m+2)*(n+1)+2*m] of real; temporary working storage
  wp:   Pointer to the elements of w

  index: Array[1..m] of integer; temporary working storage
  indexp: Pointer to the elements of index

  mode: Integer; status flag after termination
          mode=1  succesfull
          mode=2  bad dimension
          mode=3  maximum number (3*m) of iterations exceeded
          mode=4  inequality constraints are incompatible


  Subroutines called: NNLS, DIFF                                              }

{-----------------------------------------------------------------------------}

PROCEDURE NNLS ( VAR Ap; dA, m, n:integer; VAR bp ; VAR xp;
                VAR RNORM:real; VAR wp; VAR zzp; VAR INDEXp; VAR MODE:integer);

{ Nonnegative Linear Least Squares
  Lawson; Hanson: Solving Least Squares Problems

  This subroutine computes the solution of the following constrained least
  squares problem:

    Solve A * x ~= b  subject to     x >= 0



  A:    Array[1..m,1..n] of real; stored as real vector of the dimension m*n,
        Either m>=n or m<n is permitted and there is no
        restriction on the rank of A; On termination A contains Q*A, where
        Q is an m,m orthogonal matrix implicitly generated by the subroutine
  Ap:   Pointer to the elements of A

  dA:   Integer, leading dimension of A
        if A was decleared as Array[1..NMAX,1..MMAX] of real then dA:=NMAX

  m:    Integer, number of rows of A

  n:    Integer, number of columns of A

  b:    Array[1..m] of real; contains vector b of the problem; On termination
        b contains Q*b
  bp:   Pointer to the elements of b

  x:    Array[1..n] of real; on output x contains a solution of the problem
  xp:   Pointer to the elements of x

 rnorm: on output rnorm contains the norm of the final residual vector
        rnorm = || b - A *x ||

  w:    Array[1.. n] of real; On termination w contains the dual vector
        w = A^T ( b - A * x)
  wp:   Pointer to the elements of w

  zz:   Array[1..m] of real; temporary working storage
  zzp:  Pointer to the elements of z

  index: Array[1..n] of integer; temporary working storage
         On exit the contents of this array define the sets P (ositive)
         and Z (ero) as follows

             index[1] .. index[NSETP]   = set P
             index[NSETP+1] .. index[n] = set Z
  indexp: Pointer to the elements of index

  mode: Integer; status flag after termination
          mode=1  succesfull
          mode=2  bad dimension
          mode=3  maximum number (3*n) of iterations exceeded


  Subroutines called: H12, G1, G2, DIFF                                       }

{-----------------------------------------------------------------------------}

PROCEDURE H12 ( mode,lpivot,l1,m :integer; VAR U_p ; iue :integer;
                VAR up:real; VAR Cp; ice,icv,ncv : integer);

{ Construction and/or Application of a single Householder transformation

           I + U * U^T
      Q = ---------------
               B

  Lawson; Hanson: Solving Least Squares Problems

  mode:   Integer, mode = 1 or mode = 2
          if mode = 1 then
             Algorithm H1
             Construction and, if ncv > 0, application of a Householder
             transformation to a set of ncv m-vectors stored in the array C
          else
             Algorithm H2
             Application of a previously computed Householder transformation

  lpivot: Integer, index of the pivot element

  l1, m:  Integer,
          if l1 <= m the transformation will be constructed to zero elements
          indexed from l1 to m
          if l1 > m the subroutine does an identity transformation

  U:      U contains m elements with a positive storage increment IUE between
          elements

          if v is stored in a single subscripted Pascal array w with
          w[i]=v[i] then
                          "U,IUE" ----> "W,1"

          if v is stored as the j th column of a doubly subscripted Pascal
          array A with A[i,j] = v[i] and A : Array [1..MMAX,1..NMAX] of real
          then
                          "U,IUE" ----> "A[1,j],NMAX"

          if v is stored as the i th row of a doubly subscripted Pascal
          array A with A[i,j] = v[j] and A : Array [1..MMAX,1..NMAX] of real
          then
                          "U,IUE" ----> "A[i,1],1"

          Note: This is an important difference to FORTRAN!

          On entry to H1 U contains the pivot vector
          On exit from H1 U and UP contain quantities defining the vector u
          from the Householder transformation

          On entry to H2 U and UP should contain quantities previously
          computed by H1. These will not be modified by H2.

  U_p:    Pointer to the elements of U

  IUE:    Integer, Storage increment between elements of U

  UP:     Integer, see U

  C:      On entry to H1 or H2 C contains a matrix which will be regarded
          as a set of vectors to which the Householder transformation is to
          be applied
          On exit C contains the set of transformed vectors
          C is stored as a "long" vector

          C must contain initially a set of NCV m-vectors with a positive
          storage increment of ICE between elements of a vector and a storage
          increment of ICV between vectors. Thus if z[i,j] denotes the i th
          element of the j th vector, then
                  C[ 1 + (i-1) * ICE + (j-1) * ICV ] = z[i,j]

          if the vectors in C are the rows of a matrix A,
               A : Array[1..NMAX,1..MMNAX] of real
          then
                    ice:=1;
                    icv:=NMAX;

          if the vectors in C are the columns of a matrix A,
               A : Array[1..NMAX,1..MMNAX] of real
          then
                    ice:=nmax;
                    icv:=1;

  Cp:     Pointer to the elements of C

  ICE:    Integer, Storage increment between elements of vectors in C

  ICV:    Integer, Storage increment between vectors in C

  NCV:    Integer, Number of vectors in C to be transformed.
          if ncv <= 0 no operations will be done on C                         }

{-----------------------------------------------------------------------------}

FUNCTION DIFF (x,y : real) : real;
{  Function to compute x-y in floating point arithmetic to avoid explicit
   use of floating point threshhold epsilon                                   }

{-----------------------------------------------------------------------------}

PROCEDURE G1 ( VAR a,b,CosG,SinG,sig :real);

{ Compute Givens rotation
  Lawson; Hanson: Solving Least Squares Problems

  This subroutine computes a Givens rotation so that

        [  c   s ] [a] = [ sqrt( a^2 + b^2 ) ]
        [ -s   c ] [b]   [        0          ]

  and it computes sig = sqrt( a^2 + b^2 )
  sig may be in the same location as a or b                                   }

{-----------------------------------------------------------------------------}

PROCEDURE G2 ( VAR CosG,SinG,x,y :real);

{ Apply Givens rotation
  Lawson; Hanson: Solving Least Squares Problems

  This subroutine applies a Givens rotation

        [  c   s ] [x] = [  c * x + s * y ]
        [ -s   c ] [y]   [ -s * x + c * y ]                                   }

{-----------------------------------------------------------------------------}

FUNCTION Condition_Estimate( VAR Rp;dR,n : integer;VAR workp) : real;
{ Estimates l_1 condition number of upper triangular matrix R

  Algorithm CONDEST in Dennis/Schnabel:'Numerical methods for unconstrained
  optimization and nonlinear equations'
  due to Cline, Moler, Stewart and Wilkonson [1979]
  'An estimate for the condition number of a matrix'
  SIAM J.Numer.Anal. 16,368-378)


  R:    Array[1..n,1..n] of real; stored as vector with n*n elements ;
  Rp:   Pointer to the elements of R

  dR:   Integer, leading dimension of R
        if R was declared as Array[1..MMAX,1..NMAX] of real then dR:=NMAX

  n:    Integer, dimension of the system

  workp: Array[1..3*n] of real; work array                                    }

{-----------------------------------------------------------------------------}

IMPLEMENTATION

FUNCTION max_real (a,b : real) : real;

BEGIN
  IF a > b THEN
    max_real:=a
  ELSE
    max_real:=b;
END; { max_real }

{-----------------------------------------------------------------------------}

FUNCTION ind (LDA,i,j : integer) : integer;
{ Function for index computations
  Element a[i,j] corresponds to  vector [ (i-1) * LDA+j ]
  ( Arrays are stored row by row in Pascal )
  A must be declared as Array[1..m,1..LDA]                                    }

BEGIN
  ind:=(i-1)*LDA+j;
END; { ind }

{-----------------------------------------------------------------------------}

FUNCTION SolveUTS ( VAR Rp;dR,n : integer; VAR xp; VAR yp) : integer;
{ Solve upper triangular system R * x = y
  SolveUTS <> 0 if system is nearly singular

  R:    Array[1..n,1..n] of real; stored as vector with n*n elements ;
  Rp:   Pointer to the elements of R

  dR:   Integer, leading dimension of R
        if R was declared as Array[1..MMAX,1..NMAX] of real then dR:=NMAX

  n:    Integer, dimension of the system

  x:    Array[1..n] of real;
  xp:   Pointer to the elements of x

  y:    Array[1..n] of real;
  yp:   Pointer to the elements of y                                          }

TYPE vector = ARRAY[1..1] OF real;
LABEL return;
VAR R     : vector ABSOLUTE Rp;
    x     : vector ABSOLUTE xp;
    y     : vector ABSOLUTE yp;
    Flag  : Boolean;
    i,j   : integer;
    t,eps : real;

BEGIN
  eps:=100*macheps;
  Flag:=True;
  FOR i:=n DOWNTO 1 DO
    BEGIN
      t:=y[i];
      FOR j:=n DOWNTO i+1 DO
        t:=t-R[ ind(dR,i,j) ] * x[j];
      IF Abs( R[ ind(dR,i,i)] ) < eps THEN
        BEGIN
          Flag:=False;
          GOTO return;
        END;
      x[i]:=t/R[ ind(dR,i,i) ];
    END;

return:
  IF NOT Flag THEN
    SolveUTS:=1
  ELSE SolveUTS:=0;
END;

{-----------------------------------------------------------------------------}

FUNCTION Condition_Estimate( VAR Rp;dR,n : integer;VAR workp) : real;
{ Estimates l_1 condition number of upper triangular matrix R

  Algorithm CONDEST in Dennis/Schnabel:'Numerical methods for unconstrained
  optimization and nonlinear equations'
  due to Cline, Moler, Stewart and Wilkonson [1979]
  'An estimate for the condition number of a matrix'
  SIAM J.Numer.Anal. 16,368-378)


  R:    Array[1..n,1..n] of real; stored as vector with n*n elements ;
  Rp:   Pointer to the elements of R

  dR:   Integer, leading dimension of R
        if R was declared as Array[1..MMAX,1..NMAX] of real then dR:=NMAX

  n:    Integer, dimension of the system

  workp: Array[1..3*n] of real; work array
                                                                              }

TYPE vector = ARRAY[1..1] OF real;
LABEL return;
VAR R     : vector ABSOLUTE Rp;
    work  : vector ABSOLUTE workp;
    { work[1]     .. work[n]        x
      work[n+1]   .. work[2*n]      p
      work[2*n+1] .. work[3*n]      pm }

    i,j,result                 : integer;
    temp,tempm,est,xp,xm,xnorm : real;

BEGIN
  est:=Abs(R[ind(dR,1,1)]);
  FOR j:=2 TO n DO
    BEGIN
      temp:=0.0;
      FOR i:=1 TO j DO
        temp:=temp+ Abs( R[ ind(dR,i,j) ]);
      IF est < temp THEN
        est:=temp;
    END;
  { est:= || R ||_1 }

  work[1]:=1.0/R[ ind(dR,1,1) ];
  FOR i:=2 TO n DO
    work[n+i]:=R[ ind(dR,1,i) ] * work[1];

  FOR j:=2 TO n DO
    BEGIN
      xp:=(1.0-work[n+j])/R[ ind(dR,j,j) ];
      xm:=(-1.0-work[n+j])/R[ ind(dR,j,j) ];
      temp:=Abs(xp);
      tempm:=Abs(xm);
      FOR i:=j+1 TO n DO
        BEGIN
          work[2*n+i]:=work[n+i] + R[ ind(dR,j,i)] * xm;
          tempm:=tempm+ Abs(work[2*n+i]) / Abs( R[ ind(dR,i,i)] );
          work[n+i]:=work[n+i] + R[ ind(dR,j,i) ] * xp;
          temp:=temp+ Abs(work[n+i]) / Abs( R[ ind(dR,i,i) ] );
        END;
      IF temp>=tempm THEN
        work[j]:=xp
      ELSE
        BEGIN
          work[j]:=xm;
          FOR i:=j+1 TO n DO
            work[n+i]:=work[2*n+i];
        END;
    END;

  xnorm:=0.0;
  FOR j:=1 TO n DO
    xnorm:=xnorm+Abs(work[j]);

  est:=est/xnorm;
  result:=SolveUTS(R,dR,n,work[n+1],work[1]); { p:=R^(-1) * x }

  xnorm:=0.0;
  FOR j:=1 TO n DO
    xnorm:=xnorm+Abs(work[n+j]);

  est:=est*xnorm;
  Condition_Estimate:=est;
END;

{-----------------------------------------------------------------------------}

PROCEDURE LSI ( VAR Ap;dA,m,n:integer;VAR Cp;dC,ncstr:integer;VAR xp;
                VAR bp;VAR hp;VAR workrealp;VAR workintp;VAR Mode: Integer);

LABEL return;
TYPE vector=ARRAY[1..1] OF real;
     vectorint=ARRAY[1..1] OF integer;
VAR A                              : vector ABSOLUTE Ap;
    C                              : vector ABSOLUTE Cp;
    x                              : vector ABSOLUTE xp;
    b                              : vector ABSOLUTE bp;
    h                              : vector ABSOLUTE hp;
    work                           : vector ABSOLUTE workrealp;
    pivot                          : vectorint ABSOLUTE workintp;

    dummy,t,est                    : real;
    qraux_p,work_p,CTilde_p        : integer; { pointer to elements in work }

    job,i,j,k,result,joba,jobb,
    jobc,jobd,jobe                 : integer;


BEGIN
  job:=1;   { column pivoting }

{  work [ 1   ] .. work [  n  ]        qraux = Array[1..n] of real
                                       qraux_p:=1;

   work [n+1  ] .. work [ 2*n ]        work = Array[1..n] of real;
                                       work_p:=n+1;
                                                                              }

  FOR j:=1 TO n DO
    pivot[j]:=0;           { all columns are free }

  qraux_p:=1;   { points to the elements of auxilliary vector qraux }

  work_p:=n+1;  { points to temporary working space for QR-factorization }

  { Now all pointers to the necessary variables have been set. }

  QRDC(A,dA,m,n,pivot,work[qraux_p],work[work_p],job);
  { QR-decomposition with column pivoting }
  { Now the upper triangle of A contains the matrix R }

  { Condition estimation works only for regular matrices.
    Because A[n,n] is the smallest diagonal element we test
    explicitly if matrix is almost singular. }
  IF Abs(A[(n-1)*dA+n])> macheps THEN
    est:=Condition_Estimate ( A,dA,n,work[n+1])
    { Estimate l_1 Condition number of R }
  ELSE
    est:=2.0/Sqrt(macheps);

  IF est > 1.0/Sqrt(macheps) THEN
    BEGIN
      Mode:=5;  { R is nearly singular }
      GOTO return;
    END;

  { job:   Integer, decimal expansion abcde

         if a <> 0 compute QY
         if b,c,d or e <> 0 compute QTY
         if c <> 0 compute b
         if d <> 0 compute residual vector r
         if e <> 0 compute xb                                                 }

  joba:=0;
  jobb:=1;   { we need Q^T * y with y:=b }
  jobc:=0;
  jobd:=0;
  jobe:=0;

  job:=joba*10000+jobb*1000+jobc*100+jobd*10+jobe;

  result:= QRSL(A,dA,m,n,job,work[qraux_p],   { qraux }
                b,                            { right hand side }
                dummy,
                b,                            { On Output Q^T * b }
                dummy,
                dummy,
                dummy);

  { Now b contains Q^T * b }

  IF result <> 0 THEN
    BEGIN
      Mode:=5;
      GOTO Return;
    END;
  { The temporary working space for the QR-Factorization is not longer
    needed. We can use it for other purposes.
    We have to compute
           \tilde(C) = C * P *  R^(-1) and
           \tilde(h) = h - \tilde(C) * c_1 }

  FOR j:=1 TO n DO
    work[j]:=0.0;
  { work[1] .. work[n] contains the unit vector e_j }

  work_p:=n+1;
  { work[n+1] .. work[2*n] shall contain the solution of R * x = e_j
    work[2*n+1] .. work[2*n + ncstr *n] shall contain the matrix \tilde(C)
    stored row by row
    CTilde = Array[1..NCSTR,1..N] of real;

    CTilde_p:=2*n+1;

    C~(1,1) --> work(CTilde_p)
    C~(1,n) --> work(CTilde_p + n-1)
    C~(i,j) --> work(CTilde_p + (i-1) * n + j - 1 )    }

  CTilde_p:=2*n+1;  { pointer to first element of C~ }

  FOR j:=1 TO n DO  { compute column j of C~ }
    BEGIN
      work[j]:=1.0; { e_j }

      result:=SolveUTS ( A,dA,n,work[work_p], work[1]);
      { Solve R * x = e_j }

      IF result <> 0 THEN
        BEGIN
          Mode:=5;
          GOTO return;
        END;
      { work[work_p] .. work[work_p+n-1] contains R^(-1) * e_j  }

      { compute j-th column of \tilde(C) }

      FOR i:=1 TO ncstr DO
        BEGIN
           { Compute element C~(i,j)
             C~(i,j) = \sum_(k=1)^(n) C(i,k) *  ( P * R^(-1) * e_j )_k  }
           t:=0.0;
           FOR k:=1 TO n DO
             BEGIN
               t:=t + C[ ind(dC,i,pivot[k]) ] * work[ work_p + k - 1 ];
            END;
           work[CTilde_p + (i-1) * n + j-1]:=t;
        END;

      { Column j of C~ has been computed }
      work[j]:=0.0;  { reset }
    END; { for j }

  { The matrix C~ has been computed. We overwrite the original matrix C with
    C~. }
  FOR i:=1 TO ncstr DO
    FOR j:=1 TO n DO
      C[ ind(dC,i,j) ] := work[ CTilde_p + (i-1) *n + j -1 ];
  { All real temporary working space can now be reused. We have to compute
    h-C~ * c_1 }

  FOR i:=1 TO ncstr DO
    BEGIN
      t:=0.0;
      FOR j:=1 TO n DO
        t:=t + C[ ind(dC,i,j) ] * b[j];
      h[i]:=h[i] -t;
    END;
  { h is overwritten by \tilde(h) }

  { The linear least squares problem with linear inequality constraints (LSI)
    has been transformed to an equivalent least distance problem (LDP). We
    now solve this LDP. }

  LDP ( C, dC, ncstr, n, h[1] , x[1], t, work[1], pivot[n+1],result);
  Mode:=result;
  IF result <> 1 THEN
    GOTO return;

  { Now comes the backtransformation to the solution of the original problem }
  FOR j:=1 TO n DO
    BEGIN
      work[j]:=x[j] + b[j];  { z * Q_1^T * b }
    END;

  work_p:=n+1;
  result:=SolveUTS ( A,dA,n,work[work_p], work[1]);
  { Solve R * y = z + Q_1^T * b, work[n+1] .. work[2*n] contains solution y }

  FOR j:=1 TO n DO
    x[pivot[j]]:= work[work_p + j - 1];  { x = P * y }

return:

END; { LSI }

{-----------------------------------------------------------------------------}

PROCEDURE LDP ( VAR Gp; dG, m, n:integer; VAR hp ; VAR xp;
                VAR XNORM:real; VAR wp; VAR INDEXp; VAR MODE:integer);

{ Least Distance Programming
  Lawson; Hanson: Solving Least Squares Problems                              }

TYPE vector=ARRAY[1..1] OF real;
     vectorint=ARRAY[1..1] OF integer;

VAR  G                            : vector ABSOLUTE Gp;
     h                            : vector ABSOLUTE hp;
     x                            : vector ABSOLUTE xp;
     w                            : vector ABSOLUTE wp;
     index                        : vectorint ABSOLUTE indexp;

     i,j,iw,iz,iy,i_f,iwdual,np1  : integer;
     rnorm,fac                    : real;

BEGIN
  IF n<=0 THEN
    BEGIN
      mode:=2;           { bad dimension }
      Exit;
    END
  ELSE
    BEGIN
      FOR j:=1 TO n DO
        x[j]:=0.0;
      xnorm:=0.0;
      IF m<=0 THEN
        BEGIN
          mode:=1;       { nothing to do, succesful }
          Exit;
        END;

{ the declared dimension of w must be at least (n+1) * (m+2) + 2 * m

  First (n+1)*m locations of w       matrix E for problem NNLS
  next    n+1   locations of w       vector f for problem NNLS
  next    n+1   locations of w       vector z for problem NNLS
  next     m    locations of w       vector y for problem NNLS
  next     m    loactions of w       vector wdual for problem NNLS

  Copy G^T into first n rows and m columns of E
  Copy H^T into row n+1 of E                                                  }

      iw:=0;
      FOR j:=1 TO n DO
        BEGIN
          FOR i:=1 TO m DO
            BEGIN
              Inc(iw);
              W[iw]:=G[ ind(dG,i,j) ];     { G^T }
              { G^T must be stored row by row to confirm with
                Pascals memory management }
            END;
        END;

      FOR i:=1 TO m DO
        BEGIN
          Inc(iw);
          W[iw]:=H[i];     { h^T }
       END;

      i_f:=iw+1;
      FOR i:=1 TO n DO                { store f in W }
        BEGIN
          Inc(iw);
          W[iw]:=0.0;
        END;
      W[iw+1]:=1.0;                   { last element of f }

      np1:=n+1;
      iz:=iw+2;
      iy:=iz+np1;
      iwdual:=iy+m;         { set pointers to the relevant vectors in w }

      NNLS ( w, m ,np1 ,m,w[i_f],w[iy],rnorm,w[iwdual],w[iz],index,mode);

    {  leading dimension of w is not np1, but m (number of columns of E)
       This is a difference to Fortran (column by column storage) }

      IF mode <> 1 THEN Exit;  { NNLS has failed }

      IF rnorm > 0.0 THEN      { residual sum <> 0 }
        BEGIN

          fac:=1.0;
          iw:=iy-1;
          FOR i:=1 TO m DO
            BEGIN
              Inc(iw);
              fac:=fac-H[i]*w[iw];
            END;

          IF diff(1.0+fac,1.0) > 0.0 THEN
            BEGIN        { fac greater than underflow threshhold }
              fac:=1.0/fac;
              FOR j:=1 TO n DO
                BEGIN
                  iw:=iy-1;
                  FOR i:=1 TO m DO
                    BEGIN
                      Inc(iw);
                      x[j]:=x[j] + G[ ind(dG,i,j) ] * w[iw];
                    END;
                  x[j]:=x[j] * fac;
                END;

              FOR j:=1 TO n DO
                xnorm:=xnorm+Sqr(x[j]);
              xnorm:=Sqrt(xnorm);
            END;
        END;
    END; {else}
END; { LDP }

{-----------------------------------------------------------------------------}

PROCEDURE NNLS ( VAR Ap; dA, m, n:integer; VAR bp ; VAR xp;
                VAR RNORM:real; VAR wp; VAR zzp; VAR INDEXp; VAR MODE:integer);

TYPE vector = ARRAY[1..1] OF real;
     vectorint = ARRAY[1..1] OF integer;

LABEL main, findwmax, reject, accept, start2ndloop, setxzero,
      END2ndloop, finished;

VAR  A                        : vector ABSOLUTE Ap;
     b                        : vector ABSOLUTE bp;
     x                        : vector ABSOLUTE xp;
     w                        : vector ABSOLUTE wp;
     zz                       : vector ABSOLUTE zzp;
     INDEX                    : vectorint ABSOLUTE INDEXp;
     sm,wmax,asave,up,dummy,unorm,ztest,alpha,t,cc,ss
                              : real;
     i,j,ii,jj,l,iter,itmax,iz,iz1,iz2,izmax,ip,nsetp,npp1,jz
                              : integer;

  PROCEDURE triangular;
  { intern subroutine to solve a triangular system
    all variables are global }

  VAR l,ip,ii,jj       : integer;

  BEGIN
    FOR l:=1 TO nsetp DO
      BEGIN
        ip:=nsetp+1-l;
        IF l <> 1 THEN
          BEGIN
            FOR ii:=1 TO ip DO
              zz[ii]:=zz[ii] - A[ ind(dA,ii,jj) ] * zz[ip+1];
          END;
        jj:=index[ip];
        zz[ip]:=zz[ip] / A[ ind(dA,ip,jj) ];
      END;
  END; { triangular }


BEGIN
  mode:=1;
  IF NOT ((m > 0) AND (n > 0)) THEN
    BEGIN
      mode:=2;      { dimensions are wrong }
      Exit;
    END
  ELSE
    BEGIN           { dimensions ok }
      iter:=0;      { number of iterations }
      itmax:=3*n;   { max. number of iterations }

      FOR i:=1 TO n DO   { initialize the arrays index and x }
        BEGIN
          x[i]:=0.0;
          index[i]:=i;
        END;

      iz2:=n;
      iz1:=1;
      nsetp:=0;
      npp1:=1;

main: { main loop starts here }

      { quit if all coefficients are already in the solution or if m columns
        of A have been triangularized }
      IF ( iz1 > iz2 ) OR ( nsetp >= m ) THEN
        GOTO finished;

      { compute components of the dual vector w }
      FOR iz:=iz1 TO iz2 DO
        BEGIN
          j:=index[iz];
          sm:=0.0;
          FOR l:=npp1 TO m DO
            sm:=sm + A[ ind(dA,l,j) ] * b[l];
          w[j]:=sm;
        END;

findwmax:
      { find largest positive w[j] }
      wmax:=0.0;
      FOR iz:=iz1 TO iz2 DO
        BEGIN
          j:=index[iz];
          IF ( w[j] > wmax ) THEN
            BEGIN
              wmax:=w[j];
              izmax:=iz;
            END;
        END;

      { if wmax <= 0 goto termination
        This indicates satisfaction of the Kuhn-Tucker conditions }
      IF wmax <= 0.0 THEN
        GOTO finished;

      iz:=izmax;
      j:=index[iz];

      { The sign of w[j] is ok for j to be moved to set P.
        Begin the transformation and check new diagonal to avoid near
        linear dependence. }
      asave:=A[ ind(dA,npp1,j) ];

      H12(1, npp1, npp1 + 1, m, A[ ind(dA,1,j) ], dA, up, dummy, 1, 1, 0 );
      { The storage increment between elements in one column of A must
        be dA, not 1 }

      unorm:=0.0;
      IF nsetp > 0 THEN
        BEGIN
          FOR l:=1 TO nsetp DO
            unorm:=unorm + Sqr( A[ ind(dA,l,j) ] );
          unorm:=Sqrt(unorm);
        END;

      { Test for near linear dependence }
      IF diff( unorm + Abs( A[ ind(dA,npp1,j) ] * 0.01 ), unorm) > 0.0 THEN
        BEGIN
          { Column j is sufficiently independent. Copy b into zz, update
            zz and solve for ztest ( proposed new value for x[j] ) }
          FOR l:=1 TO m DO
            zz[l]:=b[l];

          H12(2,npp1,npp1+1,m,A[ ind(dA,1,j) ],dA,up,zz,1,1,1);
          { The storage increment between elements in one column of A must
            be dA, not 1. }

          ztest:=zz[npp1] / A[ ind(dA,npp1,j) ];

          IF ztest <= 0.0 THEN
            GOTO reject
          ELSE GOTO accept;
        END { sufficiently independent column }

      ELSE
        GOTO reject;

reject:
        { reject j as a candidate to be moved from set Z to set P.
          Restore A[npp1,j], set w[j]=0, and loop back to test dual
          coefficients again. }

        A[ ind(dA,npp1,j) ]:=asave;
        w[j]:=0.0;
        GOTO findwmax;

accept:

      { The index j=index[iz] has been selected to be moved from
        set Z to set P. Update b. Update indices. Apply Householder
        transformations to columns in new set Z, zero subdiagonal elements
        in column j, set w[j]:=0 }

      FOR l:=1 TO m DO
        b[l]:=zz[l];

      index[iz]:=index[iz1];
      index[iz1]:=j;
      Inc(iz1);
      nsetp:=npp1;
      Inc(npp1);

      IF iz1 <= iz2 THEN
        FOR jz:=iz1 TO iz2 DO
          BEGIN
            jj:=index[jz];
            H12(2,nsetp,npp1,m,A[ ind(dA,1,j) ],dA,up,A[ ind(dA,1,jj)],dA,1,1);
            { The storage increment between elements in a column of A
              should be dA, not 1. The increment between elements in the
              matrix C is dA, not 1. The increment between vectors in C is
              1, not dA. }
          END;

      IF nsetp <> m THEN
        BEGIN
          FOR l:=npp1 TO m DO
            A[ ind(dA,l,j) ]:=0.0;
        END;

      w[j]:=0.0;

      { solve the triangular system }
      triangular;


      { secondary loop begins here }
start2ndloop:
      Inc(iter);
      IF ( iter > itmax) THEN
        BEGIN
          mode:=3;
          Write('NNLS quitting on iteration count');
          GOTO finished;
        END;

      { See if all new coefficients are feasible. If not compute alpha }
      alpha:=2.0;
      FOR ip:=1 TO nsetp DO
        BEGIN
          l:=index[ip];
          IF zz[ip] <= 0.0 THEN
            BEGIN
              t:=-x[l] / (zz[ip]-x[l]);
              IF alpha > t THEN
                BEGIN
                  alpha:=t;
                  jj:=ip;
                END;
            END;
        END;

      { if all new constrained coefficients are feasible then alpha will
        still = 2. If so exit from secondary loop to main loop. }
      IF alpha=2.0 THEN
        GOTO END2ndloop;

      { Otherwise use alpha which will be between 0 and 1 to interpolate
        between the old x and the new zz }
      FOR ip:=1 TO nsetp DO
        BEGIN
          l:=index[ip];
          x[l]:=x[l] + alpha * (zz[ip] - x[l]);
        END;

      { Modify A and b and the index arrays to move coefficient i from
        set P to set Z }
      i:=index[jj];

setxzero:
      x[i]:=0.0;

      IF jj <> nsetp THEN
        BEGIN
          Inc(jj);
          FOR j:=jj TO nsetp DO
            BEGIN
              ii:=index[j];
              index[j-1]:=ii;
              G1(A[ind(dA,j-1,ii)],A[ind(dA,j,ii)],cc,ss,A[ind(dA,j-1,ii)]);
              A[ind(dA,j,ii)]:=0.0;
              FOR l:=1 TO n DO
                BEGIN
                  IF l <> ii THEN
                    G2(cc,ss,A[ind(dA,j-1,l)],A[ind(dA,j,l)]);
                END;
              G2(cc,ss,B[j-1],B[j]);
            END;
        END;
      npp1:=nsetp;
      Dec(nsetp);
      Dec(iz1);
      index[iz1]:=i;

      { See if the remaining coefficients in set P are feasible. They
        should be because on the way alpha was determined.
        If any are infeasible it is due to round-off error. Any that are
        nonpositive will be set to zero and moved from set P to set Z. }
      FOR jj:=1 TO nsetp DO
        BEGIN
          i:=index[jj];
          IF x[i] <= 0.0 THEN
            GOTO setxzero;
        END;

      { Copy b into zz. Then solve again and loop back }
      FOR i:=1 TO m DO
        zz[i]:=b[i];

      triangular;
      GOTO start2ndloop;

END2ndloop:
      { end of secondary loop }
      FOR ip:=1 TO nsetp DO
        BEGIN
          i:=index[ip];
          x[i]:=zz[ip];
        END;
      GOTO main;   { all new coefficients are positiv, goto main loop }

finished:
      { come here for termination, compute the norm of the residual vector }
      sm:=0.0;
      IF npp1 <= m THEN
        BEGIN
          FOR i:=npp1 TO m DO
            sm:=sm + Sqr ( B[i] )
        END
      ELSE
        BEGIN
          FOR j:=1 TO n DO
            w[j]:=0.0;
        END;
      rnorm:=Sqrt(sm);

    END; { else }
END; { NNLS }

{-----------------------------------------------------------------------------}

PROCEDURE H12 ( mode,lpivot,l1,m :integer; VAR U_p ; iue :integer;
                VAR up:real; VAR Cp; ice,icv,ncv : integer);

LABEL mode2, return;

TYPE vector=ARRAY[1..1] OF real;

VAR  U                        : vector ABSOLUTE U_p;
     C                        : vector ABSOLUTE Cp;

     sm,b,cl,clinv            : real;
     i,j,i2,i3,i4,incr        : integer;

BEGIN
{ The indizes refering to U have been changed from the original
  U[1,j] to U[ ind(iue,j,1) ] to match the Pascal memory management.
  The code has been tested under this circumstances.
  Warning: You must adapt the calls of H12.
}
  IF NOT(( 0 >= LPIVOT ) OR ( LPIVOT >= L1 ) OR ( L1 > M )) THEN
    BEGIN
      cl:=Abs( U[ind(iue,lpivot,1)] );

      IF mode=1 THEN
        BEGIN      { construct Householder transformation }
          FOR j:=l1 TO m DO
            cl:=max_real(Abs(U[ind(iue,j,1)]),cl);{ search for maximum in row }
          IF cl = 0.0 THEN
            GOTO return; { nothing to do }

          clinv:=1.0/cl;        { stable computation of a sum of squares }
          sm:=Sqr( U[ind(iue,lpivot,1)] * clinv );
          FOR j:=l1 TO m DO
            sm:=sm + Sqr( U[ ind(iue,j,1) ] * clinv );
          cl:=cl * Sqrt(sm);
          { cl now contains ( v_lpivot^2 + sum_(i=l1)^(m) v_i^2 )^0.5 }

          IF U[ ind(iue,lpivot,1) ] > 0.0 THEN
            cl:=-cl;
          up:= U[ ind(iue,lpivot,1) ] - cl;
          U[ ind(iue,lpivot,1) ]:=cl;
          GOTO mode2;
        END;      { Householder transformation has been constructed }

      IF cl <= 0.0 THEN
        GOTO return;

mode2:
      IF ncv > 0 THEN
        BEGIN { apply Householder application }

          b:=up*U[ ind(iue,lpivot,1) ];
          IF b>= 0.0 THEN
            GOTO return;

          b:=1.0/b;

          i2:=1 - icv + ice * (lpivot-1);
          incr:=ice * (l1-lpivot);
          FOR j:=1 TO ncv DO      { for each vector apply the transformation }
            BEGIN
              i2:=i2+icv;
              i3:=i2+incr;
              i4:=i3;
              sm:=C[ i2 ] * UP;           { c[lpivot,j] * up }
              FOR i:=l1 TO m DO
                BEGIN
                  sm:=sm + C[i3] * U[ ind(iue,i,1) ];
                  i3:=i3+ice;
                END;
              { sm = c[lpivot,j] * up + sum_(i=l1)^m (c[i,j] * v[i]) }
              IF sm <> 0.0 THEN
                BEGIN
                  sm:=sm*b;
                  C[i2]:=C[i2] + sm * up;
                  { c[lpivot,j] := c[lpivot,j] + sm * up }

                  FOR i:=l1 TO m DO
                    BEGIN
                      C[i4]:=C[i4] + sm * U[ ind(iue,i,1) ];
                      { c[i,j] := c[i,j] + sm * v[i] }

                      i4:=i4+ice;
                    END;
                END; { if sm <> 0 }
            END; { for j:=1 to ncv }
        END; { apply Householder transformation }
    END;
  return:    { exit the procedure }
END;  { H12 }

{-----------------------------------------------------------------------------}

PROCEDURE G1 ( VAR a,b,CosG,SinG,sig :real);
VAR xr,yr : real;
BEGIN
  IF Abs(a) > Abs (b) THEN
    BEGIN
      xr:= b/a;
      yr:=Sqrt(1.0 + Sqr(xr));
      IF a >= 0.0 THEN
        CosG:=1.0/yr
      ELSE
        CosG:=-1.0/yr;
      SinG:=CosG*xr;
      sig:=Abs(a)*yr;
    END
  ELSE
    BEGIN
      IF b<> 0.0 THEN
        BEGIN
          xr:=a/b;
          yr:=Sqrt(1.0+Sqr(xr));
          IF b>= 0.0 THEN
            SinG:=1.0/yr
          ELSE
            SinG:=-1.0/yr;
          CosG:=SinG*xr;
          sig:=Abs(b)*yr;
        END
      ELSE
        BEGIN
          CosG:=0.0;
          SinG:=1.0;
          sig:=0.0;
        END;
    END;
END; { G1 }

{-----------------------------------------------------------------------------}

PROCEDURE G2 ( VAR CosG,SinG,x,y :real);
VAR xr : real;
BEGIN
  xr:=CosG * x + SinG * y;
  y:=(-SinG) * x + CosG * y;
  x:=xr;
END; { G2 }

{-----------------------------------------------------------------------------}

FUNCTION DIFF (x,y : real) : real;
BEGIN
  diff:=x-y;
END; { DIFF }

{-----------------------------------------------------------------------------}
END. { UNIT LAWSON }
