UNIT LINPACK;
{ This unit contains procedures from LINPACK for the QR-factorization
  of a matrix and the solution of linear least squares problems               }
{$R-}
INTERFACE

USES BLAS;        { Basic Linear Algebra Subroutines                          }

{-----------------------------------------------------------------------------}

PROCEDURE QRDC(VAR Ap;dA,n,p:integer;VAR jpvtp;VAR qrauxp,workp;job:integer);

{ Computation of the QR-Decomposition of a (n,p)-Matrix A with
  Householder transformations

  On entry:

  A:    Vector of the dimension n*p, contains the elements of the Matrix;
  dA:   Integer, leading dimension of A;
  jpvt: Array[1..p] of integer, contains integers, that control the selection
        of the pivot columns

        if jpvt(j)>0, column j is an initial column of A,
        if jpvt(j)=0, column j is an free column
        if jpvt(j)<0, column j is an final column of A,

        before the decomposition is computed, initial and final columns are
        moved to the beginning and to the end of A
        At the k-th stage of the reduction, if A(k) is occupied by a free
        column it is interchanged with the free columns of largest reduced
        norm
        jpvt is not referenced if job=0
  work: Array[1..p] of real; work array
        work is not referenced if job=0
  job:  Integer,
        if job =0 no pivoting is done
        if job<>0 pivoting is done

  On return

  A:    contains in its upper triangle the upper triangular matrix R of
        the QR decomposition (with permuted columns!)
        below its diagonal A contains information for the reconstruction
        of the Householder transformations
  qraux:contains further information to recover the Householder trans-
        formations
  jpvt: jpvt(j) contains the permutation index

  BLAS-Subroutines: AXPY, DOT, SCAL, NRM2, SWAP.                              }

{-----------------------------------------------------------------------------}

FUNCTION QRSL(VAR Ap;dA,n,k,job:integer;VAR qrauxp,yp,qyp,qtyp,bp,rp,xbp):integer;

{ QRSL applies the output of QRDC to compute coordinate transformations,
  projections, and least squares solutions

  On entry:

  A:     Array[1..n,1..dA] of real, contains the output of QRDC
  dA:    Integer, leading dimension of A
  n:     Integer, number of rows of A (the same value as in QRDC;
  k:     Integer, number of columns, k<=n,k<=p;
  qraux: Array[1..p] of real, contains auxiliary output of QRDC
  y:     Array[1..n] of real, y is to be manipulated by QRSL
  job:   Integer, decimal expansion abcde

         if a <> 0 compute QY
         if b,c,d or e <> 0 compute QTY
         if c <> 0 compute b
         if d <> 0 compute residual vector
         if e <> 0 compute xb

  On return:

  qy :   array[1..n] of real, contains Q*Y
  qty:   array[1..n] of real, contains Trans(Q)*Y
  b:     array[1..k] of real, contains least squares solution of
         || y - A*b ||
  r:     array[1..n] of real, contains least squares residual y - A*b
  xb:    array[1..n] of real, contains the least squares approximation A*b

  function value:
  qrsl:  zero unless the computation of b has been requested and R is
         exactly singular
         in this case it is the index of the first zero diagonal element of R

  the parameters qy,qty,b,r and xb are not referenced if their computation
  is not requested and in this case can replaced by dummy variables

  BLAS-Subroutines: COPY, DOT, AXPY.                                          }

{-----------------------------------------------------------------------------}

FUNCTION TRSL(VAR Tp;dT,n:integer;VAR bp;job:integer):integer;

{ TRSL solves systems of the form
                          T  * x = b
               or
                    trans(T) * x = b
  where T is a triangular matrix of order n

  On entry:

  T:     Array[1..dT,1..n] of real, the zero elements of T are not referenced
  dT:    Integer, leading dimension of T
  n:     Integer, order of the system
  b:     Array[1..n] of real, right hand side of the system
  job:   Integer, specifies what kind of system is to be solved

         00       solve        T*x=b     T lower triangular
         01       solve        T*x=b     T upper triangular
         10       solve trans(T)*x=b     T lower triangular
         11       solve trans(T)*x=b     T upper triangular

  On return:

  b:     Array[1..n] of real, contains the solution if TRSL=0

  function value:
  TRSL:  integer, if TRSL=0 the system is nonsingular
         otherwise TRSL contains the index of the first zero diagonal element

                                                                              }

{-----------------------------------------------------------------------------}
IMPLEMENTATION

{-----------------------------------------------------------------------------}

FUNCTION ind (LDA,i,j : integer) : integer;
{ Function for index computations
  Element a[i,j] corresponds to  vector [ (i-1) * LDA+j ]
  ( Arrays are stored row by row in Pascal )
  A must be declared as Array[1..m,1..LDA]                                    }

BEGIN
  ind:=(i-1)*LDA+j;
END; { ind }

{-----------------------------------------------------------------------------}

PROCEDURE QRDC(VAR Ap;dA,n,p:integer;VAR jpvtp;VAR qrauxp,workp;job:integer);

LABEL L1,L2;
TYPE vector=ARRAY[1..1] OF real;
     vectorint=ARRAY[1..1] OF integer;

VAR A                         : vector ABSOLUTE Ap;
    qraux                     : vector ABSOLUTE qrauxp;
    work                      : vector ABSOLUTE workp;
    jpvt                      : vectorint ABSOLUTE jpvtp;
    maxnrm,nrmAl,t,tt,b       : real;
    j,jp,l,lp1,maxj,lup,pl,pu : integer;
    swapj,negj                : boolean;

BEGIN
  pl:=1;
  pu:=0;
  
  { Pivoting, Rearrange the columns according to jpvt }
  IF (job<>0) THEN
    BEGIN
      FOR j:=1 TO p DO
        BEGIN
          swapj:=(jpvt[j]>0) ;
          negj:= (jpvt[j]<0) ;
          jpvt[j]:=j;
          IF (negj) THEN jpvt[j]:=-j;
          IF (swapj) THEN
            BEGIN
              IF (j<>pl) THEN Swap(n,A[pl],dA,A[j],dA);
              jpvt[j]:=jpvt[pl];
              jpvt[pl]:=j;
              pl:=pl+1;
            END;
        END;
      
      pu:=p;
      FOR j:=p DOWNTO 1 DO
        BEGIN
          IF (jpvt[j]<0) THEN
            BEGIN
              jpvt[j]:=-jpvt[j];
              IF (j<>pu) THEN
                BEGIN
                  Swap(n,A[pu],dA,A[j],dA);
                  jp:=jpvt[pu];
                  jpvt[pu]:=jpvt[j];
                  jpvt[j]:=jp;
                END;
              pu:=pu-1;
            END;
        END;
    END;
  
  { Compute the norms of the free columns }
  FOR j:=pl TO pu DO
    BEGIN
      qraux[j]:=NRM2(n,A[j],dA);
      work[j]:=qraux[j]
    END;
  
  { Perform the Householder reduction of A }
  IF (n<=p) THEN lup:=n
  ELSE lup:=p;
  
  FOR l:=1 TO lup DO
    BEGIN
      IF (l>=pl) AND (l<pu) THEN
        BEGIN
          maxnrm:=0.0;
          maxj:=l;
          FOR j:=l TO pu DO
            BEGIN
              IF (qraux[j]>maxnrm) THEN
                BEGIN
                  maxnrm:=qraux[j];
                  maxj:=j
                END;
            END;
          IF (maxj<>l) THEN
            BEGIN
              Swap(n,A[l],dA,A[maxj],dA);
              qraux[maxj]:=qraux[l];
              work[maxj]:=work[l];
              jp:=jpvt[maxj];
              jpvt[maxj]:=jpvt[l];
              jpvt[l]:=jp;
            END;
        END;

      qraux[l]:=0.0;
      IF (l<>n) THEN
        BEGIN
          { compute the Householder transformation for column l }
          nrmAl:=NRM2(n-l+1,A[l+(l-1)*dA],dA);
          IF (nrmAl=0.0) THEN GOTO L2;
          IF (A[l+(l-1)*dA]<0.0) THEN nrmAl:=-nrmAl;
          b:=1.0/nrmAl;
          SCAL(n-l+1,b,A[l+(l-1)*dA],dA);
          A[l+(l-1)*dA]:=1.0+A[l+(l-1)*dA];
          
          { Apply the transformation to the remaining columns
          updating the norms                                }
          lp1:=l+1;
          FOR j:=lp1 TO p DO
            BEGIN
              t:=-DOT(n-l+1,A[l+(l-1)*dA],dA,A[j+(l-1)*dA],dA)/A[l+(l-1)*dA];
              AXPY(n-l+1,t,A[l+(l-1)*dA],dA,A[j+(l-1)*dA],dA);
              IF (j>=pl) AND (j<=pu) THEN
                BEGIN
                  IF (qraux[j]=0.0) THEN GOTO L1;
                  tt:=1.0-Sqr(Abs(A[j+(l-1)*dA])/qraux[j]);
                  IF (tt<0.0) THEN tt:=0.0;
                  t:=tt;
                  tt:=1.0+0.05*tt*Sqr(qraux[j]/work[j]);
                  IF (tt=1.0) THEN
                    BEGIN
                      qraux[j]:=NRM2(n-l,A[j+l*dA],dA);
                      work[j]:=qraux[j];
                    END
                  ELSE qraux[j]:=qraux[j]*Sqrt(t);
                  L1: 
                END;
            END;
          
          { Save the transformation }
          qraux[l]:=A[l+(l-1)*dA];
          A[l+(l-1)*dA]:=-nrmAl;
          
          L2:
        END;
    END;
END;

{-----------------------------------------------------------------------------}

FUNCTION QRSL(VAR Ap;dA,n,k,job:integer;VAR qrauxp,yp,qyp,qtyp,bp,rp,xbp):integer;

LABEL return;
TYPE vector=ARRAY[1..1] OF real;
VAR A                  : vector ABSOLUTE Ap;
    qraux              : vector ABSOLUTE qrauxp;
    y                  : vector ABSOLUTE yp;
    qy                 : vector ABSOLUTE qyp;
    qty                : vector ABSOLUTE qtyp;
    b                  : vector ABSOLUTE bp;
    r                  : vector ABSOLUTE rp;
    xb                 : vector ABSOLUTE xbp;
    t,temp             : real;
    i,j,jj,ju,kp1      : integer;
    cqy,cqty,cb,cr,cxb : boolean;

BEGIN
  qrsl:=0;

  { what is to be computed ? }
  cqy:=(Trunc(job/10000)<>0);
  cqty:=(job MOD 10000)<>0;
  cb:=(Trunc((job MOD 1000)/100)<>0);
  cr:=(Trunc(job MOD 100)/10) <>0 ;
  cxb:=(job MOD 10)<>0 ;

  IF (k<=n-1) THEN ju:=k
  ELSE ju:=n-1;

  { special action when n=1 }
  IF (ju=0) THEN
    BEGIN
      IF (cqy) THEN qy[1]:=y[1];
      IF (cqty) THEN qty[1]:=y[1];
      IF (cxb) THEN xb[1]:=y[1];
      IF (cb) THEN
        BEGIN
          IF (Abs(A[1])<0.1E-12) THEN
            qrsl:=1
          ELSE b[1]:=y[1]/A[1];
        END;
      IF (cr) THEN r[1]:=0.0;
      GOTO return;
    END;

  { set up to compute QY or TRANS(Q)*Y }
  IF (cqy) THEN Copy1(n,y,qy);
  IF (cqty) THEN Copy1(n,y,qty);

  IF (cqy) THEN
    BEGIN
      FOR j:=ju DOWNTO 1 DO
        BEGIN
          IF (qraux[j]<>0.0) THEN
            BEGIN
              i:=j+dA*(j-1);
              temp:=A[i];
              A[i]:=qraux[j];
              t:=-DOT(n-j+1,A[i],dA,qy[j],1)/A[i];
              AXPY(n-j+1,t,A[i],dA,qy[j],1);
              A[i]:=temp;
            END;
        END;
    END;

  { compute Trans(q)*y }
  IF (cqty) THEN
    BEGIN
      FOR j:=1 TO ju DO
        BEGIN
          IF (qraux[j]<>0.0) THEN
            BEGIN
              i:=j+dA*(j-1);
              temp:=A[i];
              A[i]:=qraux[j];
              t:=-DOT(n-j+1,A[i],dA,qty[j],1)/A[i];
              AXPY(n-j+1,t,A[i],dA,qty[j],1);
              A[i]:=temp;
            END;
        END;
    END;

  { set up to compute b,r or xb }
  IF (cb) THEN Copy1(k,qty,b);
  kp1:=k+1;
  IF (cxb) THEN Copy1(k,qty,xb);
  IF (cr AND (k<n)) THEN Copy1(n-k,qty[kp1],r[kp1]);
  IF (cxb AND (kp1 <= n)) THEN
    BEGIN
      FOR i:=kp1 TO n DO
        xb[i]:=0.0;
    END;

  IF (cr) THEN
    BEGIN
      FOR i:=1 TO k DO
        r[i]:=0.0;
    END;

  IF (cb) THEN
    BEGIN
      FOR jj:=1 TO k DO
        BEGIN
          j:=k-jj+1;
          i:=j+dA*(j-1);
          IF (Abs(A[i])<1.0E-12) THEN
            BEGIN
              qrsl:=j;
              WriteLn ('R singular!');
              GOTO return;
            END
          ELSE
            BEGIN
              b[j]:=b[j]/A[i];
              IF j<>1 THEN
                BEGIN
                  t:=-b[j];
                  AXPY(j-1,t,A[j],dA,b,1);
                END;
            END;
        END;
    END;

  IF (cr OR cxb) THEN
    BEGIN
      FOR jj:=1 TO ju DO
        BEGIN
          j:=ju-jj+1;
          i:=j+dA*(j-1);
          IF (qraux[j]<>0) THEN
            BEGIN
              temp:=A[i];
              A[i]:=qraux[j];
              IF (cr) THEN
                BEGIN
                  t:=-DOT(n-j+1,A[i],dA,r[j],1)/A[i];
                  AXPY(n-j+1,t,A[i],dA,r[j],1);
                END;
              IF (cxb) THEN
                BEGIN
                  t:=-DOT(n-j+1,A[i],dA,xb[j],1)/A[i];
                  AXPY(n-j+1,t,A[i],dA,xb[j],1);
                END;
              A[i]:=temp;
            END;
        END;
    END;

  return:
END;

{-----------------------------------------------------------------------------}

FUNCTION TRSL(VAR Tp;dT,n:integer;VAR bp;job:integer):integer;

LABEL return;
TYPE vector=ARRAY[1..1] OF real;
VAR T                  : vector ABSOLUTE Tp;
    b                  : vector ABSOLUTE bp;
    i, j, task         : integer;
    s                  : real;
BEGIN
  { check for zero diagonal elements }
  FOR j:=1 TO n DO
    IF T[ind(dT,j,j)]=0.0 THEN
      BEGIN
        TRSL:=j;
        GOTO return;
      END;

  { determine the task }
  IF job MOD 10 = 0 THEN
    task := 1
  ELSE
    task := 2;
  IF Trunc((job MOD 100) / 10) <> 0 THEN
    task := task+2;

  { goto relevant task }
  CASE task OF
    1: BEGIN       { job=00,  solve T*x=b, T lower triangular }
         FOR i := 1 TO n DO
           BEGIN
             s := b[i];
             FOR j := 1 TO i-1 DO
               s := s - T[ind(dT,i,j)] * b[j];
             b[i] := s / T[ind(dT,i,i)];
           END;
       END;
    2: BEGIN       { job=01,  solve T*x=b, T upper triangular }
         FOR i := n DOWNTO 1 DO
           BEGIN
             s := b[i];
             FOR j := i+1 TO n DO
               s := s - T[ind(dT,i,j)] * b[j];
             b[i] := s / T[ind(dT,i,i)];
           END;
       END;
    3: BEGIN       { job=10,  solve trans(T)*x=b, T lower triangular }
         FOR i := n DOWNTO 1 DO
           BEGIN
             s := b[i];
             FOR j := i+1 TO n DO
               s := s - T[ind(dT,j,i)] * b[j];
             b[i] := s / T[ind(dT,i,i)];
           END;
       END;
    4: BEGIN       { job=11 ,  solve trans(T)*x=b , T upper triangular }
         FOR i:=1 TO n DO
           BEGIN
             s:=b[i];
             FOR j:=1 TO i-1 DO
               s:=s-T[ind(dT,j,i)]*b[j];
             b[i]:=s/T[ind(dT,i,i)];
           END;
       END;
  END;

  TRSL:=0;
  return:
END; { TRSL }

{-----------------------------------------------------------------------------}

END.