UNIT Smooth;
{ procedures for the computation of smoothing polynomial B-Splines            }

INTERFACE
  USES def,       { definitions of constants, types and global variables      }
       givens,    { low level procedures for GIVENS plane rotations           }
       math,      { low level mathematical subroutines                        }
       bspline;   { procedures to compute B-Splines and l2-approximations     }

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Derivative_Matrix(
           VAR t  : knotsequence;     { knotsequence                          }
               n,                     { number of B-Splines                   }
               k,                     { order of B-Splines                    }
               r  : integer;          { order of derivative                   }
           VAR D  : SmoothingMatrix   { derivative matrix                     }
           );
{ Computation of a matrix D, so that
  \alpha^(r) = D_r \alpha are the coefficients of the r-th order derivative
  of the given spline of order k with coefficients \alpha belonging to the
  knotsequence t

  D is an (n-r) x n matrix with bandwidth r+1. Therefore we can store the
  upper triangular matrix D row by row in an matrix with maximal bandwidth k

  D[ row, col] = (D_r)_(i,j)       with  i = row + r, j = row + col -1        }

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Smoothing_Matrix(
           VAR t  : knotsequence;     { knotsequence                          }
               n,                     { number of B-Splines                   }
               k,                     { order of B-Splines                    }
               r  : integer;          { order of derivative in smoothing term }
           VAR S  : SmoothingMatrix   { approximation of smoothing term       }
           );
{ Computation of a matrix S, so that
  \left\|  S \alpha \right|^2 is an upper bound for the smoothing term
  \int_a^b [ s^(r) (x)]^2 dx

  This approximation of the smoothing term results from replacing the
  L_2 - norm by its discrete equivalent l_2.

  \left\| S \alpha \right|^2 =
  \sum_(i=r+1)^(n) ( \alpha_i^(r) )^2 (t[i+k-r] - t[i]) / (k-r)

  S is an (n-r) x n matrix with bandwidth r+1. Therefore we can store the
  upper triangular matrix S row by row

  S[ row, col] = \sqrt( (t[i+k-r] - t[i])/(k-r) ) (D_r)_(i,j)
  with  i = row + r
        j = row + col -1                                                      }

{-----------------------------------------------------------------------------}

PROCEDURE GIVENS_factorization_of_RBS(
               n,                     { number of B-Splines                   }
               k,                     { order of B-Splines                    }
               r  : integer;          { order of derivative in smoothing term }
              mu  : real;             { smoothing parameter                   }
          VAR RB  : leastsquarematrix;{ transformed B-Spline matrix           }
          VAR Rmu : leastsquarematrix;{ upper triangular part of QR-factor.   }
          VAR S   : SmoothingMatrix;  { approximation of smoothing term       }
          VAR c   : coefficients;     { transformed datavalues                }
          VAR G2  : coefficients;     { second part of right hand side        }
          VAR d   : coefficients;     { transformed right hand side           }
          VAR RES : real              { residual sum                          }
          );
{ produces GIVENS factorization of

  [       R_B     ]          [ c  ]
  [               ] \alpha - [    ]
  [  \sqrt(\mu) S ]          [ G2 ]

  R_B and c result from a GIVENS factorization of the B-Spline matrix B

  On Output, Rmu contains the upper triangular factor and d contains
  the transformed right side (c,G2)

  RES contains the residual sum                                               }

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Smoothing_Spline(
             m,                       { number of datavalues                  }
             n,                       { number of B-Splines                   }
             k    : integer;          { order of B-Splines                    }
        VAR t     : knotsequence;
             r    : integer;          { order of derivative in smoothing term }
            mu    : real;             { smoothing parameter                   }
        VAR tau   : dataabscissae;
        VAR y     : datavalues;
        VAR S     : SmoothingMatrix;  { Smoothing Matrix                      }
        VAR alpha : coefficients;     { to compute                            }
        VAR RES   : real              { residual sum                          }
        );

{ computes least squares solution alpha of

   || y - B * alpha ||^2 + mu * || S * alpha ||^2 --> Min !

  where || S * alpha ||^2 is an approximation of the smoothing term
  \int [ s^(r) (x) ]^2 dx

  by row by row GIVENS rotations                                              }

{-----------------------------------------------------------------------------}

PROCEDURE Get_G (
             m,                       { number of datavalues                  }
             n,                       { number of B-Splines                   }
             k,                       { order of B-Splines                    }
             r    : integer;          { order of derivative in smoothing term }
            mu    : real;             { smoothing parameter                   }
        VAR t     : knotsequence;
        VAR tau   : dataabscissae;
        VAR y     : datavalues;
        VAR alpha : coefficients;     { coefficients of smoothing spline      }
        VAR S     : SmoothingMatrix;
        VAR G     : smoothingvector   { residual vector of smoothing problem  }
        );
{ computes
               [ y ]    [      B     ]
      G(t) =   [   ] -  [            ] alpha
               [ 0 ]    [ sqrt(mu) S ]

 S is an approximation of the smoothing term
 G is of dimension m+n-r                                                      }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Smoothing_Matrix_Knots_approx(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR S     : SmoothingMatrix;
                 VAR v     : coefficients
                 );
{ computes derivative of smoothing matrix S_r with respect to t[kk] * alpha

  v:= S'(t) [e^kk] alpha \in \RealR^(n-r)
  The derivative will be approximated by finite differences.                  }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Smoothing_Matrix_Knots(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR v     : coefficients
                 );
{ computes derivative of smoothing matrix S_r with respect to t[kk] * alpha

  F:= S'(t) [e^kk] alpha \in \RealR^(n-r)                                     }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Derivative_Matrix_Knots_approx(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR D     : SmoothingMatrix;
                 VAR v     : coefficients
                 );
{ computes derivative of derivative matrix D_r with respect to t[kk] * alpha

  v:= D'(t) [e^kk] alpha \in \RealR^(n-r)
  The derivative will be approximated by finite differences.                  }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Derivative_Matrix_Knots(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR v     : coefficients
                 );
{ computes derivative of derivative matrix D_r with respect to t[kk] * alpha

  v:= D'(t) [e^kk] alpha \in \RealR^(n-r)                                     }

{-----------------------------------------------------------------------------}

PROCEDURE Get_STvec (
             m,                       { number of datavalues                  }
             n,                       { number of B-Splines                   }
             k,                       { order of B-Splines                    }
             r    : integer;          { order of derivative in smoothing term }
        VAR t     : knotsequence;
        VAR v, w  : coefficients;
        VAR S     : SmoothingMatrix
        );

{-----------------------------------------------------------------------------}

IMPLEMENTATION

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Derivative_Matrix(
           VAR t  : knotsequence;     { knotsequence                          }
               n,                     { number of B-Splines                   }
               k,                     { order of B-Splines                    }
               r  : integer;          { order of derivative                   }
           VAR D  : SmoothingMatrix   { derivative matrix                     }
           );

VAR  row, col, i, j, nu : integer;
     divisor, h : real;

BEGIN
  { Initialization }
  FOR row := 1 TO n DO
    FOR col := 1 TO k DO
      D[row,col] := 0.0;

  IF r = 0 THEN
    BEGIN
      FOR row:=1 TO n DO
      D[row,1]:=1.0;        { D_0 }
    END
  ELSE
    BEGIN
      { 1. Computation of D_1 }
      FOR row := 1 TO n-1 DO
        BEGIN
          i := row + 1;
          divisor :=  t[i+k-1] - t[i];
          D[row,1] := - (k-1) / divisor;
          D[row,2] := - D[row,1];
        END;
      { now D_1 is stored in S row by row }

      { 2. Multiplication with H_r L_r ... H_2 L_2 }

      FOR nu := 2 TO r DO
        FOR row := 1 TO n - nu DO
          BEGIN
            i := row + nu;
            h := t[i+k-nu] - t[i];
            h := (k-nu) / h;
            D[row,1] := - D[row,1] * h;
            FOR col := 2 TO nu DO
              D[row,col] := h * (D[row+1,col-1] - D[row,col]);
              D[row,nu+1] := h * D[row+1,nu];
          END;
      { now D_r is stored in D row by row }
    END; { if r>=1 }
END; { Compute_Derivative_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Smoothing_Matrix(
           VAR t  : knotsequence;     { knotsequence                          }
               n,                     { number of B-Splines                   }
               k,                     { order of B-Splines                    }
               r  : integer;          { order of derivative in smoothing term }
           VAR S  : SmoothingMatrix   { approximation of smoothing term       }
           );

VAR  row, col, i, j, nu : integer;
     divisor, h : real;

BEGIN
  { Initialization }
  FOR row := 1 TO n DO
    FOR col := 1 TO k DO
      S[row,col] := 0.0;

  IF r = 0 THEN
    BEGIN
      FOR row:=1 TO n DO
      S[row,1]:=1.0;        { D_0 }
    END
  ELSE
    BEGIN
      { 1. Computation of D_1 }
      FOR row := 1 TO n-1 DO
        BEGIN
          i := row + 1;
          divisor :=  t[i+k-1] - t[i];
          S[row,1] := - (k-1) / divisor;
          S[row,2] := - S[row,1];
        END;
      { now D_1 is stored in S row by row }

      { 2. Multiplication with H_r L_r ... H_2 L_2 }

      FOR nu := 2 TO r DO
        FOR row := 1 TO n - nu DO
          BEGIN
            i := row + nu;
            h := t[i+k-nu] - t[i];
            h := (k-nu) / h;
            S[row,1] := - S[row,1] * h;
            FOR col := 2 TO nu   DO
              S[row,col] := h * (S[row+1,col-1] - S[row,col]);
              S[row,nu+1] := h * S[row+1,nu];
          END;
      { now D_r is stored in S row by row }
    END; { if r>=1 }

  { 3. Multiplication with F_r }
  FOR row:=1 TO n-r DO
    BEGIN
      i := row + r;
      h := (t[i+k-r] - t[i]) / (k-r);
      h := Sqrt(h);
      FOR col:=1 TO r+1 DO
        S[row,col] := S[row,col] * h;
    END;

  { Computation done.}

END; { Compute_Smoothing_Matrix }

{-----------------------------------------------------------------------------}

PROCEDURE GIVENS_factorization_of_RBS(
               n,                     { number of B-Splines                   }
               k,                     { order of B-Splines                    }
               r  : integer;          { order of derivative in smoothing term }
              mu  : real;             { smoothing parameter                   }
          VAR RB  : leastsquarematrix;{ transformed B-Spline matrix           }
          VAR Rmu : leastsquarematrix;{ upper triangular part of QR-factor.   }
          VAR S   : SmoothingMatrix;  { approximation of smoothing term       }
          VAR c   : coefficients;     { transformed datavalues                }
          VAR G2  : coefficients;     { second part of right hand side        }
          VAR d   : coefficients;     { transformed right hand side           }
          VAR RES : real              { residual sum                          }
          );

VAR i,j,Length,iibeg,ii,iplusj_1       : integer;
    sqrt_mu,v,xi                       : real;
    u                                  : ARRAY[1..NBANDS] OF real;

BEGIN
  { initialization }
  FOR i:=1 TO n DO
    BEGIN
      d[i]:=0.0;
      FOR j:=1 TO k DO
        BEGIN
          Rmu[i,j]:=0.0;
        END;
    END;
  RES:=0.0;
  sqrt_mu:=Sqrt(mu);

  FOR i := 1 TO n DO
    BEGIN

      IF i <= (n-r) THEN
        iibeg := 1                { annihilate both rows of S and Rmu }
      ELSE
        iibeg := 2;               { annihilate remaining rows of Rmu  }

      FOR ii := iibeg TO 2 DO
        BEGIN
          Length := min( k, n-i+1);
          { Compute i-th row of (A,b), store it to (u,v) }

          IF ii = 1 THEN
            BEGIN
              FOR j := 1 TO Length DO
                u[ j] := sqrt_mu * S[i,j];
              v := G2[i];
            END
          ELSE
            BEGIN
              FOR j:=1 TO Length DO
                u[j] := RB[i,j];
              v := c[i];
            END; { if ii }

          { Annihilate u by GIVENS-rotations with the rows of Rmu: }

          FOR j := 1 TO Length DO
            BEGIN

              IF u[j] <> 0.0 THEN
                BEGIN
                iplusj_1 := i+j-1;
                setup_GIVENS_rotation( Rmu[ iplusj_1, 1], u[j],xi);
                                     { to annihilate u[j] }
                apply_GIVENS_rotation(Length-j+1,Rmu[iplusj_1,1],u[j]);

                apply_GIVENS_rotation( 1, d[ iplusj_1],v);

              END{ if }

            END; { for j }
          { update residual }
          RES:=RES+Sqr(v);

        END; { for ii }
    END; { for i }

  RES:=Sqrt(RES);

END; { Givens_factorization_of_RBS }

{-----------------------------------------------------------------------------}

PROCEDURE Compute_Smoothing_Spline(
             m,                       { number of datavalues                  }
             n,                       { number of B-Splines                   }
             k    : integer;          { order of B-Splines                    }
        VAR t     : knotsequence;
             r    : integer;          { order of derivative in smoothing term }
            mu    : real;             { smoothing parameter                   }
        VAR tau   : dataabscissae;
        VAR y     : datavalues;
        VAR S     : SmoothingMatrix;  { Smoothing Matrix                      }
        VAR alpha : coefficients;     { to compute                            }
        VAR RES   : real              { residual sum                              }
        );

VAR i       : integer;
    RES1    : real;
    c,d,G2  : coefficients;
    B,Rmu   : leastsquarematrix;
    Q       : OrthogonalMatrix;

BEGIN
  { QR-factorization of leastsquarematrix B }
  GIVENS_factorization(m,n,k,t,tau,B,Q);
  { B contains the upper triangular matrix R of the QR-factorization in
  packed form }

  { apply the same transformation to the righthandside vector y }
  apply_GIVENS_transformation(m,n,k,t,tau,y,Q,c,RES);
  { c contains now Q_1^T y }

  { GIVENS factorization of
  [       R_B     ]          [ c ]
  [               ] \alpha - [   ]
  [  \sqrt(\mu) S ]          [ 0 ]

  On Output, Rmu contains the upper triangular matrix of the QR
  factorization and d the transformed right hand side
  }
  FOR i:=1 TO n-r DO
    G2[i]:=0.0;
  GIVENS_factorization_of_RBS(n,k,r,mu,B,Rmu,S,c,G2,d,res1);

  RES:=Sqrt(Sqr(RES)+Sqr(RES1));

  { Solve upper triangular system Rmu * alpha = d }
  Rc_x(n,k,Rmu,alpha,d);
END; { Compute_Smoothing_spline }

{-----------------------------------------------------------------------------}

PROCEDURE Get_G (
             m,                       { number of datavalues                  }
             n,                       { number of B-Splines                   }
             k,                       { order of B-Splines                    }
             r    : integer;          { order of derivative in smoothing term }
            mu    : real;             { smoothing parameter                   }
        VAR t     : knotsequence;
        VAR tau   : dataabscissae;
        VAR y     : datavalues;
        VAR alpha : coefficients;     { coefficients of smoothing spline      }
        VAR S     : SmoothingMatrix;
        VAR G     : SmoothingVector   { residual vector of smoothing problem  }
        );

VAR i,j  : integer;
    temp : real;
    F    : datavalues;

BEGIN
  Get_F(m,n,k,t,tau,y,alpha,F);

  FOR i:= 1 TO m DO
    G[i]:=F[i];

  FOR i:=1 TO n-r DO
    BEGIN
      temp:=0.0;
      FOR j:=1 TO r+1 DO
        temp:=temp+S[i,j] * alpha[i+j-1];
      G[m+i]:=-Sqrt(mu)*temp;
    END;
END; { Get_G }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Smoothing_Matrix_Knots_approx(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR S     : SmoothingMatrix;
                 VAR v     : coefficients
                 );

VAR h,temp      : real;
    i,j         : integer;
    ttemp       : knotsequence;
    Stemp       : SmoothingMatrix;

BEGIN
  h:=EPSDiff1*(Abs(t[kk]) + EPSDiff2);
  temp:=t[kk]+h;
  h:=temp-t[kk];
  { trick to improve the accuracy of finite difference approximations }

  FOR i:=1 TO n+k DO
    ttemp[i]:=t[i];

  ttemp[ kk ]:= t[ kk ]+h;
  Compute_Smoothing_Matrix(ttemp,n,k,r,Stemp);

  FOR i:=1 TO n-r DO
    BEGIN
      v[i]:=0.0;
      FOR j:=1 TO r+1 DO
        { Summmation about all nonzero entries of S }
        v[i]:=v[i]+(Stemp[i,j] -S[i,j]) / h * alpha[i+j-1];
        { v = ( S' [e^kk] ) alpha }
      END;
END; { Derivative_Smoothing_Matrix_Knots_approx }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Matrix_Fr_Knots(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR v     : coefficients
                 );
{ computes derivative of auxilliary matrix F_r with respect to t[kk] * alpha
  alpha \in \RealR^(n-r)

  v:= F_r'(t) [e^kk] alpha \in \RealR^(n-r)                                   }

VAR i,j         : integer;
    h           : real;

BEGIN
  h:=0.5/Sqrt(k-r);
  FOR j:=1 TO n-r DO
    BEGIN
      IF j=kk-k THEN
        v[j]:=h/Sqrt(t[k+j]-t[r+j])*alpha[j]
      ELSE
      IF j=kk-r THEN
        v[j]:=-h/Sqrt(t[k+j]-t[r+j])*alpha[j]
      ELSE
        v[j]:=0.0;
    END;
END; { Derivative_Matrix_Fr_Knots }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Smoothing_Matrix_Knots(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR v     : coefficients
                 );

VAR h        : real;
    i,j      : integer;
    v1, v2   : coefficients;
    D        : SmoothingMatrix;

BEGIN
  Derivative_Derivative_Matrix_Knots(n,k,t,r,alpha,kk,v1);
                                    { v1 = D_r' [e^kk] *alpha \in \RealR^(n-r)}

  h:=1/Sqrt(k-r);
  FOR i:=1 TO n-r DO
    v1[i]:= h*Sqrt(t[k+i]-t[r+i])*v1[i];
                                { v1 = F_r D_r' [e^kk] *alpha \in \RealR^(n-r)}


  Compute_Derivative_Matrix(t,n,k,r,D);
  FOR i:=1 TO n-r DO
    BEGIN
      h:=0.0;
      FOR j:=1 TO r+1 DO
        h:=h+D[i,j]*alpha[i+j-1];
      v2[i]:=h;
    END;                                   { v2 = D_r * alpha \in \RealR^(n-r)}

  Derivative_Matrix_Fr_Knots(n,k,t,r,v2,kk,v2);
                                { v2 = F_r'[e^kk] D_r * alpha \in \RealR^(n-r)}

  FOR i:=1 TO n-r DO
    v[i]:=v1[i]+v2[i];               { v = S_r'[e^kk] * alpha \in \RealR^(n-r)}
END; { Derivative_Smoothing_Matrix_Knots }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Matrix_Hr_Knots(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR v     : coefficients
                 );
{ computes derivative of auxilliary matrix H_r with respect to t[kk] * alpha
  alpha \in \RealR^(n-r)

  v:= H_r'(t) [e^kk] alpha \in \RealR^(n-r)                                   }

VAR i,j         : integer;

BEGIN
  FOR j:=1 TO n-r DO
    BEGIN
      IF j=kk-k THEN
        v[j]:=-(k-r)/Sqr(t[k+j]-t[r+j])*alpha[j]
      ELSE
      IF j=kk-r THEN
        v[j]:=(k-r)/Sqr(t[k+j]-t[r+j])*alpha[j]
      ELSE
        v[j]:=0.0;
    END;
END; { Derivative_Matrix_Hr_Knots }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Derivative_Matrix_Knots_approx(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR D     : SmoothingMatrix;
                 VAR v     : coefficients
                 );

VAR h,temp      : real;
    i,j         : integer;
    ttemp       : knotsequence;
    Dtemp       : SmoothingMatrix;

BEGIN
  h:=EPSDiff1*(Abs(t[kk]) + EPSDiff2);
  temp:=t[kk]+h;
  h:=temp-t[kk];
  { trick to improve the accuracy of finite difference approximations }

  FOR i:=1 TO n+k DO
    ttemp[i]:=t[i];

  ttemp[ kk ]:= t[ kk ]+h;
  Compute_Derivative_Matrix(ttemp,n,k,r,Dtemp);

  FOR i:=1 TO n-r DO
    BEGIN
      v[i]:=0.0;
      FOR j:=1 TO r+1 DO
        { Summmation about all nonzero entries of D }
        v[i]:=v[i]+(Dtemp[i,j] -D[i,j]) / h * alpha[i+j-1];
        { v = ( D' [e^kk] ) alpha }
      END;
END; { Derivative_Derivative_Matrix_Knots_approx }

{-----------------------------------------------------------------------------}

PROCEDURE Derivative_Derivative_Matrix_Knots(
                 n,
                 k         : integer;
                 VAR t     : knotsequence;
                 r         : integer;
                 VAR alpha : coefficients;
                 kk        : integer;
                 VAR v     : coefficients
                 );

VAR h        : real;
    i,j,nu   : integer;
    v1, v2   : coefficients;
    D        : SmoothingMatrix;

BEGIN
  FOR i:=1 TO n DO
    v[i]:=0.0;

  FOR nu:=1 TO r DO
    BEGIN
      Compute_Derivative_Matrix(t,n,k,nu-1,D);
      FOR i:=1 TO n-nu+1 DO
        BEGIN
          h:=0.0;
          FOR j:=1 TO nu DO
            h:=h+D[i,j]*alpha[i+j-1];
          v1[i]:=h;
        END;                         { v1 = D_(nu-1)*alpha \in \RealR^(n-nu+1)}

      FOR i:=1 TO n-nu DO
        v1[i]:=v1[i+1]-v1[i];     { v1 = L_nu D_(nu-1)*alpha \in \RealR^(n-nu)}

      Derivative_Matrix_Hr_Knots(n,k,t,nu,v1,kk,v2);
      FOR i:=1 TO n-nu DO
        v1[i]:=v2[i]; { v1 = H_nu'[e^kk] L_nu D_(nu-1)*alpha \in \RealR^(n-nu)}

      FOR i:=1 TO n-nu+1 DO
        v2[i]:=v[i];        { v2 = D_(nu-1)'[e^kk] * alpha \in \RealR^(n-nu+1)}

      FOR i:=1 TO n-nu DO
        BEGIN
          h:=t[k+i]-t[nu+i];
          h:=(k-nu)/h;
          v2[i]:=h*(v2[i+1]-v2[i]);
        END;        { v2 = H_nu L_nu D_(nu-1)'[e^kk] * alpha \in \RealR^(n-nu)}

      FOR i:=1 TO n-nu DO
        v[i]:=v1[i]+v2[i];         { v = D_nu'[e^kk] * alpha \in \RealR^(n-nu)}
    END;

END; { Derivative_Derivative_Matrix_Knots }

{-----------------------------------------------------------------------------}

PROCEDURE Get_STvec (
             m,                       { number of datavalues                  }
             n,                       { number of B-Splines                   }
             k,                       { order of B-Splines                    }
             r    : integer;          { order of derivative in smoothing term }
        VAR t     : knotsequence;
        VAR v, w  : coefficients;
        VAR S     : SmoothingMatrix
        );

var i,j,ind : integer;
Begin
  for j:=1 to n do
    w[j]:=0.0;

  FOR i:=1 TO n-r DO
    FOR j:= 1 TO r+1 DO
      BEGIN
        ind:=i+j-1;
        w[ind]:=w[ind]+S[i,j]*v[i];
      END;
end; { Get_STvec }

{-----------------------------------------------------------------------------}

END. { unit Smooth }